/* $Id$ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

#ifndef __PYTHON_HH__
#define __PYTHON_HH__

#include <iosfwd>
#include <string>
#include <vector>

#include "base/str.hh"
#include "sim/host.hh"

class Python
{
  private:
    std::ostream &stream;
    int level;
    bool endline;
    char close;
    bool comma;

  protected:
    static inline std::string quote(const std::string &data)
    { return "\'" + data + "\'"; }
    static inline std::string qqquote(const std::string &data)
    { return "\'\'\'" + data + "\'\'\'"; }

    static inline std::string c2py(const std::string &str) { return str; }
    static inline std::string c2py(bool val) { return val ? "True" : "False"; }
    static inline std::string c2py(int32_t val) { return to_string(val); }
    static inline std::string c2py(uint32_t val) { return to_string(val); }
    static inline std::string c2py(int64_t val) { return to_string(val); }
    static inline std::string c2py(uint64_t val) { return to_string(val); }
    static std::string c2py(double val);

    template <class C>
    static inline std::string
    c2tuple(const C &container, bool q)
    {
	using namespace std;
	typedef typename C::const_iterator iter;

	iter i = container.begin();
	iter end = container.end();

	string str = "(";
	while (i != end) {
	    iter c = i;
	    ++i;

	    string s = c2py(*c);
	    if (q)
		s = quote(s);
	    str += s;
	    if (i != end)
		str += ", ";
	}

	str += ")";
	return str;
    }

    void next();

  public:
    explicit Python(std::ostream &s);
    ~Python();

    void newline();

    void name(const std::string &name);
    void nameEnd();

    void tuple(const std::string &name = "");
    void tupleEnd();

    void list();
    void listEnd();

    void arg(const std::string &val);
    void qarg(const std::string &val)
    { arg(quote(val)); }

    void qqqarg(const std::string &val)
    { arg(qqquote(val)); }

    template <class T>
    void arg(const std::vector<T>& val)
    { arg(c2tuple(val, false)); }

    template <class T>
    void qarg(const std::vector<T>& val)
    { arg(c2tuple(val, true)); }

    template <class T>
    void arg(const T &val)
    { arg(c2py(val)); }

    template <class T>
    void qarg(const T &val)
    { arg(quote(c2py(val))); }

    void kwarg(const std::string &name, const std::string &val);
    void qkwarg(const std::string &name, const std::string &val)
    { kwarg(name, quote(val)); }

    template <class T>
    void kwarg(const std::string &name, const std::vector<T>& val)
    { kwarg(name, c2tuple(val, false)); }

    template <class T>
    void qkwarg(const std::string &name, const std::vector<T>& val)
    { kwarg(name, c2tuple(val, true)); }

    template <class T>
    void kwarg(const std::string &name, const T &val)
    { kwarg(name, c2py(val)); }

    template <class T>
    void qkwarg(const std::string &name, const T &val)
    { kwarg(name, quote(c2py(val))); }
};

#endif // __PYTHON_HH__
