/* $Id$ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

#include <fstream>

#include "base/cprintf.hh"
#include "base/misc.hh"
#include "base/python.hh"
#include "base/statistics.hh"
#include "base/stats/python.hh"
#include "base/stats/statdb.hh"

using namespace std;

typedef Python BasePython;

namespace Stats {

struct PythonData
{
    BasePython py;
    explicit PythonData(ostream &stream) : py(stream) {}
};

Python::Python()
    : mystream(false), stream(NULL), pydata(NULL)
{
}

Python::Python(std::ostream &stream)
    : mystream(false), stream(NULL), pydata(NULL)
{
    open(stream);
}

Python::Python(const std::string &file)
    : mystream(false), stream(NULL), pydata(NULL)
{
    open(file);
}

Python::~Python()
{
    if (stream) {
	ccprintf(*stream, "\n\nif __name__ == '__main__':\n");
	ccprintf(*stream, "    program_display()\n");
    }

    if (pydata)
	delete pydata;

    if (mystream) {
	assert(stream);
	delete stream;
    }
}

void
Python::open()
{
    Time now(true);
    date = now.date();

    pydata = new PythonData(*stream);
    ccprintf(*stream, "import sys\n");
    ccprintf(*stream, "sys.path.append('.')\n");
    ccprintf(*stream, "from m5stats import *\n\n");
}

void
Python::open(std::ostream &_stream)
{
    if (stream)
	panic("stream already set!");

    mystream = false;
    stream = &_stream;

    open();
}

void
Python::open(const std::string &file)
{
    if (stream)
	panic("stream already set!");

    mystream = true;
    stream = new ofstream(file.c_str(), ios::trunc);

    open();
}

bool
Python::valid() const
{
    return stream != NULL;
}

void
Python::output(const string &bin)
{
    using namespace Database;
    BasePython &py = pydata->py;

    py.name("collections.append");
    py.newline();
    py.name("Collection");
    py.newline();
    py.qarg(simname);
    py.newline();
    py.qarg(curTick);
    py.newline();
    py.qarg(bin);
    py.newline();
    py.qarg(simhost);
    py.newline();
    py.qarg(project);
    py.newline();
    py.qarg(date);
    py.newline();
    py.list();

    stat_list_t::iterator i, end = stats().end();
    for (i = stats().begin(); i != end; ++i)
	(*i)->visit(*this);

    py.newline();
    py.listEnd();
    py.newline();
    py.nameEnd();
    py.newline();
    py.nameEnd();
    py.newline();
}

void
Python::output()
{
    using namespace Database;
    assert(valid());

    if (bins().empty()) {
	output(string(""));
    } else {
	bin_list_t::iterator i, end = bins().end();
	for (i = bins().begin(); i != end; ++i) {
	    MainBin *bin = *i;
	    bin->activate();
	    output(bin->name());
	}
    }
}

void
Python::visit(const ScalarData &data)
{
    BasePython &py = pydata->py;

    py.newline();
    py.name("Scalar");
    py.qarg(data.name);
    py.qqqarg(data.desc);
    py.kwarg("binned", data.binned());
    py.kwarg("precision", data.precision);
    py.kwarg("flags", data.flags);
    if (data.prereq)
	py.qkwarg("prereq", data.prereq->name);
    py.kwarg("value", data.value());
    py.nameEnd();
}

void
Python::visit(const VectorData &data)
{
    BasePython &py = pydata->py;

    py.newline();
    py.name("Vector");
    py.qarg(data.name);
    py.qqqarg(data.desc);
    py.kwarg("binned", data.binned());
    py.kwarg("precision", data.precision);
    py.kwarg("flags", data.flags);
    if (data.prereq)
	py.qkwarg("prereq", data.prereq->name);
    py.kwarg("value", data.value());
    if (!data.subnames.empty())
	py.qkwarg("subnames", data.subnames);
    if (!data.subdescs.empty())
	py.qkwarg("subdescs", data.subdescs);
    py.nameEnd();
}

void
Python::distData(const DistDataData &data, const string &name) const
{
    string s = name.empty() ? "" : name + "=";

    if (data.samples == 0 || data.fancy)
	s += "SimpleDist";
    else
	s += "FullDist";

    BasePython &py = pydata->py;

    py.name(s);
    py.arg(data.sum);
    py.arg(data.squares);
    py.arg(data.samples);
    if (data.samples && !data.fancy) {
	py.arg(data.min_val);
	py.arg(data.min_val);
	py.arg(data.underflow);
	py.arg(data.cvec);
	py.arg(data.overflow);
	py.arg(data.min);
	py.arg(data.max);
	py.arg(data.bucket_size);
	py.arg(data.size);
    }
    py.nameEnd();
}

void
Python::visit(const DistData &data)
{
    BasePython &py = pydata->py;

    py.newline();
    py.name("Dist");
    py.qarg(data.name);
    py.qqqarg(data.desc);
    py.kwarg("binned", data.binned());
    py.kwarg("precision", data.precision);
    py.kwarg("flags", data.flags);
    if (data.prereq)
	py.qkwarg("prereq", data.prereq->name);
    distData(data.data, "dist");
    py.nameEnd();
}

void
Python::visit(const VectorDistData &data)
{
    BasePython &py = pydata->py;

    py.newline();
    py.name("VectorDist");
    py.qarg(data.name);
    py.qqqarg(data.desc);
    py.kwarg("binned", data.binned());
    py.kwarg("precision", data.precision);
    py.kwarg("flags", data.flags);
    if (data.prereq)
	py.qkwarg("prereq", data.prereq->name);
    if (!data.subnames.empty())
	py.qkwarg("subnames", data.subnames);
    if (!data.subdescs.empty())
	py.qkwarg("subdescs", data.subdescs);

    py.tuple("dist");
    vector<DistDataData>::const_iterator i, end = data.data.end();
    for (i = data.data.begin(); i != end; ++i)
	distData(*i, "");

    py.tupleEnd();
    py.nameEnd();
}

void
Python::visit(const Vector2dData &data)
{
    BasePython &py = pydata->py;

    py.newline();
    py.name("Vector2d");
    py.qarg(data.name);
    py.qqqarg(data.desc);
    py.kwarg("binned", data.binned());
    py.kwarg("precision", data.precision);
    py.kwarg("flags", data.flags);
    if (data.prereq)
	py.qkwarg("prereq", data.prereq->name);

    py.kwarg("value", data.cvec);
    if (!data.subnames.empty())
	py.qkwarg("subnames", data.subnames);
    if (!data.subdescs.empty())
	py.qkwarg("subdescs", data.subdescs);
    if (!data.y_subnames.empty())
	py.qkwarg("ysubnames", data.y_subnames);

    py.kwarg("x", data.x);
    py.kwarg("y", data.y);
    py.nameEnd();
}

void
Python::visit(const FormulaData &data)
{
    BasePython &py = pydata->py;

    py.newline();
    py.name("Formula");
    py.qarg(data.name);
    py.qqqarg(data.desc);
    py.kwarg("binned", data.binned());
    py.kwarg("precision", data.precision);
    py.kwarg("flags", data.flags);
    if (data.prereq)
	py.qkwarg("prereq", data.prereq->name);
    py.qkwarg("formula", data.str());
    if (!data.subnames.empty())
	py.qkwarg("subnames", data.subnames);
    if (!data.subdescs.empty())
	py.qkwarg("subdescs", data.subdescs);
    py.nameEnd();
}

/* namespace Stats */ }
