/* $Id: s.exec_context.cc 1.60 04/05/28 11:29:33-04:00 ktlim@zamp.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

#include <string>

#include "cpu/base_cpu.hh"
#include "cpu/exec_context.hh"

#ifdef FULL_SYSTEM
#include "sim/system.hh"
#else
#include "sim/process.hh"
#endif

using namespace std;

// constructor
#ifdef FULL_SYSTEM
ExecContext::ExecContext(BaseCPU *_cpu, int _thread_num, System *_sys,
			 AlphaITB *_itb, AlphaDTB *_dtb,
			 FunctionalMemory *_mem)
    : _status(ExecContext::Unallocated),
      kernelStats(this, _cpu), cpu(_cpu), thread_num(_thread_num),
      cpu_id(-1), mem(_mem), itb(_itb), dtb(_dtb), system(_sys),
      memCtrl(_sys->memCtrl), physmem(_sys->physmem),
      swCtx(NULL), func_exe_inst(0), storeCondFailures(0)
{
    memset(&regs, 0, sizeof(RegFile));
}
#else
ExecContext::ExecContext(BaseCPU *_cpu, int _thread_num,
			 Process *_process, int _asid)
    : _status(ExecContext::Unallocated),
      cpu(_cpu), thread_num(_thread_num), cpu_id(-1),
      process(_process), mem(process->getMemory()), asid(_asid),
      func_exe_inst(0), storeCondFailures(0)
{
    memset(&regs, 0, sizeof(RegFile));
}

ExecContext::ExecContext(BaseCPU *_cpu, int _thread_num,
			 FunctionalMemory *_mem, int _asid)
    : cpu(_cpu), thread_num(_thread_num), process(0), mem(_mem), asid(_asid),
      func_exe_inst(0), storeCondFailures(0)
{
    memset(&regs, 0, sizeof(RegFile));
}
#endif


void
ExecContext::takeOverFrom(ExecContext *oldContext)
{
    // some things should already be set up
    assert(mem == oldContext->mem);
#ifdef FULL_SYSTEM
    assert(system == oldContext->system);
#else
    assert(process == oldContext->process);
#endif

    // copy over functional state
    _status = oldContext->_status;
#ifdef FULL_SYSTEM
    kernelStats = oldContext->kernelStats;
#endif
    regs = oldContext->regs;
    cpu_id = oldContext->cpu_id;
    func_exe_inst = oldContext->func_exe_inst;

    storeCondFailures = 0;

    oldContext->_status = ExecContext::Unallocated;
}


void
ExecContext::serialize(ostream &os)
{
    SERIALIZE_ENUM(_status);
    regs.serialize(os);
    // thread_num and cpu_id are deterministic from the config
    SERIALIZE_SCALAR(func_exe_inst);
    SERIALIZE_SCALAR(inst);

#ifdef FULL_SYSTEM
    bool ctx = false;
    if (swCtx) {
	ctx = true;
	SERIALIZE_SCALAR(ctx);
	SERIALIZE_SCALAR(swCtx->calls);
	std::stack<fnCall *> *stack = &(swCtx->callStack);
	fnCall *top;
	int size = stack->size();
	SERIALIZE_SCALAR(size);

	for (int j=0; j<size; ++j) {
	    top = stack->top();
	    paramOut(os, csprintf("stackpos[%d]",j), top->name);    
	    delete top;
	    stack->pop();
	}	
    } else {
	SERIALIZE_SCALAR(ctx);
    }
    if (system->bin) {
	Stats::MainBin *cur = Stats::MainBin::curBin();	
	string bin_name = cur->name();
	SERIALIZE_SCALAR(bin_name);
    }
#endif //FULL_SYSTEM
}


void
ExecContext::unserialize(Checkpoint *cp, const std::string &section)
{
    UNSERIALIZE_ENUM(_status);
    regs.unserialize(cp, section);
    // thread_num and cpu_id are deterministic from the config
    UNSERIALIZE_SCALAR(func_exe_inst);
    UNSERIALIZE_SCALAR(inst);

#ifdef FULL_SYSTEM
    bool ctx;
    UNSERIALIZE_SCALAR(ctx);
    if (ctx) {
	swCtx = new SWContext;
	UNSERIALIZE_SCALAR(swCtx->calls);
	int size;
	UNSERIALIZE_SCALAR(size);

	vector<fnCall *> calls;
	fnCall *call;
	for (int i=0; i<size; ++i) {
	    call = new fnCall;
	    paramIn(cp, section, csprintf("stackpos[%d]",i), call->name);
	    call->myBin = system->getBin(call->name);
	    calls.push_back(call);
	}

	for (int i=size-1; i>=0; --i) {
	    swCtx->callStack.push(calls[i]);
	}
	
    }

    if (system->bin) {
	string bin_name;
	UNSERIALIZE_SCALAR(bin_name);
	system->getBin(bin_name)->activate();	
    }
#endif //FULL_SYSTEM
}


void
ExecContext::activate(int delay)
{
    if (status() == Active)
	return;

    _status = Active;
    cpu->activateContext(thread_num, delay);
}

void
ExecContext::suspend()
{
    if (status() == Suspended)
	return;

#ifdef FULL_SYSTEM
    // Don't change the status from active if there are pending interrupts
    if (cpu->check_interrupts()) {
	assert(status() == Active);
	return;
    }
#endif

    _status = Suspended;
    cpu->suspendContext(thread_num);
}

void
ExecContext::deallocate()
{
    if (status() == Unallocated)
	return;

    _status = Unallocated;
    cpu->deallocateContext(thread_num);
}

void
ExecContext::halt()
{
    if (status() == Halted)
	return;

    _status = Halted;
    cpu->haltContext(thread_num);
}


void
ExecContext::regStats(const string &name)
{
#ifdef FULL_SYSTEM
    kernelStats.regStats(name + ".kern");
#endif
}

void
ExecContext::trap(Fault fault)
{
    //TheISA::trap(fault);    //One possible way to do it...

    /** @todo: Going to hack it for now.  Do a true fixup later. */
#ifdef FULL_SYSTEM
    ev5_trap(fault);
#else
    fatal("fault (%d) detected @ PC 0x%08p", fault, readPC());
#endif
}
