/* $Id: s.dyn_inst.cc 1.60 04/03/04 14:57:56-05:00 ehallnor@zizzer.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

#include <iostream>
#include <string>
#include <sstream>

#include "base/cprintf.hh"
#include "cpu/full_cpu/full_cpu.hh"
#include "cpu/full_cpu/dyn_inst.hh"
#include "cpu/exetrace.hh"
#include "cpu/full_cpu/fetch.hh"
#include "mem/mem_req.hh"
#include "cpu/full_cpu/spec_state.hh"

using namespace std;

#define NOHASH
#ifndef NOHASH

#include "base/hashmap.hh"

unsigned int MyHashFunc(const DynInst *addr)
{
  unsigned a = (unsigned)addr;
  unsigned hash = (((a >> 14) ^ ((a >> 2) & 0xffff))) & 0x7FFFFFFF;

  return hash;
}

typedef m5::hash_map<const DynInst *, const DynInst *, MyHashFunc> my_hash_t;
my_hash_t thishash;
#endif


int DynInst::instcount = 0;

//int break_inst = -1;

void
DynInst::squash()
{

    if (isCondCtrl())
	update_thread_confidence(thread_number, &dir_update);

    if (blocking_inst) {
	cpu->thread_unblock(thread_number);
	blocking_inst = false;
    }

    if (recover_inst) {
	xc->spec_mode--;
	recover_inst = false;
    }
}


void
DynInst::setCPSeq(InstSeqNum seq) {
    correctPathSeq = seq;
    if (trace_data) {
	trace_data->setCPSeq(seq);
    }
}


void
do_break_inst(int print = 0)
{
    if (print)
	cprintf("Break here\n");
}


DynInst::DynInst(StaticInstPtr<TheISA> &_staticInst)
    : staticInst(_staticInst), trace_data(NULL)
{
    /**
     * @todo this actually is in error since we're incorrectly reading
     * the variable valid before it is set  (the first time it is newed)
     */
    if (valid == 0x1234abcd)
	fatal("Not freed properly\n");

    if (valid == 0xdcba4321)
	do_break_inst();

    eff_addr = MemReq::inval_addr;

    valid = 0x1234abcd;
    spec_mem_write = false;

    blocking_inst = false;
    recover_inst = false;
    spec_mode = false;
    btb_missed = false;

#ifndef NOHASH
    thishash.insert(this, this);

    if (++instcount > 2048) {
	my_hash_t::iterator iter = thishash.begin();
	my_hash_t::iterator end = thishash.end();
	while (iter != end) {
	    cprintf("hash_t, addr = %#x\n"
		    "        valid = %#08x\n"
		    "        spec_mode = %d\n",
		    (void*)*iter, (*iter).second->valid,
		    (*iter).second->spec_mode);
	    ++iter;
	}
	//thishash.DumpData();
	panic("Too many insts.\n\tcycle = %n\n\tdeletecount = %d\n"
	      "\tinstcount = %d\n", curTick, deletecount, instcount);
    }
#endif
}

DynInst::~DynInst()
{
    if (valid != 0x1234abcd)
	panic("Invalid Instruction!\n");
    valid = 0xdcba4321;

    if (spec_mem_write) {
	// Remove effects of this instruction from speculative memory
	xc->spec_mem->erase(eff_addr);
    }

#ifndef NOHASH
    my_hash_t::iterator i = thishash.find(this);
    thishash.Remove(i);
    --instcount;
#endif
}

FunctionalMemory *DynInst::getMemory(void)
{
    return (spec_mode ? xc->spec_mem : xc->mem);
}

IntReg *DynInst::getIntegerRegs(void)
{
    return (spec_mode ? xc->specIntRegFile : xc->regs.intRegFile);
}


void
DynInst::prefetch(Addr addr, unsigned flags)
{
    // This is the "functional" implementation of prefetch.  Not much
    // happens here since prefetches don't affect the architectural
    // state.

    // Generate a MemReq so we can translate the effective address.
    MemReqPtr req = new MemReq(addr, xc, 1, flags);
    req->asid = asid;

    // Prefetches never cause faults.
    fault = No_Fault;

    // note this is a local, not DynInst::fault
    Fault trans_fault = xc->translateDataReadReq(req);

    if (trans_fault == No_Fault) {
	// It's a valid address.  Record key MemReq parameters so we
	// can generate another one just like it for the timing access
	// without calling translate() again (which might mess up the
	// TLB).
	eff_addr = req->vaddr;
	phys_eff_addr = req->paddr;
	mem_req_flags = req->flags;
    } else {
	// bogus address.  Normally DynInst::fault would get set, but
	// we don't want the prefetch to cause a fault.  We mark it by
	// setting the eff_addr to InvalidAddr.
	eff_addr = phys_eff_addr = MemReq::inval_addr;
    }

    /**
     * @todo
     * Replace the disjoint functional memory with a unified one and remove
     * this hack.
     */
#ifndef FULL_SYSTEM
    req->paddr = req->vaddr;
#endif

    if (trace_data) {
	trace_data->setAddr(addr);
    }
}

void
DynInst::writeHint(Addr addr, int size)
{
    // Need to create a MemReq here so we can do a translation.  This
    // will casue a TLB miss trap if necessary... not sure whether
    // that's the best thing to do or not.  We don't really need the
    // MemReq otherwise, since wh64 has no functional effect.
    MemReqPtr req = new MemReq(addr, xc, size, 0);
    req->asid = asid;

    fault = xc->translateDataWriteReq(req);

    if (fault != No_Fault) {
	// ignore faults... treat as no-op
	eff_addr = phys_eff_addr = MemReq::inval_addr;
    } else {
	// Record key MemReq parameters so we can generate another one
	// just like it for the timing access without calling translate()
	// again (which might mess up the TLB).
	eff_addr = req->vaddr;
	phys_eff_addr = req->paddr;
	mem_req_flags = req->flags;
    }

    store_size = size;
    store_data = 0;
}

/**
 * @todo Need to find a way to get the cache block size here.
 */
Fault
DynInst::copySrcTranslate(Addr src)
{
    MemReqPtr req = new MemReq(src, xc, 64);
    req->asid = asid;

    // translate to physical address
    Fault fault = xc->translateDataReadReq(req);
    
    if (fault == No_Fault) {
	xc->copySrcAddr = src;
	xc->copySrcPhysAddr = req->paddr;
    } else {
	xc->copySrcAddr = 0;
	xc->copySrcPhysAddr = 0;
    }
    return fault;
}

/**
 * @todo Need to find a way to get the cache block size here.
 */
Fault
DynInst::copy(Addr dest)
{
    uint8_t data[64];
    FunctionalMemory *mem = (xc->misspeculating()) ? xc->spec_mem : xc->mem;
    assert(xc->copySrcPhysAddr || xc->misspeculating());
    MemReqPtr req = new MemReq(dest, xc, 64);
    req->asid = asid;

    // translate to physical address
    Fault fault = xc->translateDataWriteReq(req);

    if (fault == No_Fault) {
	Addr dest_addr = req->paddr;
	// Need to read straight from memory since we have more than 8 bytes.
	req->paddr = xc->copySrcPhysAddr;
	mem->read(req, data);
	req->paddr = dest_addr;
	mem->write(req, data);
    }
    return fault;
}

void
DynInst::dump()
{
    cprintf("T%d : %#08d `", thread_number, PC);
    cout << staticInst->disassemble(PC);
    cprintf("'\n");
}

void DynInst::dump(std::string &outstring)
{
    std::ostringstream s;
    s << "T" << thread_number << " : 0x" << PC << " "
      << staticInst->disassemble(PC);

    outstring = s.str();
}


#if 0
Fault
DynInst::mem_access(mem_cmd cmd, Addr addr, void *p, int nbytes)
{
    Fault fault;

    // check alignments, even speculative this test should always pass
    if ((nbytes & nbytes - 1) != 0 || (addr & nbytes - 1) != 0) {
	for (int i = 0; i < nbytes; i++)
	    ((char *) p)[i] = 0;

	// I added the following because according to the comment above,
	// we should never get here.  The comment lies
#if 0
	panic("unaligned access. Cycle = %n", curTick);
#endif
	return No_Fault;
    }

    MemReqPtr req = new MemReq(addr, thread, nbytes);
    switch(cmd) {
      case Read:
	fault = spec_mem->read(req, (uint8_t *)p);
	break;

      case Write:
	fault = spec_mem->write(req, (uint8_t *)p);
	if (fault != No_Fault)
	    break;

	spec_mem_write = true;
	store_size = nbytes;
	switch(nbytes) {
	  case sizeof(uint8_t):
	    *(uint8_t)&store_data = (uint8_t *)p;
	    break;
	  case sizeof(uint16_t):
	    *(uint16_t)&store_data = (uint16_t *)p;
	    break;
	  case sizeof(uint32_t):
	    *(uint32_t)&store_data = (uint32_t *)p;
	    break;
	  case sizeof(uint64_t):
	    *(uint64_t)&store_data = (uint64_t *)p;
	    break;
	}
	break;

      default:
	fault = Machine_Check_Fault;
	break;
    }

    trace_mem(fault, cmd, addr, p, nbytes);

    return fault;
}
#endif
