/* $Id$ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

#include "base/statistics.hh"
#include "base/trace.hh"
#include "cpu/base_cpu.hh"
#include "cpu/sampling_cpu/sampling_cpu.hh"
#include "sim/builder.hh"
#include "sim/serialize.hh"
#include "sim/sim_events.hh"

using namespace std;

class CpuSwitchEvent : public Event
{
  private:
    
    SamplingCPU *sampler;
    
  public:
    
    CpuSwitchEvent(Tick _when, SamplingCPU *sampler);
    CpuSwitchEvent(SamplingCPU *sampler);
    virtual ~CpuSwitchEvent() {}
    
    void process();
    
    void serialize(std::ostream &os);
    void unserialize(Checkpoint *cp, const std::string &section);
    static Serializable *createForUnserialize(Checkpoint *cp, 
					       const std::string &section);
};


CpuSwitchEvent::CpuSwitchEvent(Tick when, SamplingCPU *_sampler)
    : Event(&mainEventQueue, CPU_Switch_Pri), sampler(_sampler)
{
    DPRINTF(Sampler, "New switch event curTick=%d when=%d\n", curTick, when);
    setFlags(AutoDelete);
    schedule(when);
}

//non-scheduling version for createForUnserialized()
CpuSwitchEvent::CpuSwitchEvent(SamplingCPU *_sampler)
    : Event(&mainEventQueue), sampler(_sampler)
{
    setFlags(AutoDelete);
}

void
CpuSwitchEvent::process()
{
    sampler->switchCPUs();
}

void
CpuSwitchEvent::serialize(ostream &os)
{
    paramOut(os, "type", string("CpuSwitchEvent"));
    Event::serialize(os);
    SERIALIZE_OBJPTR(sampler);
}

void
CpuSwitchEvent::unserialize(Checkpoint *cp, const string &section)
{
    Event::unserialize(cp,section);
}

Serializable *
CpuSwitchEvent::createForUnserialize(Checkpoint *cp, 
				     const string &section)
{
    SamplingCPU *sampler;
    UNSERIALIZE_OBJPTR(sampler);
    return new CpuSwitchEvent(sampler);
}

REGISTER_SERIALIZEABLE("CpuSwitchEvent", CpuSwitchEvent)

SamplingCPU::SamplingCPU(const std::string &_name,
			 vector<BaseCPU *> &_cpus, vector<Tick> &_periods)
    : SimObject(_name), cpus(_cpus), periods(_periods), curIndex(0)
{
    assert(curTick == 0);
}

SamplingCPU::~SamplingCPU()
{
}

void
SamplingCPU::init()
{
    // set up only the first CPU to run
    cpus[0]->registerExecContexts();
    new CpuSwitchEvent(curTick + periods[0], this);
}
/**
 * @todo This is just a quick hack that makes the simulator exit after
 * the last sampling phase, but the intent is to have it repeat the
 * samples.  In order to do this, we need to make the sampling time
 * be relative to the inital curTick on an unserialization, we should
 * also put in a periods parameter to prevent it from running forever.
 */
void
SamplingCPU::switchCPUs()
{
    DPRINTF(Sampler, "switching CPUs curIndex=%d\n", curIndex);
    int oldIndex = curIndex;
    if (++curIndex >= periods.size()) {
	new SimExitEvent(curTick, "Done Sampling");
//	curIndex = 0;
    } else {
	cpus[curIndex]->takeOverFrom(cpus[oldIndex]);

	Stats::reset();

	new CpuSwitchEvent(curTick + periods[curIndex], this);
    }
}

void
SamplingCPU::serialize(ostream &os)
{
    SERIALIZE_SCALAR(curIndex);
}

void
SamplingCPU::unserialize(Checkpoint *cp, const string &section)
{
    UNSERIALIZE_SCALAR(curIndex);
}

////////////////////////////////////////////////////////////////////////
//
//  SamplingCPU Simulation Object
//
//  Since SamplingCPU isn't really a CPU at all, but a proxy for the
//  set of CPUs we'll be switching among, we have to declare these
//  SimObject functions as applying to BaseCPU.  This is really a minor
//  bug in how these macros work.  The whole SimObject builder thing
//  needs an overhaul...
//

BEGIN_DECLARE_SIM_OBJECT_PARAMS(SamplingCPU)

    SimObjectVectorParam<BaseCPU *> cpus;
    VectorParam<Tick> periods;

END_DECLARE_SIM_OBJECT_PARAMS(SamplingCPU)

BEGIN_INIT_SIM_OBJECT_PARAMS(SamplingCPU)

    INIT_PARAM(cpus, "vector of actual CPUs to sample among"),
    INIT_PARAM(periods, "vector of per-cpu sample periods")

END_INIT_SIM_OBJECT_PARAMS(SamplingCPU)


CREATE_SIM_OBJECT(SamplingCPU)
{
    if (cpus.size() != periods.size()) {
	fatal("SamplingCPU: 'cpus' and 'periods' vector lengths must match");
    }

    return new SamplingCPU(getInstanceName(), cpus, periods);
}

REGISTER_SIM_OBJECT("SamplingCPU", SamplingCPU)
