/* $Id: s.dma.cc 1.60 04/03/02 02:49:16-05:00 rdreslin@zamp.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* @file
 * Device dma access definition
 */

#include <climits>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <deque>
#include <string>
#include <vector>

#include "base/misc.hh"
#include "base/trace.hh"
#include "cpu/full_cpu/smt.hh"
#include "dev/dma.hh"
#include "mem/bus/bus.hh"
#include "mem/bus/dma_interface.hh"
#include "mem/functional_mem/physical_memory.hh"
#include "sim/builder.hh"
#include "sim/eventq.hh"
#include "sim/sim_object.hh"

using namespace std;

#if TRACING_ON
Counter DmaRequest::numbers = 0;
#endif

// Initialize all of the DMA channels
DmaEngine::DmaEngine(const string &name, Bus *b, PhysicalMemory *pmem,
		     unsigned nchan)
    : SimObject(name), queue_len(nchan), bus(b), interface(NULL),
      physmem(pmem), processing(false)
{
    if (bus)
	interface = new DMAInterface<Bus>(name + ".dma", bus, bus, 1);
}

DmaEngine::~DmaEngine()
{
    if (interface)
	delete interface;
}

// Perform the specified DMA operation.
void
DmaEngine::doTransfer(DmaRequest *req)
{
#ifndef NDEBUG
    req->pending = true;
#endif
    DPRINTF(DMA, "doTransfer req=%d queue size=%d\n",
	    req->number(), dma_queue.size());
    if (dma_queue.size() >= queue_len)
	panic("DMA queue full\n");

    dma_queue.push_back(req);
    process();
}

void
DmaEngine::serialize(std::ostream &os)
{
    int numReqs=0;
    for (queue_t::iterator it=dma_queue.begin(); it != dma_queue.end();
         ++it) {
        paramOut(os, csprintf("dmaReq%d", numReqs++), (*it)->name());    
    }
    SERIALIZE_SCALAR(numReqs);
}

void
DmaEngine::unserialize(Checkpoint *cp, const std::string &section)
{
    int numReqs;

    UNSERIALIZE_SCALAR(numReqs);

    std::string dmaReqName;
    uint32_t reqId;
    Serializable *reqSer;
    DmaHolder *reqParent;
    DmaRequest *reqPtr;

    for (int i=0; i < numReqs; i++) {
        // read the name of the dma request from our section
        paramIn(cp, section, csprintf("dmaReq%d", i), dmaReqName);
        
        // read the request's parent and id from its global section
        objParamIn(cp, dmaReqName, csprintf("parent"), reqSer); 
        paramIn(cp, dmaReqName, csprintf("id"), reqId);

        // recast the parent pointer as a DmaHolder and query w/ id
        reqParent = dynamic_cast<DmaHolder *>(reqSer);
        reqPtr = reqParent->find_dmareq(reqId);

        if (reqPtr)
            dma_queue.push_back(reqPtr);
    }
}

class DmaRequestEvent : public Event
{
  protected:
    DmaEngine *engine;

    // non-scheduling version for createForUnserialize()
    DmaRequestEvent(DmaEngine *e)
	: Event(&mainEventQueue), engine(e)
    {
	setFlags(AutoSerialize);
    }

  public:
    DmaRequestEvent(DmaEngine *e, Tick when)
	: Event(&mainEventQueue), engine(e)
    {
	setFlags(AutoSerialize);
	schedule(when);
    }

    void process();
    virtual const char *description();

    virtual void serialize(std::ostream &os);
    virtual void unserialize(Checkpoint *cp, const std::string &section);

    static Serializable *createForUnserialize(Checkpoint *cp,
					       const std::string &section);
};

void
DmaRequestEvent::process()
{
    engine->process();
    delete this;
}


const char *
DmaRequestEvent::description()
{
    return "dma complete";
}


void
DmaRequestEvent::serialize(std::ostream &os)
{
    paramOut(os, "type", string("DmaRequestEvent"));
    Event::serialize(os);
    SERIALIZE_OBJPTR(engine);
}


void
DmaRequestEvent::unserialize(Checkpoint *cp, const std::string &section)
{
    Event::unserialize(cp, section);
}


Serializable *
DmaRequestEvent::createForUnserialize(Checkpoint *cp,
				       const std::string &section)
{
    DmaEngine *engine;
    UNSERIALIZE_OBJPTR(engine);
    return new DmaRequestEvent(engine);
}

REGISTER_SERIALIZEABLE("DmaRequestEvent", DmaRequestEvent)

class DmaTransferEvent : public Event
{
  protected:
    DmaEngine *engine;

  public:
    DmaTransferEvent(DmaEngine *e)
	: Event(&mainEventQueue), engine(e)
    {
	setFlags(AutoSerialize);
    }

    void process();
    virtual const char *description();

    virtual void serialize(std::ostream &os);
    virtual void unserialize(Checkpoint *cp, const std::string &section);

    static Serializable *createForUnserialize(Checkpoint *cp,
					       const std::string &section);
};

void
DmaTransferEvent::process()
{
    DmaRequest *req = engine->dma_queue.front();
    DPRINTF(DMA, "req=%d\n", req->number());
    req->start = curTick;
    engine->process();
    delete this;
}


const char *
DmaTransferEvent::description()
{
    return "dma transfer";
}


void
DmaTransferEvent::serialize(std::ostream &os)
{
    paramOut(os, "type", string("DmaTransferEvent"));
    Event::serialize(os);
    SERIALIZE_OBJPTR(engine);
}


void
DmaTransferEvent::unserialize(Checkpoint *cp, const std::string &section)
{
    Event::unserialize(cp, section);
}

Serializable *
DmaTransferEvent::createForUnserialize(Checkpoint *cp,
				       const std::string &section)
{
    DmaEngine *engine;
    UNSERIALIZE_OBJPTR(engine);
    return new DmaTransferEvent(engine);
}

REGISTER_SERIALIZEABLE("DmaTransferEvent", DmaTransferEvent)

void
DmaEngine::process(Tick when)
{
    if (when < curTick)
	panic("scheduling event in the past");

    if (when == curTick)
	process();

    new DmaRequestEvent(this, when);
}

// Process DMA to start or continue the current request.
void
DmaEngine::process()
{
    if (processing)
	return;

    processing = true;

    if (dma_queue.empty())
	DPRINTF(DMA, "dma_queue empty.  nothing to do\n");

    while (!dma_queue.empty()) {
	DmaRequest *req = dma_queue.front();

	if (req->start == ULL(0x7fffffffffffffff)) {
	    DPRINTF(DMA, "DMA Engine process, request not ready. req=%d\n",
		    req->number());
	    processing = false;
	    return;
	}

	// wait until we can start our next transaction
	if (req->start > curTick) {
	    DPRINTF(DMA, "process start at +%d req=%d\n",
		    req->start - curTick, req->number());
	    process(req->start);
	    processing = false;
	    return;
	}

	assert(req->moved <= req->total_len);

	DPRINTF(DMA, "process, req=%d total=%d remaining=%d\n",
		req->number(), req->total_len, req->total_len - req->moved);

	if (req->moved == req->total_len) {
	    DPRINTF(DMA, "process, done with req=%d\n", req->number());

	    dma_queue.pop_front();
#ifndef NDEBUG
	    req->pending = false;
#endif

	    // This DMA is complete
	    if (req->dmaDoneCB)
		req->dmaDoneCB->process();

	    continue;
	}

	if (req->write && req->bufferCB &&
	    (req->dataoff == req->datalen ||
	     (req->curphys == 0 && req->dataoff == 0))) {
	    DPRINTF(DMA, "write callback\n");
	    req->bufferCB->process();
	    req->dataoff = 0;
	}


	DmaPhys *phys = &req->phys[req->curphys];
	Addr paddr = (Addr)(phys->addr + req->curoff);
	int transfer_len = min(req->datalen - req->dataoff,
			   min(phys->length - req->curoff,
			       req->total_len - req->moved));

	if (transfer_len <= 0)
	    panic("invalid dma length");

	uint8_t *mem_addr = physmem->dma_addr(paddr, transfer_len);

	req->start = ULL(0x7fffffffffffffff);

	if (req->write) {
	    DPRINTF(DMAWriteVerbose, "write paddr=%#x len=%d\n",
		    paddr, transfer_len);
	    DDUMP(DMAWriteVerbose, req->data + req->dataoff, transfer_len);

	    memcpy(mem_addr, req->data + req->dataoff, transfer_len);

	    if (interface) {
		if (req->use_interface)
		    interface->doDMA(WriteInvalidate, paddr, transfer_len,
				     curTick, new DmaTransferEvent(this),
				     NULL);
		else
		    req->start = curTick + transfer_len / 4;
	    } else
		req->start = curTick;
	} else {
	    memcpy(req->data + req->dataoff, mem_addr, transfer_len);

	    if (interface) {
		if (req->use_interface)
		    interface->doDMA(Read, paddr, transfer_len,
				     curTick, new DmaTransferEvent(this),
				     NULL);
		else
		    req->start = curTick + 100 + (transfer_len / 4);
	    } else
		req->start = curTick;

	    DPRINTF(DMAReadVerbose, "read  paddr=%#x len=%d\n",
		    paddr, transfer_len);
	    DDUMP(DMAReadVerbose, req->data + req->dataoff, transfer_len);
	}

	req->curoff += transfer_len;
	req->moved += transfer_len;
	req->dataoff += transfer_len;

	if (!req->write && req->dataoff == req->datalen && req->bufferCB) {
	    DPRINTF(DMA, "read callback\n");
	    req->bufferCB->process();
	    req->dataoff = 0;
	}

	// Go over page boundary. Adjust offset
	if (req->curoff > VMPageSize)
	    DPRINTF(DMA, "request offset(%d) > VMPageSize(%d)\n",
		    req->curoff, VMPageSize);
	if (req->curoff == phys->length) {
	    req->curoff = 0;
	    req->curphys++;

	    if (req->moved < req->total_len &&
		req->nphys == req->curphys)
		panic("Data left, but no physical addrs");
	}
    }

    processing = false;
}

void 
DmaRequest::serialize(ostream &os)
{
#ifdef NDEBUG
    bool pending = true;
#endif
    SERIALIZE_SCALAR(pending);
    SERIALIZE_SCALAR(curphys);
    SERIALIZE_SCALAR(curoff);
    SERIALIZE_SCALAR(dataoff);
    SERIALIZE_SCALAR(moved);
    SERIALIZE_SCALAR(write);
    SERIALIZE_SCALAR(nphys);
    SERIALIZE_SCALAR(total_len);
    SERIALIZE_SCALAR(datalen);
    SERIALIZE_SCALAR(start);
}
    
void 
DmaRequest::unserialize(Checkpoint *cp, const string &section)
{
#ifndef NDEBUG
    UNSERIALIZE_SCALAR(pending);
#endif
    UNSERIALIZE_SCALAR(curphys);
    UNSERIALIZE_SCALAR(curoff);
    UNSERIALIZE_SCALAR(dataoff);
    UNSERIALIZE_SCALAR(moved);
    UNSERIALIZE_SCALAR(write);
    UNSERIALIZE_SCALAR(nphys);
    UNSERIALIZE_SCALAR(total_len);
    UNSERIALIZE_SCALAR(datalen);
    UNSERIALIZE_SCALAR(start);
}

BEGIN_DECLARE_SIM_OBJECT_PARAMS(DmaEngine)

    SimObjectParam<Bus *> bus;
    SimObjectParam<PhysicalMemory *> physmem;
    Param<int> channels;

END_DECLARE_SIM_OBJECT_PARAMS(DmaEngine)

BEGIN_INIT_SIM_OBJECT_PARAMS(DmaEngine)

    INIT_PARAM_DFLT(bus, "bus that we're attached to", NULL),
    INIT_PARAM(physmem, "physical memory"),
    INIT_PARAM_DFLT(channels, "number of dma channels", 36)

END_INIT_SIM_OBJECT_PARAMS(DmaEngine)

CREATE_SIM_OBJECT(DmaEngine)
{
    return new DmaEngine(getInstanceName(), bus, physmem, channels);
}

REGISTER_SIM_OBJECT("DmaEngine", DmaEngine)
