/* $Id: s.dma.hh 1.38 04/02/04 21:41:58-08:00 stever@isabel.reinhardt.house $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* @file
 * Device dma access declaration
 */

#ifndef __DMA_HH__
#define __DMA_HH__

#include "sim/sim_object.hh"
#include "base/trace.hh"
#include "targetarch/isa_traits.hh"	// for Addr

class DmaChannel;

/*
 * Callback event
 */
class DmaCallback
{
  public:
    DmaCallback() {}
    virtual ~DmaCallback() {}
    virtual void process() = 0;
};

/*
 * The DmaRequest structure is initialized by the caller for a set of
 * DMA requests and is subsequently updated by the dma model.
 * (scsi_disk is THE example)
 */
struct DmaPhys {
    Addr addr;
    int length;
};

/*
 * Data representing a DMA Request from a device including status
 * information and relevant parameters
 */
class DmaRequest
{
#if TRACING_ON
  private:
    /** A counter to provide unique IDs to each request. */
    static Counter numbers;
    /** The ID of this request. */
    Counter num;

  public:
    /**
     * Get the unique ID of this request.
     * @return The unique ID.
     */
    Counter number() const 
    { 
	return num; 
    }
#endif

#ifndef NDEBUG
  private:
    /** True if this request has not been completed. */
    bool pending;
#endif

  private:
    /** If true, use the DMAInterface to the memory hierarchy. */
    bool use_interface;

  private:
    friend class DmaEngine;

    /** The current physical address. */
    int curphys;
    /** The current offset. */
    int curoff;
    /** The offset in the current data buffer. */
    int dataoff;
    /** The number of bytes transfered so far. */
    int moved;

  public:
    /** Callback for when the DMA completes. */
    DmaCallback *dmaDoneCB;
    /** Callback when the data buffer is empty. */
    DmaCallback *bufferCB;
    /** True if the transfer is a write. */
    bool write;
    /** The physical address and length to transfer. */
    DmaPhys *phys;
    /** The number of physical adresses to transfer. */
    int nphys;
    /** The total number of bytes to transfer. */
    int total_len;
    /** The current data buffer. */
    uint8_t *data;
    /** The number of bytes in the data buffer. */
    int datalen;
    /** Earliest time this transfer can begin. */
    Tick start;

  public:
    /**
     * Construct and initialize this DMA request
     * @param _use_interface True if we want to use the DMAInterface to mem 
     * hierarchy.
     */
    DmaRequest(bool _use_interface)
	: use_interface(_use_interface)
    {
#ifndef NDEBUG
	pending = false;
#endif
	init(0, 0, false, 0, 0, 0, 0, 0, 0);
    }

    /**
     * Reset this DMA request.
     */
    void init(DmaCallback *_dmaDoneCB, DmaCallback *_bufferCB,
	      bool _write, DmaPhys *_phys, int _nphys, int _total_len,
	      uint8_t *_data, int _datalen, Tick _start)
    {
	assert(!pending);
	curphys = 0;
	curoff = 0;
	dataoff = 0;
	moved = 0;

	dmaDoneCB = _dmaDoneCB;
	bufferCB = _bufferCB;
	write = _write;
	phys = _phys;
	nphys = _nphys;
	total_len = _total_len;
	data = _data;
	datalen = _datalen;
	start = _start;

#if TRACING_ON
	num = numbers++;
#endif
    }

    /**
     * Return the total number of bytes transfered to far.
     * @return The number of bytes moved.
     */
    int totalMoved() const 
    { 
	return moved; 
    }

    const std::string name() const {
        return csprintf("%s_%x", "DmaRequest", (uintptr_t)this);
    }

    /**
     * Serialize this object to the given output stream.
     * @param os The stream to serialize to.
     * @warning Does not serialize the callback pointers, the physical pages, 
     * or the data buffer, must be done by the object containing the request.
     */
    void serialize(std::ostream &os);

    /**
     * Reconstruct the state of this object from a checkpoint.
     * @param cp The checkpoint to use.
     * @param section The section name describing this object.
     * @warning Does not serialize the callback pointers, the physical pages,
     * or the data buffer, must be done by the object containing the request.
     */
    void unserialize(Checkpoint *cp, const std::string &section);
};

class DmaHolder
{
  public:
    virtual DmaRequest *find_dmareq(uint32_t &id) = 0;
};

class Bus;
class PhysicalMemory;
template<class Bus> class DMAInterface;

/*
 * Engine for managing DMA requests queueing them, handling the data
 * transfer from physical memory and handing them off to the memory
 * system if necessary.
 */
class DmaEngine : public SimObject
{
  protected:
    typedef std::deque<DmaRequest *> queue_t;
    queue_t dma_queue;
    const unsigned queue_len;

  protected:
    Bus *bus;
    DMAInterface<Bus> *interface;
    PhysicalMemory *physmem;
    bool processing;

    friend class DmaTransferEvent;
    friend class DmaRequestEvent;
    void process(Tick when);
    void process();

  public:
    DmaEngine(const std::string &name, Bus *bus, PhysicalMemory *pmem,
	      unsigned num_channels);
    ~DmaEngine();

    void doTransfer(DmaRequest *req);

    virtual void serialize(std::ostream &os);
    virtual void unserialize(Checkpoint *cp, const std::string &section);
};
#endif // __DMA_HH__
