/* $Id: s.etherdev.cc 1.93 04/05/21 13:58:39-04:00 binkertn@ziff.eecs.umich.edu $ */

/*
 * Copyright (c) 2003-2004
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* @file
 * Device module for modelling an ethernet controller
 */
#include <cstdio>
#include <deque>
#include <string>

#include "base/inet.hh"
#include "base/stats/events.hh"
#include "cpu/exec_context.hh"
#include "cpu/intr_control.hh"
#include "dev/dma.hh"
#include "dev/etherdev.hh"
#include "dev/etherlink.hh"
#include "dev/tlaser_pcidev.hh"
#include "mem/bus/bus.hh"
#include "mem/bus/dma_interface.hh"
#include "mem/bus/pio_interface.hh"
#include "mem/bus/pio_interface_impl.hh"
#include "mem/functional_mem/memory_control.hh"
#include "mem/functional_mem/physical_memory.hh"
#include "sim/builder.hh"
#include "sim/host.hh"
#include "sim/stats.hh"
#include "targetarch/vtophys.hh"


const char *EsRxStateStrings[] = 
{
    "rxIdle",
    "rxDescRefr",
    "rxDescRead",
    "rxFifoBlock",
    "rxFragWrite",
    "rxDescWrite",
    "rxAdvance"
};

const char *EsTxStateStrings[] =
{
    "txIdle",
    "txDescRefr",
    "txDescRead",
    "txFifoBlock",
    "txFragRead",
    "txDescWrite",
    "txAdvance"
};

const char *EsDmaState[] =
{
    "dmaIdle",
    "dmaReading",
    "dmaWriting",
    "dmaReadWaiting",
    "dmaWriteWaiting"
};

using namespace std;

///////////////////////////////////////////////////////////////////////
//
// EtherDev PCI Device
//
EtherDev::EtherDev(const string &name, IntrControl *i, Tick intr_delay,
		   PhysicalMemory *pmem,
		   Tick tx_delay, Tick rx_delay, bool rx_filter,
		   const int eaddr[6], MemoryController *mmu, Addr a,
		   HierParams *hier, Bus *header_bus, Bus *payload_bus,
		   Tick lat, bool desc_free, bool data_free,
		   Tick dma_read_delay, Tick dma_write_delay,
		   Tick dma_read_factor, Tick dma_write_factor)
    : DmaDevice(name), addr(a),
      rxState(rxIdle), rxEnable(false), rxRing(0), rxRingLen(0),
      rxRingPtr(0), rxDescDone(false), rxDescOff(0), rxDmaState(dmaIdle),
      rxDmaFree(false), rxXferLen(0), rxXferOff(0), rxBufLen(0),
      rxPacketBytes(0),
      txState(txIdle), txEnable(false), txRing(0), txRingLen(0),
      txRingPtr(0), txDescDone(false), txDescOff(0), txDmaState(dmaIdle),
      txDmaFree(false), txXferLen(0), txXferOff(0), txBufLen(0),
      txPacketBytes(0),
      rxDmaReadEvent(this), rxDmaWriteEvent(this),
      txDmaReadEvent(this), txDmaWriteEvent(this),
      dmaDescFree(desc_free), dmaDataFree(data_free),
      rxDelay(rx_delay), txDelay(tx_delay), rxKickTick(0), txKickTick(0),
      txEvent(this), rxFilterEnable(rx_filter),
      physmem(pmem), intctrl(i), intrTick(0),
      cpuPendingIntr(false), intrEvent(0), interface(0), pioLatency(lat)
{
    mmu->add_child(this, Range<Addr>(addr, addr + size));

    if (header_bus) {
	pioInterface = newPioInterface(name, hier, header_bus, this,
				       &EtherDev::cacheAccess);
	pioInterface->addAddrRange(addr, addr + size - 1);
	if (payload_bus)
	    dmaInterface = new DMAInterface<Bus>(name + ".dma", 
						 header_bus, payload_bus, 1);
	else
	    dmaInterface = new DMAInterface<Bus>(name + ".dma", 
						 header_bus, header_bus, 1);
    } else if (payload_bus) {
	pioInterface = newPioInterface(name, hier, payload_bus, this,
				       &EtherDev::cacheAccess);
	pioInterface->addAddrRange(addr, addr + size - 1);
	dmaInterface = new DMAInterface<Bus>(name + ".dma", 
					     payload_bus, payload_bus, 1);
	
    }

    intrDelay = US2Ticks(intr_delay);
    dmaReadDelay = dma_read_delay;
    dmaWriteDelay = dma_write_delay;
    dmaReadFactor = dma_read_factor;
    dmaWriteFactor = dma_write_factor;

    memset(&regs, 0, sizeof(regs));
    regs.rom[0] = eaddr[0];
    regs.rom[1] = eaddr[1];
    regs.rom[2] = eaddr[2];
    regs.rom[3] = eaddr[3];
    regs.rom[4] = eaddr[4];
    regs.rom[5] = eaddr[5];
}

EtherDev::~EtherDev()
{
    if (rxRing)
	delete [] rxRing;

    if (txRing)
	delete [] txRing;
}

void
EtherDev::regStats()
{
    txBytes
	.name(name() + ".txBytes")
	.desc("Bytes Transmitted")
	.prereq(txBytes)
	;

    rxBytes
	.name(name() + ".rxBytes")
	.desc("Bytes Received")
	.prereq(rxBytes)
	;

    txPackets
	.name(name() + ".txPackets")
	.desc("Number of Packets Transmitted")
	.prereq(txBytes)
	;

    rxPackets
	.name(name() + ".rxPackets")
	.desc("Number of Packets Received")
	.prereq(rxBytes)
	;

    txBandwidth
	.name(name() + ".txBandwidth")
	.desc("Transmit Bandwidth (bits/s)")
	.precision(0)
	.prereq(txBytes)
	;

    rxBandwidth
	.name(name() + ".rxBandwidth")
	.desc("Receive Bandwidth (bits/s)")
	.precision(0)
	.prereq(rxBytes)
	;

    txPacketRate
	.name(name() + ".txPPS")
	.desc("Packet Tranmission Rate (packets/s)")
	.precision(0)
	.prereq(txBytes)
	;

    rxPacketRate
	.name(name() + ".rxPPS")
	.desc("Packet Reception Rate (packets/s)")
	.precision(0)
	.prereq(rxBytes)
	;

    txBandwidth = txBytes * Stats::constant(8) / simSeconds;
    rxBandwidth = rxBytes * Stats::constant(8) / simSeconds;
    txPacketRate = txPackets / simSeconds;
    rxPacketRate = rxPackets / simSeconds;
}

Fault
EtherDev::read(MemReqPtr &req, uint8_t *data)
{
    Addr daddr = req->paddr - addr;
    DPRINTF(EthernetPIO, "read  da=%#x pa=%#x va=%#x size=%d\n",
	    daddr, req->paddr, req->vaddr, req->size);

    switch (req->size) {
      case sizeof(uint32_t):
	{
	    uint32_t &reg = *(uint32_t *)data;

	    switch (daddr) {
	      case ES_CMD:
		reg = regs.command;
		break;

	      case ES_CFG:
		reg = regs.config;
		break;

	      case ES_ISR:
		reg = regs.isr;
		devIntrClear(ES_INT_ALL);
		break;

	      case ES_IMR:
		reg = regs.imr;
		break;

	      case ES_RX_RING_LEN:
		reg = regs.rxRingLen;
		break;

	      case ES_TX_RING_LEN:
		reg = regs.txRingLen;
		break;

	      case ES_ROM_ADDR:
		reg = regs.romAddr;
		break;

	      case ES_ROM_DATA:
		if (regs.romAddr >= ES_ROM_SIZE)
		    panic("Can't read past end of rom");

		reg = regs.rom[regs.romAddr++];
		break;

	      case ES_RXFILT_CTL:
		reg = regs.rxFiltCtrl;
		break;

	      case ES_RXFILT_DATA:
		reg = regs.rxFiltData;
		break;

	      default:
		panic("accessing register with invalid size: addr=%#x", daddr);
	    }
	    DPRINTF(EthernetPIO, "read  data=%d data=%#x\n", reg, reg);
	}
	break;

      case sizeof(uint64_t):
	{
	    uint64_t &reg = *(uint64_t *)data;

	    switch (daddr) {
	      case ES_RX_RING:
		reg = regs.rxRing;
		break; 
	      case ES_TX_RING:
		reg = regs.txRing;
		break;

	      default:
		panic("accessing register with invalid size: addr=%#x", daddr);
	    }
	    DPRINTF(EthernetPIO, "read  data=%d data=%#x\n", reg, reg);
	}
	break;

      default:
	panic("accessing register with invalid size: addr=%#x, size=%d",
	      daddr, req->size);
    }


    return No_Fault;
}

Fault
EtherDev::write(MemReqPtr &req, const uint8_t *data)
{
    Addr daddr = req->paddr - addr;
    DPRINTF(EthernetPIO, "write da=%#x pa=%#x va=%#x size=%d\n",
	    daddr, req->paddr, req->vaddr, req->size);

    Addr addr;

    if (req->size == sizeof(uint32_t)) {
	uint32_t reg = *(uint32_t *)data;
	DPRINTF(EthernetPIO, "write data=%d data=%#x\n", reg, reg);

	switch (daddr) {
	  case ES_CMD:
	    {
		if (reg & ES_CMD_RES)
		    panic("writing reserved command bit");

		if ((reg & (ES_CMD_RX_ENABLE | ES_CMD_RX_DISABLE)) ==
		    (ES_CMD_RX_ENABLE | ES_CMD_RX_DISABLE))
		    panic("Cannot enable and disable recv at the same time.");

		if ((reg & (ES_CMD_TX_ENABLE | ES_CMD_TX_DISABLE)) ==
		    (ES_CMD_TX_ENABLE | ES_CMD_TX_DISABLE))
		    panic("Cannot enable and disable xmit at the same time.");

		bool old_rxEnable = rxEnable;
		bool old_txEnable = txEnable;
		if (reg & ES_CMD_TX_ENABLE)
		    txEnable = true;

		if (reg & ES_CMD_TX_DISABLE)
		    txEnable = false;

		if (reg & ES_CMD_RX_ENABLE)
		    rxEnable = true;

		if (reg & ES_CMD_RX_DISABLE)
		    rxEnable = false;

		bool rx_reset = reg & ES_CMD_RX_RESET;
		bool tx_reset = reg & ES_CMD_TX_RESET;
		bool chip_reset = reg & ES_CMD_CHIP_RESET;
		bool start = reg & ES_CMD_START;

		if (start)
		    DPRINTF(Ethernet, "start bit set!\n");

		if (rx_reset)
		    rxReset();

		if (tx_reset)
		    txReset();

		if (chip_reset)
		    chipReset();

		if (rxEnable && (rx_reset || !old_rxEnable))
		    rxKick();

		if (txEnable && (start || tx_reset || !old_txEnable))
		    txKick();
	    }
	    break;

	  case ES_CFG:
	    {
		if (reg & ES_CFG_RES)
		    panic("writing reserved chip configuration bit");

		// Figure out which things actually changed
		uint32_t xor_reg = reg ^ regs.config;

		if (xor_reg & ES_CFG_SPEED_MASK) {
		    uint32_t speed = reg & ES_CFG_SPEED_MASK;
		    regs.config &= ~ES_CFG_SPEED_MASK;
		    regs.config |= speed;

		    setSpeed(speed);
		}

		if (xor_reg & ES_CFG_DUPLEX) {
		    uint32_t duplex = reg & ES_CFG_DUPLEX;
		    regs.config &= ~ES_CFG_DUPLEX;
		    regs.config |= duplex;

		    setDuplex((bool)duplex);
		}

		if (xor_reg & ES_CFG_LINK) {
		    uint32_t link = reg & ES_CFG_LINK;
		    regs.config &= ~ES_CFG_LINK;
		    regs.config |= link;

		    setLink((bool)link);
		}
	    }
	    break;

	  case ES_ISR:
	    if (reg & ES_INT_RES)
		panic("writing reserved interrupt status bit");

	    if (reg & ES_INT_RX_OK)
		regs.isr &= ~ES_INT_RX_OK;

	    if (reg & ES_INT_RX_ERR)
		regs.isr &= ~ES_INT_RX_ERR;

	    if (reg & ES_INT_RX_OFLOW)
		regs.isr &= ~ES_INT_RX_OFLOW;

	    if (reg & ES_INT_TX_OK)
		regs.isr &= ~ES_INT_TX_OK;

	    if (reg & ES_INT_TX_ERR)
		regs.isr &= ~ES_INT_TX_ERR;

	    break;

	  case ES_IMR:
	    if (reg & ES_INT_RES)
		panic("writing reserved interrupt mask bit");

	    devIntrChangeMask(reg);
	    break;

	  case ES_RX_RING_LEN:
	    regs.rxRingLen = reg;
	    break;

	  case ES_TX_RING_LEN:
	    regs.txRingLen = reg;
	    break;

	  case ES_ROM_ADDR:
	    regs.romAddr = reg;
	    break;

	  case ES_ROM_DATA:
	    panic("Cannot write to rom");
	    break;

	  case ES_RXFILT_CTL:
	    regs.rxFiltCtrl = reg;
	    break;

	  case ES_RXFILT_DATA:
	    addr = (regs.rxFiltCtrl & ES_RXFILTCTL_ADDR);
	    if (addr < ES_FILTADDR_MCAST_LO) {
		switch (addr) {
		  case ES_RXFILTADDR_PAR0:
		    ((uint16_t *)regs.perfect)[0] = reg & 0xffff;
		    break;

		  case ES_RXFILTADDR_PAR1:
		    ((uint16_t *)regs.perfect)[1] = reg & 0xffff;
		    break;

		  case ES_RXFILTADDR_PAR2:
		    ((uint16_t *)regs.perfect)[2] = reg & 0xffff;
		    break;

		  default:
		    panic("Invalid Receive filter address");
		}
	    } else if (addr <= ES_FILTADDR_MCAST_HI) {
		int offset = (addr - ES_FILTADDR_MCAST_LO) / sizeof(uint16_t);
		regs.hash[offset] = reg & 0xffff;
	    } else
		panic("Invalid Receive filter address");
	    break;

	  default:
	    panic("accessing register with invalid size: addr=%#x", daddr);
	}
    } else if (req->size == sizeof(uint64_t)) {
	uint32_t reg = *(uint64_t *)data;
	DPRINTF(EthernetPIO, "EtherDev write data=%d data=%#x\n", reg, reg);

	switch (daddr) {
	  case ES_RX_RING:
	    regs.rxRing = reg;
	    break;

	  case ES_TX_RING:
	    regs.txRing = reg;
	    break;

	  default:
	    panic("accessing register with invalid size: addr=%#x", daddr);
	}
    } else
	panic("Invalid Request Size");

    return No_Fault;
}

void
EtherDev::devIntrPost(uint32_t interrupts)
{
    DPRINTF(Ethernet, "interrupt posted intr=%x isr=%x imr=%x\n",
	    interrupts, regs.isr, regs.imr);

    if (interrupts & ES_INT_RES)
	panic("Cannot set a reserved interrupt");

    bool delay = true;
    if (interrupts & ES_INT_TX_OK) {
	regs.isr |= ES_INT_TX_OK;
    }

    if (interrupts & ES_INT_TX_ERR) {
	regs.isr |= ES_INT_TX_ERR;
	delay = false;
    }

    if (interrupts & ES_INT_RX_OK) {
	regs.isr |= ES_INT_RX_OK;
    }

    if (interrupts & ES_INT_RX_ERR) {
	regs.isr |= ES_INT_RX_ERR;
	delay = false;
    }

    if (interrupts & ES_INT_RX_OFLOW) {
	regs.isr |= ES_INT_RX_OFLOW;
	delay = false;
    }

    if ((regs.isr & regs.imr)) {
	Tick when = curTick;
	if (delay)
	    when += intrDelay;
	cpuIntrPost(when);
    }
}

void
EtherDev::devIntrClear(uint32_t interrupts)
{
    DPRINTF(Ethernet, "interrupt cleared intr=%x isr=%x imr=%x\n",
	    interrupts, regs.isr, regs.imr);

    if (interrupts & ES_INT_RES)
	panic("Cannot clear a reserved interrupt");

    if (interrupts & ES_INT_TX_OK)
	regs.isr &= ~ES_INT_TX_OK;

    if (interrupts & ES_INT_TX_ERR)
	regs.isr &= ~ES_INT_TX_ERR;

    if (interrupts & ES_INT_RX_OK)
	regs.isr &= ~ES_INT_RX_OK;

    if (interrupts & ES_INT_RX_ERR)
	regs.isr &= ~ES_INT_RX_ERR;

    if (interrupts & ES_INT_RX_OFLOW)
	regs.isr &= ~ES_INT_RX_OFLOW;

    if (!(regs.isr & regs.imr))
	cpuIntrClear();
}

void
EtherDev::devIntrChangeMask(uint32_t mask)
{
    DPRINTF(Ethernet, "iterrupt mask changed\n");
    regs.imr = mask;

    if (regs.isr & regs.imr)
	cpuIntrPost(curTick);
    else
	cpuIntrClear();
}

void
EtherDev::cpuIntrPost(Tick when)
{
    if (when > intrTick && intrTick != 0)
	return;

    intrTick = when;

    if (intrEvent) {
	intrEvent->squash();
	intrEvent = 0;
    }

    if (when < curTick) {
	cpuInterrupt();
    } else {
	intrEvent = new IntrEvent(this, true);
	intrEvent->schedule(intrTick);
    }
}

void
EtherDev::cpuInterrupt()
{
    // Don't send an interrupt if there's already one
    if (cpuPendingIntr)
	return;

    // Don't send an interrupt if it's supposed to be delayed
    if (intrTick > curTick)
	return;

    // Whether or not there's a pending interrupt, we don't care about
    // it anymore
    intrEvent = 0;
    intrTick = 0;

    // Send interrupt
    cpuPendingIntr = true;
    intctrl->post(TheISA::INTLEVEL_IRQ1, TheISA::INTINDEX_ETHERNET);
}

void
EtherDev::cpuIntrClear()
{
    intrTick = 0;
    if (intrEvent) {
	intrEvent->squash();
	intrEvent = 0;
    }
	
    if (cpuPendingIntr) {
	cpuPendingIntr = false;
	intctrl->clear(TheISA::INTLEVEL_IRQ1, TheISA::INTINDEX_ETHERNET);
    }
}

bool
EtherDev::cpuIntrPending() const
{ return cpuPendingIntr; }

void
EtherDev::rxReset()
{
    DPRINTF(Ethernet, "receive reset\n");

    rxState = rxIdle;
    rxRingLen = regs.rxRingLen;
    if (rxRing)
	delete [] rxRing;
    rxRing = new es_desc[rxRingLen];
    memset(rxRing, 0, sizeof(es_desc) * rxRingLen);
    rxRingPtr = 0;
    rxDescDone = false;
}

void
EtherDev::txReset()
{
    DPRINTF(Ethernet, "transmit reset\n");

    txState = txIdle;
    txRingLen = regs.txRingLen;
    if (txRing)
	delete [] txRing;
    txRing = new es_desc[txRingLen];
    memset(txRing, 0, sizeof(es_desc) * txRingLen);
    txRingPtr = 0;
    txDescDone = 0;
}

void
EtherDev::chipReset()
{
    DPRINTF(Ethernet, "chip reset\n");
}

void
EtherDev::rxDmaReadCopy()
{
    assert(rxDmaState == dmaReading);

    memcpy(rxDmaData, physmem->dma_addr(rxDmaAddr, rxDmaLen), rxDmaLen);
    rxDmaState = dmaIdle;

    DPRINTF(EthernetDMA, "rx dma read  paddr=%#x len=%d\n",
	    rxDmaAddr, rxDmaLen);
    DDUMP(EthernetDMA, rxDmaData, rxDmaLen);
}

bool
EtherDev::doRxDmaRead()
{
    assert(rxDmaState == dmaIdle || rxDmaState == dmaReadWaiting);
    rxDmaState = dmaReading;

    if (dmaInterface && !rxDmaFree) {
	if (dmaInterface->busy())
	    rxDmaState = dmaReadWaiting;
	else
	    dmaInterface->doDMA(Read, rxDmaAddr, rxDmaLen, curTick,
				&rxDmaReadEvent);
	return true;
    }

    if (dmaReadDelay == 0 && dmaReadFactor == 0) {
	rxDmaReadCopy();
	return false;
    }

    Tick factor = ((rxDmaLen + ULL(63)) >> ULL(6)) * dmaReadFactor;
    Tick start = curTick + dmaReadDelay + factor;
    rxDmaReadEvent.schedule(start);
    return true;
}

void
EtherDev::rxDmaReadDone()
{
    assert(rxDmaState == dmaReading);
    rxDmaReadCopy();

    // If the transmit state machine has a pending DMA, let it go first
    if (txDmaState == dmaReadWaiting || txDmaState == dmaWriteWaiting)
	txKick();

    rxKick();
}

void
EtherDev::rxDmaWriteCopy()
{
    assert(rxDmaState == dmaWriting);

    memcpy(physmem->dma_addr(rxDmaAddr, rxDmaLen), rxDmaData, rxDmaLen);
    rxDmaState = dmaIdle;

    DPRINTF(EthernetDMA, "rx dma write paddr=%#x len=%d\n",
	    rxDmaAddr, rxDmaLen);
    DDUMP(EthernetDMA, rxDmaData, rxDmaLen);
}

bool
EtherDev::doRxDmaWrite()
{
    assert(rxDmaState == dmaIdle || rxDmaState == dmaWriteWaiting);
    rxDmaState = dmaWriting;

    if (dmaInterface && !rxDmaFree) {
	if (dmaInterface->busy())
	    rxDmaState = dmaWriteWaiting;
	else
	    dmaInterface->doDMA(WriteInvalidate, rxDmaAddr, rxDmaLen, curTick,
				&rxDmaWriteEvent);
	return true;
    }

    if (dmaWriteDelay == 0 && dmaWriteFactor == 0) {
	rxDmaWriteCopy();
	return false;
    }

    Tick factor = ((rxDmaLen + ULL(63)) >> ULL(6)) * dmaWriteFactor;
    Tick start = curTick + dmaWriteDelay + factor;
    rxDmaWriteEvent.schedule(start);
    return true;
}

void
EtherDev::rxDmaWriteDone()
{
    assert(rxDmaState == dmaWriting);
    rxDmaWriteCopy();

    // If the transmit state machine has a pending DMA, let it go first
    if (txDmaState == dmaReadWaiting || txDmaState == dmaWriteWaiting)
	txKick();

    rxKick();
}

void
EtherDev::rxKick()
{
    DPRINTF(Ethernet, "receive kick state=%s (rxBuf.size=%d)\n",
	    EsRxStateStrings[rxState], rxBuf.size());

    if (rxKickTick > curTick) {
	DPRINTF(Ethernet, "receive kick exiting, can't run till %d\n",
		rxKickTick);

	return;
    }

    es_desc *desc_data = &rxRing[rxRingPtr];
    Addr desc_addr = regs.rxRing + rxRingPtr * sizeof(es_desc);

  next:
    switch(rxDmaState) {
      case dmaReadWaiting:
	if (doRxDmaRead())
	    goto exit;
	break;
      case dmaWriteWaiting:
	if (doRxDmaWrite())
	    goto exit;
	break;
      default:
	break;
    }

    switch (rxState) {
      case rxIdle:
	if (!rxEnable) {
	    DPRINTF(Ethernet, "Receive disabled.  Nothing to do.\n");
	    goto exit;
	}

	if (rxDescDone) {
	    rxState = rxDescRefr;

	    /**
	     * @todo Fix this to only DMA read the link address.
	     * We don't use the linked list field, so we don't need
	     * to referesh the link field, and as a result just skip
	     * to the next state.
	     */
	    panic("We should never transition to this state!\n");
#if 0
	    rxDmaAddr = desc_data->link & 0x3fffffff;
	    rxDmaData = desc_addr + offsetof(es_desc, link);
	    rxDmaLen = sizeof(desc_data->link);
	    rxDmaFree = dmaDescFree;

	    if (doRxDmaRead())
		goto exit;
#endif
	} else {
	    rxState = rxDescRead;

	    rxDmaAddr = desc_addr & 0x3fffffff;
	    rxDmaData = desc_data;
	    rxDmaLen = sizeof(es_desc);
	    rxDmaFree = dmaDescFree;

	    if (doRxDmaRead())
		goto exit;
	}
	break;

      case rxDescRefr:
	if (rxDmaState != dmaIdle)
	    goto exit;

	rxState = rxAdvance;
	break;

      case rxDescRead:
	if (rxDmaState != dmaIdle)
	    goto exit;

	if (desc_data->flags & ES_DESC_OWN) {
	    rxState = rxIdle;
	    /**
	     * @todo set the receive status reg idle bit
	     */
	} else {
	    rxDescDone = true;
	    rxState = rxFifoBlock;
	    rxDescOff = 0;
	}
	break;

      case rxFifoBlock:
	if (!rxPacket) {
	    /**
	     * @todo in reality, we should be able to start processing
	     * the packet as it arrives, and not have to wait for the
	     * full packet ot be in the receive fifo.
	     */
	    if (rxBuf.empty())
		goto exit;

	    // If we don't have a packet, grab a new one from the fifo.
	    rxPacket = rxBuf.front();
	    rxPacketBytes = rxPacket->length;
	    rxXferOff = 0;

	    // Must clear the value before popping to decrement the
	    // reference count
	    rxBuf.front() = NULL;
	    rxBuf.pop_front();
	}

	/**
	 * @todo there is support here for partial writes of
	 * descriptors, but it is not currently used because the
	 * receive FIFO doesn't hold partial packets just yet.
	 */
	if (rxPacketBytes > 0 && rxDescOff < desc_data->length) {
	    // Now that we have a packet, the next thing to do is copy
	    // the data to the the descriptor
	    rxState = rxFragWrite;

	    rxXferLen = min<int>(rxPacketBytes, desc_data->length - rxDescOff);

	    rxDmaAddr = (desc_data->addr + rxDescOff) & 0x3fffffff;
	    rxDmaData = rxPacket->data + rxXferOff;
	    rxDmaLen = rxXferLen;
	    rxDmaFree = dmaDataFree;

	    if (doRxDmaWrite())
		goto exit;
	} else {
	    rxState = rxDescWrite;

	    desc_data->flags |= ES_DESC_OWN;

	    if (rxPacketBytes == 0) {
		rxBufLen -= rxPacket->length;
		rxPacket = 0;
		desc_data->flags &= ~ES_DESC_MORE;
	    } else {
		desc_data->flags |= ES_DESC_MORE;
	    }

	    desc_data->length = rxDescOff;

	    /**
	     * @todo in the real NS part, we only need to update the
	     * cmdsts register.
	     */
	    rxDmaAddr = desc_addr & 0x3fffffff;
	    rxDmaData = desc_data;
	    rxDmaLen = sizeof(es_desc);
	    rxDmaFree = dmaDescFree;

	    if (doRxDmaWrite())
		goto exit;
	}
	break;

      case rxFragWrite:
	if (rxDmaState != dmaIdle)
	    goto exit;

	// Now that the data has been transferred, update counters.
	rxXferOff += rxXferLen;
	rxDescOff += rxXferLen;
	rxPacketBytes -= rxXferLen;

	rxState = rxFifoBlock;
	break;

      case rxDescWrite:
	if (rxDmaState != dmaIdle)
	    goto exit;

	if (rxPacket == 0)
	    devIntrPost(ES_INT_RX_OK);

	rxState = rxAdvance;
	break;

      case rxAdvance:
	/**
	 * @todo in the real NS chip, there's a link list, and you
	 * should check the link field.
	 */
	if (true /* link */) {
	    rxState = rxDescRead;

	    rxDescDone = false;
	    if (++rxRingPtr >= rxRingLen)
		rxRingPtr = 0;

	    desc_data = &rxRing[rxRingPtr];
	    desc_addr = regs.rxRing + rxRingPtr * sizeof(es_desc);

	    rxDmaAddr = desc_addr & 0x3fffffff;
	    rxDmaData = desc_data;
	    rxDmaLen = sizeof(es_desc);
	    rxDmaFree = dmaDescFree;

	    if (doRxDmaRead())
		goto exit;
	} else {
	    rxDescDone = true;
	    rxState = rxIdle;
	}

	break;

      default:
	panic("invalid state");
    }

    DPRINTF(Ethernet, "entering next rx state=%s\n",
	    EsRxStateStrings[rxState]);
    goto next;

  exit:
    /**
     * @todo do we want to schedule a future kick?
     */
    DPRINTF(Ethernet, "rx state machine exited state=%s\n",
	    EsRxStateStrings[rxState]);
}

void
EtherDev::transmit()
{
    if (txBuf.empty())
	return;

    if (interface->sendPacket(txBuf.front())) {
	Stats::recordEvent("Packet Transmitted");
	DPRINTF(Ethernet, "transmit packet\n");
	DDUMP(Ethernet, txBuf.front()->data, txBuf.front()->length);
	txBytes += txBuf.front()->length;
	txPackets++;

	txBufLen -= txBuf.front()->length;
	txBuf.front() = NULL;
	txBuf.pop_front();
	devIntrPost(ES_INT_TX_OK);
    }

    if (!txBuf.empty() && !txEvent.scheduled()) {
	DPRINTF(Ethernet, "reschedule transmit\n");
	txEvent.schedule(curTick + 1000);
    }
}

void
EtherDev::txDmaReadCopy()
{
    assert(txDmaState == dmaReading);

    memcpy(txDmaData, physmem->dma_addr(txDmaAddr, txDmaLen), txDmaLen);
    txDmaState = dmaIdle;

    DPRINTF(EthernetDMA, "tx dma read  paddr=%#x len=%d\n",
	    txDmaAddr, txDmaLen);
    DDUMP(EthernetDMA, txDmaData, txDmaLen);
}

bool
EtherDev::doTxDmaRead()
{
    assert(txDmaState == dmaIdle || txDmaState == dmaReadWaiting);
    txDmaState = dmaReading;

    if (dmaInterface && !txDmaFree) {
	if (dmaInterface->busy())
	    txDmaState = dmaReadWaiting;
	else
	    dmaInterface->doDMA(Read, txDmaAddr, txDmaLen, curTick,
				&txDmaReadEvent);
	return true;
    }

    if (dmaReadDelay == 0 && dmaReadFactor == 0.0) {
	txDmaReadCopy();
	return false;
    }

    Tick factor = ((txDmaLen + ULL(63)) >> ULL(6)) * dmaReadFactor;
    Tick start = curTick + dmaReadDelay + factor;
    txDmaReadEvent.schedule(start);
    return true;
}

void
EtherDev::txDmaReadDone()
{
    assert(txDmaState == dmaReading);
    txDmaReadCopy();

    // If the receive state machine  has a pending DMA, let it go first
    if (rxDmaState == dmaReadWaiting || rxDmaState == dmaWriteWaiting)
	rxKick();

    txKick();
}

void
EtherDev::txDmaWriteCopy()
{
    assert(txDmaState == dmaWriting);

    memcpy(physmem->dma_addr(txDmaAddr, txDmaLen), txDmaData, txDmaLen);
    txDmaState = dmaIdle;

    DPRINTF(EthernetDMA, "tx dma write paddr=%#x len=%d\n",
	    txDmaAddr, txDmaLen);
    DDUMP(EthernetDMA, txDmaData, txDmaLen);
}

bool
EtherDev::doTxDmaWrite()
{
    assert(txDmaState == dmaIdle || txDmaState == dmaWriteWaiting);
    txDmaState = dmaWriting;

    if (dmaInterface && !txDmaFree) {
	if (dmaInterface->busy())
	    txDmaState = dmaWriteWaiting;
	else
	    dmaInterface->doDMA(WriteInvalidate, txDmaAddr, txDmaLen, curTick,
				&txDmaWriteEvent);
	return true;
    }

    if (dmaWriteDelay == 0 && dmaWriteFactor == 0.0) {
	txDmaWriteCopy();
	return false;
    }

    Tick factor = ((txDmaLen + ULL(63)) >> ULL(6)) * dmaWriteFactor;
    Tick start = curTick + dmaWriteDelay + factor;
    txDmaWriteEvent.schedule(start);
    return true;
}

void
EtherDev::txDmaWriteDone()
{
    assert(txDmaState == dmaWriting);
    txDmaWriteCopy();

    // If the receive state machine  has a pending DMA, let it go first
    if (rxDmaState == dmaReadWaiting || rxDmaState == dmaWriteWaiting)
	rxKick();

    txKick();
}

void
EtherDev::txKick()
{
    DPRINTF(Ethernet, "transmit kick state=%s\n", EsTxStateStrings[txState]);

    transmit();

    if (rxKickTick > curTick) {
	DPRINTF(Ethernet, "receive kick exiting, can't run till %d\n",
		rxKickTick);

	return;
    }

    es_desc *desc_data = &txRing[txRingPtr];
    Addr desc_addr = regs.txRing + txRingPtr * sizeof(es_desc);

  next:
    switch(txDmaState) {
      case dmaReadWaiting:
	if (doTxDmaRead())
	    goto exit;
	break;
      case dmaWriteWaiting:
	if (doTxDmaWrite())
	    goto exit;
	break;
      default:
	break;
    }

    switch (txState) {
      case txIdle:
	if (!txEnable) {
	    DPRINTF(Ethernet, "Transmit disabled.  Nothing to do.\n");
	    goto exit;
	}

	if (txDescDone) {
	    txState = txDescRefr;

	    /**
	     * @todo Fix this to only DMA read the link address.
	     * We don't use the linked list field, so we don't need
	     * to referesh the link field, and as a result just skip
	     * to the next state.
	     */
	    panic("We should never transition to this state!\n");
#if 0
	    txDmaAddr = (desc_addr + offsetof(es_desc, link)) & 0x3fffffff;
	    txDmaData = desc_data->link;
	    txDmaLen = sizeof(desc_data->link);
	    txDmaFree = dmaDescFree;

	    if (doTxDmaRead())
		goto exit;
#endif
	} else {
	    txState = txDescRead;

	    txDmaAddr = desc_addr & 0x3fffffff;
	    txDmaData = desc_data;
	    txDmaLen = sizeof(es_desc);
	    txDmaFree = dmaDescFree;

	    if (doTxDmaRead())
		goto exit;
	}
	break;

      case txDescRefr:
	if (txDmaState != dmaIdle)
	    goto exit;

	txState = txAdvance;
	break;

      case txDescRead:
	if (txDmaState != dmaIdle)
	    goto exit;

	if (desc_data->flags & ES_DESC_OWN) {
	    txDescDone = true;
	    txState = txFifoBlock;
	    txDescOff = 0;
	} else {
	    txState = txIdle;
	    /**
	     * @todo set the transmit status reg idle bit
	     */

	    goto exit;
	}
	break;

      case txFifoBlock:
	// If there's no room in the transmit buffer, there's nothing
	// to do.
	if (txBufLen >= txBufMaxLen) {
	    DPRINTF(Ethernet, "Transmit buffer is full\n");
	    goto exit;
	}

	if (!txPacket) {
	    txPacket = new EtherPacket;
	    txPacket->data = new uint8_t[16384];
	    txXferOff = 0;
	}

	if (txDescOff < desc_data->length) {
	    txState = txFragRead;

	    // The number of bytes transfered is the lower of the
	    // number of bytes free in the transmit FIFO and the
	    // number of bytes left to transfer in this descriptor
	    txXferLen = min<int>(desc_data->length - txDescOff,
				 txBufMaxLen - txBufLen);

	    txDmaAddr = (desc_data->addr + txDescOff) & 0x3fffffff;
	    txDmaData = txPacket->data + txXferOff;
	    txDmaLen = txXferLen;
	    txDmaFree = dmaDataFree;

	    if (doTxDmaRead())
		goto exit;
	} else {
	    if (desc_data->flags & ES_DESC_MORE) {
		desc_data->flags &= ~ES_DESC_OWN;
		txState = txDescWrite;
	    } else {
		/**
		 * @todo in reality, we should go to the txAdvance
		 * state here an queue the descriptor until it has
		 * completed transmission.  Once the transmission is
		 * complete, we should then clear the OWN bit and
		 * raise the interrupt, but we'll do it here instead.
		 */
#if 0
		txPacket->priv = txRingPtr;
		txState = txAdvance;
		// set some cmdsts, extsts flags
#endif

		txPacket->length = txXferOff;
		txBuf.push_back(txPacket);
		txPacket = 0;

		/**
		 * @todo we should really only do this for descriptors
		 * with the MORE bit set.
		 */
		desc_data->flags &= ~ES_DESC_OWN;

		txState = txDescWrite;
	    }

	    /**
	     * @todo in the real NS part, we only need to update the
	     * cmdsts register.
	     */
	    txDmaAddr = desc_addr & 0x3fffffff;
	    txDmaData = desc_data;
	    txDmaLen = sizeof(es_desc);
	    txDmaFree = dmaDescFree;

	    if (doTxDmaWrite())
		goto exit;
	}
	break;

      case txFragRead:
	if (txDmaState != dmaIdle)
	    goto exit;

	txXferOff += txXferLen;
	txDescOff += txXferLen;
	txBufLen += txXferLen;

	txState = txFifoBlock;
	break;

      case txDescWrite:
	if (txDmaState != dmaIdle)
	    goto exit;

	transmit();

	txState = txAdvance;
	break;

      case txAdvance:
	/**
	 * @todo in the real NS chip, there's a link list, and you
	 * should check the link field.
	 */
	if (true /* link */) {
	    txState = txDescRead;

	    txDescDone = false;
	    if (++txRingPtr >= txRingLen)
		txRingPtr = 0;

	    desc_data = &txRing[txRingPtr];
	    desc_addr = regs.txRing + txRingPtr * sizeof(es_desc);


	    txDmaAddr = desc_addr & 0x3fffffff;
	    txDmaData = desc_data;
	    txDmaLen = sizeof(es_desc);
	    txDmaFree = dmaDescFree;

	    if (doTxDmaRead())
		goto exit;
	} else {
	    txDescDone = true;
	    txState = txIdle;
	}
	break;

      default:
	panic("invalid state");
    }

    DPRINTF(Ethernet, "entering next tx state=%s\n",
	    EsTxStateStrings[txState]);
    goto next;

  exit:
    /**
     * @todo do we want to schedule a future kick?
     */
    DPRINTF(Ethernet, "tx state machine exited state=%s\n",
	    EsTxStateStrings[txState]);
}

void
EtherDev::txDump() const
{
    int i = txRingPtr;
    for (int loop = 0; loop < txRingLen; loop++) {
	es_desc *desc = &txRing[i];

	if (desc->addr)
	    cprintf("desc[%d]: addr=%#x, len=%d, flags=%#x\n",
		    i, desc->addr, desc->length, desc->flags);

	if (++i >= txRingLen)
	    i = 0;
    }
}

void
EtherDev::rxDump() const
{
    int i = rxRingPtr;
    for (int loop = 0; loop < rxRingLen; loop++) {
	es_desc *desc = &rxRing[i];

	if (desc->addr)
	    cprintf("desc[%d]: addr=%#x, len=%d, flags=%#x\n",
		    i, desc->addr, desc->length, desc->flags);

	if (++i >= rxRingLen)
	    i = 0;
    }
}

bool
EtherDev::rxHashMatch(PacketPtr &packet)
{
    uint32_t crc = crc32be(packet->data, 6);
    int index = (crc >> 25) & 0x7f;
    int bit = (crc >> 21) & 0xf;
    bool match = ((regs.hash[index] >> bit) & 0x01);

    DPRINTF(Ethernet, "etherdev hash %s for %s\n",
	    match ? "match" : "no-match", eaddr_string(packet->data));

    return match;
}

bool
EtherDev::rxFilter(PacketPtr &packet)
{
    bool drop = true;
    string type;

    if (packet->IsUnicast()) {
	type = "unicast";

	// If we're accepting all unicast addresses
	if (regs.rxFiltCtrl & ES_RXFILTCTL_ALLUNI)
	    drop = false;

	// If we make a perfect match
	if ((regs.rxFiltCtrl & ES_RXFILTCTL_PERFECT) &&
	    (memcmp(regs.perfect, packet->data, sizeof(regs.perfect)) == 0))
	    drop = false;

	if ((regs.rxFiltCtrl & ES_RXFILTCTL_UCHASH) && rxHashMatch(packet))
	    drop = false;

    } else if (packet->IsBroadcast()) {
	type = "broadcast";

	// if we're accepting broadcasts
	if (regs.rxFiltCtrl & ES_RXFILTCTL_BCAST)
	    drop = false;

    } else if (packet->IsMulticast()) {
	type = "multicast";

	// if we're accepting all multicasts
	if (regs.rxFiltCtrl & ES_RXFILTCTL_ALLMULTI)
	    drop = false;

	if ((regs.rxFiltCtrl & ES_RXFILTCTL_MCHASH) && rxHashMatch(packet))
	    drop = false;
    } else {
	type = "unknown";

	// oh well, punt on this one
    }

    if (drop) {
	DPRINTF(Ethernet, "rxFilter drop\n");
	DDUMP(EthernetData, packet->data, packet->length);
    }

    return drop;
}

bool
EtherDev::recvPacket(PacketPtr &packet)
{
    Stats::recordEvent("Packet Received");
    rxBytes += packet->length;
    rxPackets++;

    if (!rxEnable) {
	DPRINTF(Ethernet, "receive disabled...packet dropped\n");
	interface->recvDone();
	return true;
    }

    if (rxFilterEnable && rxFilter(packet)) {
	DPRINTF(Ethernet, "packet filtered...dropped\n");
	interface->recvDone();
	return true;
    }

    if (rxBufLen + packet->length >= rxBufMaxLen) {
	DPRINTF(Ethernet,
		"packet will not fit in receive buffer...packet dropped\n");
	devIntrPost(ES_INT_RX_OFLOW);
	return false;
    }

    rxBuf.push_back(packet);
    rxBufLen += packet->length;

    interface->recvDone();

    rxKick();
    return true;
}

void
EtherDev::transferDone()
{
    if (txBuf.empty())
	return;

    DPRINTF(Ethernet, "schedule transmit\n");

    if (txEvent.scheduled())
	txEvent.reschedule(curTick + 1);
    else
	txEvent.schedule(curTick + 1);
}

void
EtherDev::setSpeed(int speed)
{}

void
EtherDev::setLink(bool link_up)
{}

void
EtherDev::setDuplex(bool fdx)
{}

void
EtherDev::serialize(ostream &os)
{
    /*
     * Finalize any DMA events now.
     */
    if (rxDmaReadEvent.scheduled())
	rxDmaReadCopy();
    if (rxDmaWriteEvent.scheduled())
	rxDmaWriteCopy();
    if (txDmaReadEvent.scheduled())
	txDmaReadCopy();
    if (txDmaWriteEvent.scheduled())
	txDmaWriteCopy();

    /*
     * Serialize the device registers
     */
    SERIALIZE_SCALAR(regs.command);
    SERIALIZE_SCALAR(regs.config);
    SERIALIZE_SCALAR(regs.isr);
    SERIALIZE_SCALAR(regs.imr);
    SERIALIZE_SCALAR(regs.txRing);          
    SERIALIZE_SCALAR(regs.rxRing);          
    SERIALIZE_SCALAR(regs.txRingLen);
    SERIALIZE_SCALAR(regs.rxRingLen);
    SERIALIZE_SCALAR(regs.romAddr);
    SERIALIZE_SCALAR(regs.romData);
    SERIALIZE_SCALAR(regs.rxFiltCtrl);
    SERIALIZE_SCALAR(regs.rxFiltData);

    SERIALIZE_ARRAY(regs.perfect, EADDR_LEN);
    SERIALIZE_ARRAY(regs.hash, ES_HASH_SIZE);

    /*
     * Serialize the various parts of the receive state machine
     */
    int rxState = this->rxState;
    SERIALIZE_SCALAR(rxState);
    SERIALIZE_SCALAR(rxEnable);
    for (int i = 0; i < rxRingLen; ++i) {
	paramOut(os, csprintf("rxDesc%d.addr", i), rxRing[i].addr);
	paramOut(os, csprintf("rxDesc%d.length", i), rxRing[i].length);
	paramOut(os, csprintf("rxDesc%d.flags", i), rxRing[i].flags);
    }
    SERIALIZE_SCALAR(rxRingLen);
    SERIALIZE_SCALAR(rxRingPtr);
    SERIALIZE_SCALAR(rxDescDone);
    SERIALIZE_SCALAR(rxDescOff);
    int rxDmaState = this->rxDmaState;
    SERIALIZE_SCALAR(rxDmaState);
    SERIALIZE_SCALAR(rxXferLen);
    SERIALIZE_SCALAR(rxXferOff);

    int rxNumPkts = rxBuf.size();
    SERIALIZE_SCALAR(rxNumPkts);
    SERIALIZE_SCALAR(rxBufLen);
    //SERIZLIZE_SCALAR(rxBufMaxLen);

    bool rxPacketExists = rxPacket;
    SERIALIZE_SCALAR(rxPacketExists);
    SERIALIZE_SCALAR(rxPacketBytes);
 
    /*
     * Serialize the various parts of the transmit state machine
     */
    int txState = this->txState;
    SERIALIZE_SCALAR(txState);
    SERIALIZE_SCALAR(txEnable);
    for (int i = 0; i < txRingLen; ++i) {
	paramOut(os, csprintf("txDesc%d.addr", i), txRing[i].addr);
	paramOut(os, csprintf("txDesc%d.length", i), txRing[i].length);
	paramOut(os, csprintf("txDesc%d.flags", i), txRing[i].flags);
    }
    SERIALIZE_SCALAR(txRingLen);
    SERIALIZE_SCALAR(txRingPtr);
    SERIALIZE_SCALAR(txDescDone);
    SERIALIZE_SCALAR(txDescOff);
    int txDmaState = this->txDmaState;
    SERIALIZE_SCALAR(txDmaState);
    SERIALIZE_SCALAR(txXferLen);
    SERIALIZE_SCALAR(txXferOff);

    int txNumPkts = txBuf.size();
    SERIALIZE_SCALAR(txNumPkts);
    SERIALIZE_SCALAR(txBufLen);
    //SERIZLIZE_SCALAR(txBufMaxLen);

    bool txPacketExists = txPacket;
    SERIALIZE_SCALAR(txPacketExists);
    SERIALIZE_SCALAR(txPacketBytes);

    /*
     * If there's a pending transmit, store the time so we can
     * reschedule it later
     */
    Tick transmitTick = txEvent.scheduled() ? txEvent.when() - curTick : 0;
    SERIALIZE_SCALAR(transmitTick);

    /*
     * Keep track of pending interrupt status.
     */
    SERIALIZE_SCALAR(intrTick);
    SERIALIZE_SCALAR(cpuPendingIntr);
    Tick intrEventTick = 0;
    if (intrEvent)
	intrEventTick = intrEvent->when();
    SERIALIZE_SCALAR(intrEventTick);

    int i = 0;
    for (pktiter_t p = rxBuf.begin(); p != rxBuf.end(); ++p) {
        nameOut(os, csprintf("%s.rxBuf%d", name(), i++));
        (*p)->serialize(os);
    }
    if (rxPacketExists) {
        nameOut(os, csprintf("%s.rxPacket", name()));
        rxPacket->serialize(os);
    }
    i = 0;
    for (pktiter_t p = txBuf.begin(); p != txBuf.end(); ++p) {
        nameOut(os, csprintf("%s.txBuf%d", name(), i++));
        (*p)->serialize(os);
    }
    if (txPacketExists) {
        nameOut(os, csprintf("%s.txPacket", name()));
        txPacket->serialize(os);
    }
}

void
EtherDev::unserialize(Checkpoint *cp, const std::string &section)
{
    /*
     * unserialize device registers
     */
    UNSERIALIZE_SCALAR(regs.command);
    UNSERIALIZE_SCALAR(regs.config);
    UNSERIALIZE_SCALAR(regs.isr);
    UNSERIALIZE_SCALAR(regs.imr);
    UNSERIALIZE_SCALAR(regs.txRing);
    UNSERIALIZE_SCALAR(regs.rxRing);
    UNSERIALIZE_SCALAR(regs.txRingLen);
    UNSERIALIZE_SCALAR(regs.rxRingLen);
    UNSERIALIZE_SCALAR(regs.romAddr);
    UNSERIALIZE_SCALAR(regs.romData);
    UNSERIALIZE_SCALAR(regs.rxFiltCtrl);
    UNSERIALIZE_SCALAR(regs.rxFiltData);

    UNSERIALIZE_ARRAY(regs.perfect, EADDR_LEN);
    UNSERIALIZE_ARRAY(regs.hash, ES_HASH_SIZE);

    /*
     * unserialize the receive state machine
     */
    int rxState;
    UNSERIALIZE_SCALAR(rxState);
    this->rxState = (EsRxState)rxState;
    UNSERIALIZE_SCALAR(rxEnable);
    UNSERIALIZE_SCALAR(rxRingLen);
    rxRing = new es_desc[rxRingLen];
    for (int i = 0; i < rxRingLen; ++i) {
	paramIn(cp, section, csprintf("rxDesc%d.addr", i), rxRing[i].addr);
	paramIn(cp, section, csprintf("rxDesc%d.length", i), rxRing[i].length);
	paramIn(cp, section, csprintf("rxDesc%d.flags", i), rxRing[i].flags);
    }
    UNSERIALIZE_SCALAR(rxRingPtr);
    UNSERIALIZE_SCALAR(rxDescDone);
    UNSERIALIZE_SCALAR(rxDescOff);
    int rxDmaState;
    UNSERIALIZE_SCALAR(rxDmaState);
    this->rxDmaState = (EsDmaState)rxDmaState;
    UNSERIALIZE_SCALAR(rxXferLen);
    UNSERIALIZE_SCALAR(rxXferOff);

    int rxNumPkts;
    UNSERIALIZE_SCALAR(rxNumPkts);
    for (int i = 0; i < rxNumPkts; ++i) {
        PacketPtr p = new EtherPacket;
	p->unserialize(cp, csprintf("%s.rxBuf%d", section, i));
	rxBuf.push_back(p);
    }
    UNSERIALIZE_SCALAR(rxBufLen);
    //UNSERIZLIZE_SCALAR(rxBufMaxLen);

    bool rxPacketExists;
    UNSERIALIZE_SCALAR(rxPacketExists);
    
    rxPacket = NULL;
    if (rxPacketExists) {
        rxPacket = new EtherPacket;
        rxPacket->unserialize(cp, csprintf("%s.rxPacket", section));
    }
    UNSERIALIZE_SCALAR(rxPacketBytes);


    /*
     * unserialize the transmit state machine
     */
    int txState;
    UNSERIALIZE_SCALAR(txState);
    this->txState = (EsTxState)txState;
    UNSERIALIZE_SCALAR(txEnable);
    UNSERIALIZE_SCALAR(txRingLen);
    txRing = new es_desc[txRingLen];
    for (int i = 0; i < txRingLen; ++i) {
	paramIn(cp, section, csprintf("txDesc%d.addr", i), txRing[i].addr);
	paramIn(cp, section, csprintf("txDesc%d.length", i), txRing[i].length);
	paramIn(cp, section, csprintf("txDesc%d.flags", i), txRing[i].flags);
    }
    UNSERIALIZE_SCALAR(txRingPtr);
    UNSERIALIZE_SCALAR(txDescDone);
    UNSERIALIZE_SCALAR(txDescOff);
    int txDmaState;
    UNSERIALIZE_SCALAR(txDmaState);
    this->txDmaState = (EsDmaState)txDmaState;
    UNSERIALIZE_SCALAR(txXferLen);
    UNSERIALIZE_SCALAR(txXferOff);

    int txNumPkts;
    UNSERIALIZE_SCALAR(txNumPkts);
    for (int i = 0; i < txNumPkts; ++i) {
        PacketPtr p = new EtherPacket;
	p->unserialize(cp, csprintf("%s.txBuf%d", section, i));
	txBuf.push_back(p);
    }
    UNSERIALIZE_SCALAR(txBufLen);
    //UNSERIZLIZE_SCALAR(txBufMaxLen);

    bool txPacketExists;
    UNSERIALIZE_SCALAR(txPacketExists);
    
    txPacket = NULL;
    if (txPacketExists) {
        txPacket = new EtherPacket;
        txPacket->unserialize(cp, csprintf("%s.txPacket", section));
    }
    UNSERIALIZE_SCALAR(txPacketBytes);


    /*
     * If there's a pending transmit, store the time so we can
     * reschedule it later
     */
    Tick transmitTick;
    UNSERIALIZE_SCALAR(transmitTick);
    if (transmitTick)
	txEvent.schedule(curTick + transmitTick);

    /*
     * Keep track of pending interrupt status.
     */
    UNSERIALIZE_SCALAR(intrTick);
    UNSERIALIZE_SCALAR(cpuPendingIntr);
    Tick intrEventTick;
    UNSERIALIZE_SCALAR(intrEventTick);
    if (intrEventTick) {
	intrEvent = new IntrEvent(this, true);
	intrEvent->schedule(intrEventTick);
    }
}

Tick
EtherDev::cacheAccess(MemReqPtr &req)
{
    DPRINTF(EthernetPIO, "timing access to paddr=%#x (daddr=%#x)\n",
	    req->paddr, req->paddr - addr);
    return curTick + pioLatency;
}

//=====================================================================


BEGIN_DECLARE_SIM_OBJECT_PARAMS(EtherDevInt)

    SimObjectParam<EtherInt *> peer;
    SimObjectParam<EtherDev *> device;

END_DECLARE_SIM_OBJECT_PARAMS(EtherDevInt)

BEGIN_INIT_SIM_OBJECT_PARAMS(EtherDevInt)

    INIT_PARAM_DFLT(peer, "peer interface", NULL),
    INIT_PARAM(device, "Ethernet device of this interface")

END_INIT_SIM_OBJECT_PARAMS(EtherDevInt)

CREATE_SIM_OBJECT(EtherDevInt)
{
    EtherDevInt *dev_int = new EtherDevInt(getInstanceName(), device);

    EtherInt *p = (EtherInt *)peer;
    if (p) {
	dev_int->setPeer(p);
	p->setPeer(dev_int);
    }

    return dev_int;
}

REGISTER_SIM_OBJECT("EtherDevInt", EtherDevInt)


BEGIN_DECLARE_SIM_OBJECT_PARAMS(EtherDev)

    Param<Tick> tx_delay;
    Param<Tick> rx_delay;
    SimObjectParam<IntrControl *> intr_ctrl;
    Param<Tick> intr_delay;
    SimObjectParam<MemoryController *> mmu;
    SimObjectParam<PhysicalMemory *> physmem;
    Param<Addr> addr;
    SimObjectParam<TlaserPciDev *> pci_dev;
    Param<bool> rx_filter;
    Param<string> hardware_address;
    SimObjectParam<Bus*> header_bus;
    SimObjectParam<Bus*> payload_bus;
    SimObjectParam<HierParams *> hier;
    Param<Tick> pio_latency;
    Param<bool> dma_desc_free;
    Param<bool> dma_data_free;
    Param<Tick> dma_read_delay;
    Param<Tick> dma_write_delay;
    Param<Tick> dma_read_factor;
    Param<Tick> dma_write_factor;

END_DECLARE_SIM_OBJECT_PARAMS(EtherDev)

BEGIN_INIT_SIM_OBJECT_PARAMS(EtherDev)

    INIT_PARAM_DFLT(tx_delay, "Transmit Delay", 1000),
    INIT_PARAM_DFLT(rx_delay, "Receive Delay", 1000),
    INIT_PARAM(intr_ctrl, "Interrupt Controller"),
    INIT_PARAM_DFLT(intr_delay, "Interrupt Delay in microseconds", 0),
    INIT_PARAM(mmu, "Memory Controller"),
    INIT_PARAM(physmem, "Physical Memory"),
    INIT_PARAM(addr, "Device Address"),
    INIT_PARAM(pci_dev, "PCI Device Configuration"),
    INIT_PARAM_DFLT(rx_filter, "Enable Receive Filter", true),
    INIT_PARAM_DFLT(hardware_address, "Ethernet Hardware Address",
		    "00:99:00:00:00:01"),
    INIT_PARAM_DFLT(header_bus, "The IO Bus to attach to for headers", NULL),
    INIT_PARAM_DFLT(payload_bus, "The IO Bus to attach to for payload", NULL),
    INIT_PARAM_DFLT(hier, "Hierarchy global variables", &defaultHierParams),
    INIT_PARAM_DFLT(pio_latency, "Programmed IO latency", 1000),
    INIT_PARAM_DFLT(dma_desc_free, "DMA of Descriptors is free", false),
    INIT_PARAM_DFLT(dma_data_free, "DMA of Data is free", false),
    INIT_PARAM_DFLT(dma_read_delay, "fixed delay for dma reads", 0),
    INIT_PARAM_DFLT(dma_write_delay, "fixed delay for dma writes", 0),
    INIT_PARAM_DFLT(dma_read_factor, "multiplier for dma reads", 0),
    INIT_PARAM_DFLT(dma_write_factor, "multiplier for dma writes", 0)

END_INIT_SIM_OBJECT_PARAMS(EtherDev)


CREATE_SIM_OBJECT(EtherDev)
{
    int eaddr[6];
    sscanf(((string)hardware_address).c_str(), "%x:%x:%x:%x:%x:%x",
	   &eaddr[0], &eaddr[1], &eaddr[2], &eaddr[3], &eaddr[4], &eaddr[5]);

    return new EtherDev(getInstanceName(), intr_ctrl, intr_delay, physmem,
			tx_delay, rx_delay, rx_filter, eaddr,
			mmu, addr, hier, header_bus, payload_bus, 
			pio_latency,
			dma_desc_free, dma_data_free,
			dma_read_delay, dma_write_delay,
			dma_read_factor, dma_write_factor);
}

REGISTER_SIM_OBJECT("EtherDev", EtherDev)
