/* $Id: s.etherdev.hh 1.54 04/05/21 13:03:14-04:00 binkertn@ziff.eecs.umich.edu $ */

/*
 * Copyright (c) 2003-2004
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* @file
 * Device module for modelling an ethernet controller
 */

#ifndef __ETHERDEV_HH__
#define __ETHERDEV_HH__

#include "base/range.hh"
#include "base/statistics.hh"
#include "dev/etherint.hh"
#include "dev/etherpkt.hh"
#include "dev/if_esreg.h"
#include "dev/io_device.hh"
#include "sim/eventq.hh"

#define EADDR_LEN 6
#define ES_ROM_SIZE 6
#define ES_HASH_SIZE 128

/*
 * Ethernet device registers
 */
struct EsRegs {
    uint32_t	command;
    uint32_t	config;
    uint32_t	isr;
    uint32_t	imr;
    uint64_t	txRing;
    uint64_t	rxRing;
    uint32_t	txRingLen;
    uint32_t	rxRingLen;
    uint32_t	romAddr;
    uint32_t	romData;
    uint32_t	rxFiltCtrl;
    uint32_t	rxFiltData;

    uint8_t	perfect[EADDR_LEN];
    uint8_t	rom[ES_ROM_SIZE];
    uint16_t	hash[ES_HASH_SIZE];
};

class IntrControl;
class EtherDevInt;
class PhysicalMemory;
class BaseInterface;
class HierParams;
class Bus;

/*
 * Ethernet device model
 */
class EtherDev : public DmaDevice
{
  public:
    /* Receive State Machine states */
    enum EsRxState
    {
	rxIdle,
	rxDescRefr,
	rxDescRead,
	rxFifoBlock,
	rxFragWrite,
	rxDescWrite,
	rxAdvance
    };

    /* Transmit State Machine states */
    enum EsTxState
    {
	txIdle,
	txDescRefr,
	txDescRead,
	txFifoBlock,
	txFragRead,
	txDescWrite,
	txAdvance
    };

    enum EsDmaState
    {
	dmaIdle,
	dmaReading,
	dmaWriting,
	dmaReadWaiting,
	dmaWriteWaiting
    };

  private:
    Addr addr;
    static const Addr size = sizeof(EsRegs);

  protected:
    typedef std::deque<PacketPtr> pktbuf_t;
    typedef pktbuf_t::iterator pktiter_t;

  protected:
    EsRegs regs;

    EsRxState rxState;
    bool rxEnable;
    es_desc *rxRing;
    int rxRingLen;
    int rxRingPtr;
    bool rxDescDone;
    int rxDescOff;
    EsDmaState rxDmaState;
    bool rxDmaFree;
    int rxXferLen;
    int rxXferOff;
    pktbuf_t rxBuf;
    uint32_t rxBufLen;
    static const uint32_t rxBufMaxLen = 1024*1024;
    PacketPtr rxPacket;
    int rxPacketBytes;

    EsTxState txState;
    bool txEnable;
    es_desc *txRing;
    int txRingLen;
    int txRingPtr;
    bool txDescDone;
    int txDescOff;
    EsDmaState txDmaState;
    bool txDmaFree;
    int txXferLen;
    int txXferOff;
    pktbuf_t txBuf;
    uint32_t txBufLen;
    static const uint32_t txBufMaxLen = 1024*1024;
    PacketPtr txPacket;
    int txPacketBytes;

  protected:
    Tick dmaReadDelay;
    Tick dmaWriteDelay;

    Tick dmaReadFactor;
    Tick dmaWriteFactor;

    void *rxDmaData;
    Addr  rxDmaAddr;
    int   rxDmaLen;
    bool  doRxDmaRead();
    bool  doRxDmaWrite();
    void  rxDmaReadCopy();
    void  rxDmaWriteCopy();
    
    void *txDmaData;
    Addr  txDmaAddr;
    int   txDmaLen;
    bool  doTxDmaRead();
    bool  doTxDmaWrite();
    void  txDmaReadCopy();
    void  txDmaWriteCopy();

    void rxDmaReadDone();
    friend class EventWrapper<EtherDev, &EtherDev::rxDmaReadDone>;
    EventWrapper<EtherDev, &EtherDev::rxDmaReadDone> rxDmaReadEvent;

    void rxDmaWriteDone();
    friend class EventWrapper<EtherDev, &EtherDev::rxDmaWriteDone>;
    EventWrapper<EtherDev, &EtherDev::rxDmaWriteDone> rxDmaWriteEvent;

    void txDmaReadDone();
    friend class EventWrapper<EtherDev, &EtherDev::txDmaReadDone>;
    EventWrapper<EtherDev, &EtherDev::txDmaReadDone> txDmaReadEvent;

    void txDmaWriteDone();
    friend class EventWrapper<EtherDev, &EtherDev::txDmaWriteDone>;
    EventWrapper<EtherDev, &EtherDev::txDmaWriteDone> txDmaWriteEvent;

    bool dmaDescFree;
    bool dmaDataFree;

  protected:
    Tick rxDelay;
    Tick txDelay;

    void txReset();
    void rxReset();
    void chipReset();

    void rxKick();
    Tick rxKickTick;
    typedef EventWrapper<EtherDev, &EtherDev::rxKick> RxKickEvent;
    friend class RxKickEvent;

    void txKick();
    Tick txKickTick;
    typedef EventWrapper<EtherDev, &EtherDev::txKick> TxKickEvent;
    friend class TxKickEvent;

    /*
     * Retransmit event
     */
    void transmit();
    typedef EventWrapper<EtherDev, &EtherDev::transmit> TxEvent;
    friend class TxEvent;
    TxEvent txEvent;

    void rxDump() const;
    void txDump() const;

    /**
     * receive address filter
     */
    bool rxFilterEnable;
    bool rxHashMatch(PacketPtr &packet);
    bool rxFilter(PacketPtr &packet);

    void setSpeed(int speed);
    void setLink(bool link_up);
    void setDuplex(bool fdx);

    PhysicalMemory *physmem;

    /**
     * Interrupt management
     */
    IntrControl *intctrl;
    void devIntrPost(uint32_t interrupts);
    void devIntrClear(uint32_t interrupts);
    void devIntrChangeMask(uint32_t mask);

    Tick intrDelay;
    Tick intrTick;
    bool cpuPendingIntr;
    void cpuIntrPost(Tick when);
    void cpuInterrupt();
    void cpuIntrClear();

    typedef EventWrapper<EtherDev, &EtherDev::cpuInterrupt> IntrEvent;
    friend class IntrEvent;
    IntrEvent *intrEvent;

    EtherDevInt *interface;

  public:
    EtherDev(const std::string &name, IntrControl *i, Tick intr_delay,
	     PhysicalMemory *physmem, Tick tx_delay, Tick rx_delay,
	     bool rx_filter, const int eaddr[6],
	     MemoryController *mmu, Addr addr, HierParams *hier,
	     Bus *header_bus, Bus *payload_bus, Tick pio_latency,
	     bool dma_desc_free, bool dma_data_free,
	     Tick dma_read_delay, Tick dma_write_delay,
	     Tick dma_read_factor, Tick dma_write_factor);
    ~EtherDev();

    Fault read(MemReqPtr &req, uint8_t *data);
    Fault write(MemReqPtr &req, const uint8_t *data);

    bool cpuIntrPending() const;
    void cpuIntrAck() { cpuIntrClear(); }

    bool recvPacket(PacketPtr &packet);
    void transferDone();

    void setInterface(EtherDevInt *i) { assert(!interface); interface = i; }

    virtual void serialize(std::ostream &os);
    virtual void unserialize(Checkpoint *cp, const std::string &section);

  private:
    Stats::Scalar<> txBytes;
    Stats::Scalar<> rxBytes;
    Stats::Scalar<> txPackets;
    Stats::Scalar<> rxPackets;
    Stats::Formula txBandwidth;
    Stats::Formula rxBandwidth;
    Stats::Formula txPacketRate;
    Stats::Formula rxPacketRate;

  public:
    void regStats();

  private:
    Tick pioLatency;

  public:
    Tick cacheAccess(MemReqPtr &req);
};

/*
 * Ethernet Interface for an Ethernet Device
 */
class EtherDevInt : public EtherInt
{
  private:
    EtherDev *dev;

  public:
    EtherDevInt(const std::string &name, EtherDev *d)
	: EtherInt(name), dev(d) { dev->setInterface(this); }

    virtual bool recvPacket(PacketPtr &pkt) { return dev->recvPacket(pkt); }
    virtual void sendDone() { dev->transferDone(); }
};

#endif // __ETHERDEV_HH__
