/* $Id: s.if_esreg.h 1.7 03/09/05 23:05:32-00:00 binkertn $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* @file
 * Ethernet device register definitions
 */

#ifndef _IF_ESREG_H_
#define _IF_ESREG_H_

/*
 * ethsim register definitions
 */
#define ES_CMD			0x00	/* chip command register */
#define ES_CFG			0x04	/* configuration register */
#define ES_ISR			0x08	/* interrupt rtatus register */
#define ES_IMR			0x0c	/* interrupt mask register */
#define ES_TX_RING		0x10	/* first desc in tx ring */
#define ES_RX_RING		0x18	/* first desc in rx ring */
#define ES_TX_RING_LEN		0x20	/* number of tx descriptors */
#define ES_RX_RING_LEN		0x24	/* number of rx descriptors */
#define ES_ROM_ADDR		0x28	/* rom address (incr on data read) */
#define ES_ROM_DATA		0x2c	/* rom data at above address */
#define ES_RXFILT_CTL		0x30	/* receive filter control */
#define ES_RXFILT_DATA		0x34	/* receive filter data */
#define ES_LAST_REG		0x34

/* chip command register */
#define ES_CMD_TX_ENABLE	0x00000001	/* enable transmit */
#define ES_CMD_TX_DISABLE	0x00000002	/* disable transmit */
#define ES_CMD_RX_ENABLE	0x00000004	/* enable receive */
#define ES_CMD_RX_DISABLE	0x00000008	/* disable receive */
#define ES_CMD_TX_RESET		0x00000010	/* reset tx state machine */
#define ES_CMD_RX_RESET		0x00000020	/* reset rx state machine */
#define ES_CMD_CHIP_RESET	0x00000040	/* reset the chip */
#define ES_CMD_START		0x00000080	/* start transmit */
#define ES_CMD_RES		0xffffff00	/* reserved bits */
#define ES_CMD_RESET_ALL	(ES_CMD_TX_RESET | ES_CMD_RX_RESET | \
				 ES_CMD_CHIP_RESET)

/* configuration register */
#define ES_CFG_SPEED_10		0x00000000	/* baud 10MB/s */
#define ES_CFG_SPEED_100	0x00000001	/* baud 100MB/s */
#define ES_CFG_SPEED_1000	0x00000002	/* baud 1GB/s */
#define ES_CFG_SPEED_10000	0x00000003	/* baud 10GB/s */
#define ES_CFG_SPEED_MASK	0x00000003	/* speed mask */
#define ES_CFG_DUPLEX		0x00000004	/* full duplex */
#define ES_CFG_LINK		0x00000008	/* link status */
#define ES_CFG_RES		0xfffffff0	/* reserved bits */

/* interrupt status/mask register */
#define ES_INT_TX_OK		0x00000001	/* transmit finished */
#define ES_INT_TX_ERR		0x00000002	/* transmit error */
#define ES_INT_RX_OK		0x00000004	/* data received */
#define ES_INT_RX_ERR		0x00000008	/* receive error */
#define ES_INT_RX_OFLOW		0x00000010	/* receive buffer overrun */
#define ES_INT_RES		0xffffffe0	/* reserved bits */
#define ES_INT_ALL		(ES_INT_RX_OK|ES_INT_RX_ERR|ES_INT_RX_OFLOW| \
                                 ES_INT_TX_OK|ES_INT_TX_ERR)

/* rom addresses */
#define ES_ROM_EADDR		0x00		/* ethernet address offset */

/* receive filter control register */
#define ES_RXFILTCTL_ADDR	0x000003ff	/* filter reg offset */
#define ES_RXFILTCTL_RES	0x01fffc00	/* reserved bits */
#define ES_RXFILTCTL_UCHASH	0x02000000	/* use unicast hash */
#define ES_RXFILTCTL_MCHASH	0x04000000	/* use multicast hash */
#define ES_RXFILTCTL_PERFECT	0x08000000	/* accept perfect match*/
#define ES_RXFILTCTL_ALLUNI	0x10000000	/* accept all unicast */
#define ES_RXFILTCTL_ALLMULTI	0x20000000	/* accept all multicast */
#define ES_RXFILTCTL_BCAST	0x40000000	/* accept all broadcast */
#define ES_RXFILTCTL_ENABLE	0x80000000	/* enable receive filter */

#define ES_RXFILTADDR_PAR0	0x00000000
#define ES_RXFILTADDR_PAR1	0x00000002
#define ES_RXFILTADDR_PAR2	0x00000004
#define ES_FILTADDR_MCAST_LO	0x00000100 /* start of multicast filter */
#define ES_FILTADDR_MCAST_HI	0x000001FE /* end of multicast filter */
#define ES_MCAST_FILTER_LEN	256	   /* bytes */

/*
 * ethsim descriptor (tx and rx)
 */
struct es_desc {
    uint64_t	addr;		/* address of tx/rx data buffer */
    uint32_t	length;		/* length of tx/rx data buffer */
    uint32_t	flags;
};

/*
 * The way the ethsim device works is that a ring of descriptors is
 * created by simply allocating a contiguous block of physical
 * memory.  This block is organized in a ring structure.  The ring
 * length register indicates where the ring should wrap.
 */

/* Common flags */
#define ES_DESC_OWN		0x0001	/* descriptor is in use */
#define ES_DESC_MORE		0x0002	/* packet cont'd in  next descriptor */
#define ES_DESC_INT		0x0004	/* interrupt after access */

/* Receive flags */
#define ES_DESC_OFLOW		0x0008	/* receive buffer overflow */
#define ES_DESC_CRC		0x0010
#define ES_DESC_SHORT		0x0020	/* short packet received */
#define ES_DESC_LONG		0x0040	/* long packet received */

#define ES_DESC_RXERR		(ES_DESC_OFLOW)

/* Transmit flags */
#define ES_DESC_EC		0x0008	/* excessive collisions */
#define ES_DESC_CCNT		0x0070	/* collision count */
#define ES_DESC_UNDER		0x0010	/* transmit FIFO underrun */

#define ES_DESC_CC(X)	(((X)->flags & ES_DESC_CCNT) >> 4)

#define ES_DESC_TXERR		(ES_DESC_EC | ES_DESC_UNDER)

#endif /* _IF_ESREG_H_ */
