/* $Id: s.scsi_ctrl.cc 1.48 04/02/29 16:18:39-05:00 ehallnor@zazzer.eecs.umich.edu $ */

/** @file
 * Device definition for the scsi controller
 */

/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */

#include <cstddef>
#include <cstdlib>
#include <deque>
#include <string>
#include <vector>

#include "base/trace.hh"
#include "cpu/full_cpu/smt.hh"
#include "cpu/intr_control.hh"
#include "dev/disk_image.hh"
#include "dev/dma.hh"
#include "dev/pcireg.h"
#include "dev/scsi_ctrl.hh"
#include "dev/scsi_disk.hh"
#include "dev/scsi_none.hh"
#include "dev/tlaser_pcidev.hh"
#include "mem/bus/bus.hh"
#include "mem/bus/pio_interface.hh"
#include "mem/bus/pio_interface_impl.hh"
#include "mem/functional_mem/memory_control.hh"
#include "sim/builder.hh"
#include "sim/sim_object.hh"

using namespace std;

ScsiController::ScsiController(const string &name, DmaEngine *de,
			       IntrControl *i, const vector<ScsiDisk *> &disks,
			       MemoryController *mmu, Addr a,
			       HierParams *hier, Bus *bus)
    : DmaDevice(name), addr(a), pending(false), num_in_progress(0),
      dma(de), intctrl(i)
{
    mmu->add_child(this, Range<Addr>(addr, addr + size));

    if (bus) {
	pioInterface = newPioInterface(name, hier, bus, this,
				       &ScsiController::cacheAccess);
	pioInterface->addAddrRange(addr, addr + size - 1);
    }

    memset(&scsi_regs, 0, sizeof(scsi_regs));
    memset(device, 0, sizeof(ScsiDevice *) * SCSI_MAXTARG);

    for (int i = 0; i < disks.size(); i++) {
	int target = disks[i]->getTarget();

	if (target > SCSI_MAXTARG)
	    panic("Target too large!");

	device[target] = disks[i];
	device[target]->setController(this);
    }

    for (int i = 0; i < SCSI_MAXTARG; i++) {
	if (device[i] == NULL) {
	    device[i] = new ScsiNone("none", i);
	    device[i]->setController(this);
	}
    }

    memset(in_progress, 0, sizeof(in_progress));
}

ScsiController::~ScsiController()
{
    for (int i = 0; i < SCSI_MAXTARG; i++)
	if (device[i])
	    delete device[i];
}

Fault
ScsiController::read(MemReqPtr &req, uint8_t *data)
{
    Addr daddr = req->paddr - addr;

    if (daddr >= offsetof(ScsiData, done) &&
	daddr < offsetof(ScsiData, target)) {
	int target = (daddr - offsetof(ScsiData, done)) / sizeof(uint64_t);

	if (!device[target])
	    panic("trying to access uninitialized device!");

	int done = device[target]->isDone();

	DPRINTF(ScsiCtrl,
		"probe target=%d lun=%d done=%d txBytes=%d errno=%d\n",
		target, scsi_regs.lun, done,
		device[target]->bytesTX(), device[target]->getError());

	*(uint64_t *)data = done;
    }
    else
	panic("Invalid read access!\n");

    DPRINTF(ScsiCtrl, "read  va=%#x off=%#x value=%#x\n",
	    req->vaddr, daddr, *(uint64_t *)data);

    return No_Fault;
}

Fault
ScsiController::write(MemReqPtr &req, const uint8_t *data)
{
    Addr daddr = req->paddr - addr;

    DPRINTF(ScsiCtrl, "write va=%#x off=%#x value=%#x\n",
	    req->vaddr, daddr, *(uint64_t *)data);

    if (daddr == offsetof(ScsiData, start_io)) {
	int i;
	ScsiCommand cmd;
	DmaPhys pages[SCSI_MAXDMA_LEN];

	// start io
	DPRINTF(ScsiCtrl, "start io to target=%d lun=%d cmd=%#x sglen=%d\n",
		scsi_regs.target, scsi_regs.lun, scsi_regs.cmd[0],
		scsi_regs.sg_len);

	if (in_progress[scsi_regs.target]) 
	    panic("io already in progress!");

	in_progress[scsi_regs.target] = 1;
	num_in_progress++;
	if (scsi_regs.sg_len >= SCSI_MAXDMA_LEN)
	    panic("Transfer length exceeds maximum DMA size");

	for (i = 0; i < scsi_regs.sg_len; i++) {
	    DPRINTF(ScsiCtrl, "sg paddr=%#x len=%d\n",
		    scsi_regs.sg_map[i].paddr, scsi_regs.sg_map[i].len);

	    if (scsi_regs.sg_map[i].paddr == 0)
		panic("invalid dma access!");
	}

	for (i = 0; i < SCSI_COMMAND_LEN; i++) {
	    cmd[i] = scsi_regs.cmd[i];
	}

	for (i = 0; i < scsi_regs.sg_len; i++) {
	    DPRINTF(ScsiCtrl, "paddr[%d] = %#x len -> %d\n", i,
		    scsi_regs.sg_map[i].paddr, scsi_regs.sg_map[i].len);
	    pages[i].addr = scsi_regs.sg_map[i].paddr;
	    Addr offset = scsi_regs.sg_map[i].paddr & (VMPageSize - 1);
	    pages[i].length = VMPageSize;
	    if (offset)
		pages[i].length -= offset;

	    if (!pages[i].addr)
		panic("Cannot access page zero");

	    if (i && offset)
		panic("only first page may be unaligned");
	}
	pages[i].addr = 0;

	device[scsi_regs.target]->startIO(cmd, pages, scsi_regs.lun);
    }
    else if (daddr >= offsetof(ScsiData, done) &&
	     daddr < offsetof(ScsiData, target)) {
	int target = (daddr - offsetof(ScsiData, done)) / sizeof(uint64_t);

	DPRINTF(ScsiCtrl, "ack target=%d lun=%d\n", target, scsi_regs.lun);

	if (!in_progress[target])
	    panic("No scsi access in progress to ack!");

	in_progress[target] = 0;
	num_in_progress--;

	device[target]->ioDone();
    }
    else {
	uint64_t *ptr = (uint64_t *) &scsi_regs;
	ptr[daddr / sizeof(uint64_t)] = *(uint64_t *)data;
    }

    return No_Fault;
}

// check the interrupt level
void
ScsiController::intr_post()
{
    pending = (num_in_progress > 0);
    if (pending)
	intctrl->post(TheISA::INTLEVEL_IRQ1, TheISA::INTINDEX_SCSI);
}

void
ScsiController::intr_clear()
{
    pending = (num_in_progress > 0);
    if (!pending)
	intctrl->clear(TheISA::INTLEVEL_IRQ1, TheISA::INTINDEX_SCSI);
}

bool
ScsiController::intr_pending()
{ 
    return pending; 
}

Tick
ScsiController::cacheAccess(MemReqPtr &req)
{
    return curTick + 1000;
}

void 
ScsiController::serialize(std::ostream &os)
{
    SERIALIZE_SCALAR(pending);
    SERIALIZE_ARRAY(in_progress, SCSI_MAXTARG);
    SERIALIZE_SCALAR(num_in_progress);
    scsi_regs.serialize(os);
}

void 
ScsiController::unserialize(Checkpoint *cp, const std::string &section)
{
    UNSERIALIZE_SCALAR(pending);
    UNSERIALIZE_ARRAY(in_progress, SCSI_MAXTARG);
    UNSERIALIZE_SCALAR(num_in_progress);
    scsi_regs.unserialize(cp, section);
}

void 
ScsiData::serialize(std::ostream &os)
{
    SERIALIZE_SCALAR(start_io);
    SERIALIZE_ARRAY(done, SCSI_MAXTARG);
    SERIALIZE_SCALAR(target);
    SERIALIZE_SCALAR(lun);
    SERIALIZE_ARRAY(cmd, SCSI_COMMAND_LEN);
    SERIALIZE_SCALAR(length);
    SERIALIZE_SCALAR(sg_len);
    for (int i = 0; i < SCSI_MAXDMA_LEN; ++i) {
	paramOut(os, csprintf("sg_map_paddr%d",i), sg_map[i].paddr);
	paramOut(os, csprintf("sg_map_len%d", i), sg_map[i].len);
    }
}

void 
ScsiData::unserialize(Checkpoint *cp, const std::string &section)
{
    UNSERIALIZE_SCALAR(start_io);
    UNSERIALIZE_ARRAY(done, SCSI_MAXTARG);
    UNSERIALIZE_SCALAR(target);
    UNSERIALIZE_SCALAR(lun);
    UNSERIALIZE_ARRAY(cmd, SCSI_COMMAND_LEN);
    UNSERIALIZE_SCALAR(length);
    UNSERIALIZE_SCALAR(sg_len);
    for (int i = 0; i < SCSI_MAXDMA_LEN; ++i) {
	paramIn(cp, section, csprintf("sg_map_paddr%d",i), sg_map[i].paddr);
	paramIn(cp, section, csprintf("sg_map_len%d", i), sg_map[i].len);
    }
}

#ifndef DOXYGEN_SHOULD_SKIP_THIS

BEGIN_DECLARE_SIM_OBJECT_PARAMS(ScsiController)

    SimObjectParam<DmaEngine *> engine;
    SimObjectParam<IntrControl *> intr_ctrl;
    SimObjectParam<MemoryController *> mmu;
    Param<Addr> addr;
    SimObjectParam<TlaserPciDev *> pci_dev;
    SimObjectVectorParam<ScsiDisk *> disks;
    SimObjectParam<Bus*> io_bus;
    SimObjectParam<HierParams *> hier;

END_DECLARE_SIM_OBJECT_PARAMS(ScsiController)

BEGIN_INIT_SIM_OBJECT_PARAMS(ScsiController)

    INIT_PARAM(engine, "DMA Engine"),
    INIT_PARAM(intr_ctrl, "Interrupt Controller"),
    INIT_PARAM(mmu, "Memory Controller"),
    INIT_PARAM(addr, "Device Address"),
    INIT_PARAM(pci_dev, "PCI Device Configuration"),
    INIT_PARAM(disks, "Disks attached to this controller"),
    INIT_PARAM_DFLT(io_bus, "The IO Bus to attach to", NULL),
    INIT_PARAM_DFLT(hier, "Hierarchy global variables", &defaultHierParams)

END_INIT_SIM_OBJECT_PARAMS(ScsiController)


CREATE_SIM_OBJECT(ScsiController)
{
    return new ScsiController(getInstanceName(), engine, intr_ctrl, disks,
			      mmu, addr, hier, io_bus);
}

REGISTER_SIM_OBJECT("ScsiController", ScsiController)

#endif //DOXYGEN_SHOULD_SKIP_THIS
