/* $Id: s.scsi_ctrl.hh 1.29 04/02/11 16:07:49-05:00 binkertn@zizzer.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/** @file
 * Simos SCSI device declaration
 * @todo Check docs for correctness.
 */

// Some information gleaned from simos_scsi.h
#ifndef __SCSI_CTRL_HH__
#define __SCSI_CTRL_HH__

#include "dev/pcireg.h"
#include "dev/io_device.hh"
#include "dev/scsi_device.hh" // For SCSI_COMMAND_LEN
#include "sim/sim_object.hh"

// #define SIM_MAXCPUS	(1)
// #define SCSI_ADDR       (ULL(0xfffffcc500000000))
/** The maximum length of a DMA transaction. */
#define SCSI_MAXDMA_LEN (128)
/** The The maximum number of SCSI devices. */
#define SCSI_MAXTARG    (16)
//#define SCSI_MAXLUN     (16)

/**
 * Device register interface to the kernel
 */
struct ScsiData {
    /** Write to this location to start a DMA transaction. */
    uint64_t start_io;			// 0x00 write-only
    /** 
     * Read locations to check if an I/O operation is finished. 
     * Write the location to acknowledge the access.
     */
    uint64_t done[SCSI_MAXTARG];	// 0x08 read-write (write=ack)
    /** The target for the next I/O operation. */
    uint64_t target;			// 0x88 data fields
    /** The LUN for the next I/O operation. */
    uint64_t lun;			// 0x90
    /** The commands to use for the next I/O operation */
    uint64_t cmd[SCSI_COMMAND_LEN];	// 0x98
    /** The length of the next I/O operation. */
    uint64_t length;			// 0xF8
    /** The number of pages in the next I/O operation. */
    uint64_t sg_len;			// 0x100
    /** Starting addresses and lengths for each page in the next operation. */
    struct {
	/** The starting address. */
	uint64_t paddr;			// 0x108, 0x118 ...
	/** The length of the access. */
	uint64_t len;			// 0x110, 0x120 ...
    } sg_map[SCSI_MAXDMA_LEN];

    /**
     * Serialize this object to the given output stream.
     * @param os The stream to serialize to.
     */
    void serialize(std::ostream &os);

    /**
     * Reconstruct the state of this object from a checkpoint.
     * @param cp The checkpoint to use.
     * @param section The section name describing this object.
     */
    void unserialize(Checkpoint *cp, const std::string &section);
};

class DmaEngine;
class IntrControl;
class ScsiDevice;
class DiskImage;

/**
 * Device model for a very simple SCSI controller
 */
class ScsiController : public DmaDevice
{
    /** A Typical ScsiDisk */
    friend class ScsiDisk;
    /** A placeholder for an empty target. */
    friend class ScsiNone;

  protected:
    /** Base address of data */
    Addr addr;
    /** Size of scsi controller data */
    static const int size = sizeof(ScsiData);

  protected:
    /** There are outstanding requests. */
    bool pending;
    /** The in progress flag for each target.*/ 
    int in_progress[SCSI_MAXTARG];
    /** The total number of outstanding I/O requests. */
    int num_in_progress;

  protected:
    /** The dma controller. */
    DmaEngine *dma;
    /** 
     * The interrupt controller. 
     * @todo Can store in a smaller structure.
     */
    IntrControl *intctrl;
    /** The internal I/O control registers. */
    ScsiData scsi_regs;

  protected:
    /** Pointers to all the connected devices. */
    ScsiDevice *device[SCSI_MAXTARG];

  public:
    /**
     * Constructs and initialize this controller. 
     * @param name The name of this controller.
     * @param de The dma controller.
     * @param i The interrupt controller.
     * @param disks The connected devices.
     * @param mmu The memory controller.
     * @param addr The starting address of the memory map control regs.
     */
    ScsiController(const std::string &name, DmaEngine *de, IntrControl *i,
		   const std::vector<ScsiDisk *> &disks,
		   MemoryController *mmu, Addr addr, HierParams *hier,
		   Bus *bus);
    /**
     * Deletes the connected devices.
     */
    ~ScsiController();

    /**
     * Read a done field for a given target.
     * @param req Contains the address of the field to read.
     * @param data Return the field read.
     * @return The fault condition of the access.
     */
    Fault read(MemReqPtr &req, uint8_t *data);
    
    /**
     * Write to the mmapped I/O control registers.
     * @param req Contains the address to write to.
     * @param data The data to write.
     * @return The fault condition of the access.
     */
    Fault write(MemReqPtr &req, const uint8_t *data);

    /**
     * Post an interupt if there are outstanding requests.
     */
    void intr_post();
    
    /**
     * Clear an interupt if there are no outstanding requests.
     */
    void intr_clear();
    
    /**
     * Return true if there are outstanding requests.
     */
    bool intr_pending();
    
    /**
     * Serialize this object to the given output stream.
     * @param os The stream to serialize to.
     */
    void serialize(std::ostream &os);

    /**
     * Reconstruct the state of this object from a checkpoint.
     * @param cp The checkpoint use.
     * @param section The section name of this object
     */
    void unserialize(Checkpoint *cp, const std::string &section);

  public:
    Tick cacheAccess(MemReqPtr &req);
};
#endif // __SCSI_CTRL_HH__
