/* $Id: s.scsi_device.hh 1.12 03/10/29 23:30:18-05:00 ehallnor@zizzer.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/** @file
 * Base definition for a scsi device attached to the scsi bus.
 */

#ifndef __SCSI_DEVICE_HH__
#define __SCSI_DEVICE_HH__

#include "base/misc.hh"
#include "sim/sim_object.hh"

/** The maximum number of SCSI commands for an I/O request. */
#define SCSI_COMMAND_LEN	12

/** An array encapsulation of a SCSI I/O command. */
struct ScsiCommand {
    /** Storage for the commands. */
    uint8_t data[SCSI_COMMAND_LEN];
    /**
     * Return the command byte at the given index.
     * @param index The command byte to get.
     * @return The command byte.
     */
    uint8_t &operator[](int index)
    { 
	return data[index]; 
    }
};

class DmaPhys;
class ScsiController;

/**
 * A base class for SCSI devices.
 */
class ScsiDevice : public SimObject
{
  protected:
    /** The SCSI controller for this device. */
    ScsiController *ctrl;
    /** The target number of this device. */
    int target;

  public:
    /**
     * Construct and initialize this device.
     * @param name The name of this device.
     * @param tgt The target number for this device.
     */
    ScsiDevice(const std::string &name, int tgt)
	: SimObject(name), ctrl(NULL), target(tgt) 
    {
    }
    
    /**
     * Destroy this device.
     */
    virtual ~ScsiDevice() {}

    /**
     * Set the controller for this device.
     * @param c The SCSI controller.
     */
    void setController(ScsiController *c) {
	if (ctrl) panic("Cannot change controller!");
	ctrl = c;
    }

    /**
     * Return true if the last I/O request has finished.
     * @return True if the I/O request is finished.
     */
    virtual bool isDone() const = 0;

    /**
     * Return the number of bytes transmitted.
     * @return The number of bytes transmitted.
     */
    virtual int bytesTX() const = 0;

    /**
     * Return the last error code.
     * @return The last error code.
     */
    virtual int getError() const = 0;

    /**
     * Return the target number for this device.
     * @return The target number.
     */
    int getTarget() const { return target; }

    /**
     * Start an I/O transaction with the provided parameters.
     * @param cmd The SCSI command to perform.
     * @param pages The pages to transfer to/from
     * @param lun The LUN to access.
     */
    virtual void startIO(ScsiCommand cmd, DmaPhys *pages, int lun) = 0;

    /**
     * Called when the OS is finished with the current I/O request.
     */
    virtual void ioDone() = 0;
};

#endif // __SCSI_DEVICE_HH__
