/* $Id: s.scsi_disk.cc 1.54 04/04/06 10:01:58-07:00 stever@isabel.reinhardt.house $ */
/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */

/** @file
 * Definitions of a SCSI disk.
 */

///////////////////////////////////////////////////////////////////////
// Simple SCSI disk device emulation. Emulates a number of independent
// disks. Correctly models disk latencies and DMA transfers, but does
// not model disk controller contention (in fact, there is no real
// notion of a disk controller in this model).
//
// Data handling now supported. The DMA uses in all cases an
// intermediate buffer for the scatter/gather

#include <cerrno>
#include <cstring>
#include <deque>
#include <string>

#include "base/cprintf.hh" // csprintf
#include "base/trace.hh"
#include "cpu/full_cpu/smt.hh"
#include "dev/disk_image.hh"
#include "dev/scsi.hh"
#include "dev/scsi_ctrl.hh"
#include "dev/scsi_device.hh"
#include "dev/scsi_disk.hh"
#include "sim/builder.hh"
#include "sim/sim_object.hh"
#include "sim/universe.hh"

using namespace std;

ScsiDisk::ScsiDisk(const string &name, DiskImage *img, int delay, int secsize,
		   int tgt, const string &ser, bool use_interface)
    : ScsiDevice(name, tgt), dmareq(use_interface), inFixedLatency(0),
      image(img), serial(ser), sectorSize(secsize),
      readDoneCB(this), writeDoneCB(this), diskDoneCB(this), fixedDoneCB(this)
{
    diskDelay = delay * ticksPerSecond / 1000;
    // initialize device
    isRead = false;
    size = 0;
    sector = 0;
    done = false;
    txBytes = 0;
    error = 0;
    currSector = 0;
    lun = 0;

    data = new uint8_t[sectorSize];
    memset(data, 0, sectorSize);
    memset(&command, 0, sizeof(command));
    memset(phys, 0, sizeof(phys));
}

ScsiDisk::~ScsiDisk()
{
    delete [] data;
}

void
ScsiDisk::writeDone()
{
    DPRINTF(DMAReadVerbose, "write new sector\n");

    // this is a writable disk
    int len = image->write(data, currSector++);
    if (len != sectorSize)
	panic("Can't write to %s. only %d of %d written. errno=%d\n",
	      name(), len, sectorSize, errno);
}

void
ScsiDisk::readDone()
{
    DPRINTF(DMAWriteVerbose, "read new sector\n");
    int len = image->read(data, currSector++);
    if (len != sectorSize)
	panic("Can't read from %s. only %d of %d read. errno=%d\n",
	      name(), len, sectorSize, errno);
}

// Start disk i/o. Arguments are:
//  - cmd         = SCSI command
//  - pages       = array of page pointers for transfer
void
ScsiDisk::startIO(ScsiCommand c, DmaPhys *pg, int l)
{
    int i;
    string op;

    // XXXXXXXXXXX
    // Unfortunately, the kernel does not give a list of phsysical
    // pages to us. If the buffer was originally in kseg0, the actual
    // physical address (not paged aligned) is given to us).
    // simscsi.c (line 674) should eventually be fixed. For now,
    // we align the pages ourselves
    //
    // Actually, it would be even better if we had physical
    // addresses. For now we convert it here.
    // XXXXXXXXXXX
    for (i = 0; i < SCSI_MAXDMA_LEN && pg[i].addr; i++ ) {
	DPRINTF(ScsiDisk, "paddr[%d] = %#x len -> %d\n", i, pg[i].addr,
		pg[i].length);
	phys[i] = pg[i];
    }
		
    int nphys = i;
		
    if (i >= SCSI_MAXDMA_LEN)
	panic("DMA size of %d pages too large.", i);
    phys[i].addr = 0;
		
    DPRINTF(ScsiDisk, "Disk startIO\n", curTick);
    DPRINTF(ScsiDisk, "paddrs=%#x, %#x, %#x, %#x\n", phys[0].addr,
	    phys[1].addr, phys[2].addr, phys[3].addr);
		
    // Decode the command and execute it, if it doesn't require a data
    // transfer. Otherwise (for I/O commands), set up parameters. Actual
    // I/O will be kicked below.
    command = c;
    lun = l;
		
    doCommand();
    if (done) {
	DPRINTF(ScsiDisk, "diskio done\n");
				
	// raise interrupt to signal operation done
	if (ctrl)
	    ctrl->intr_post();
	return;
    }

    // Set up DMA request area.  Start time (last param) is set to 0
    // here, but will be overridden later.
    dmareq.init(doneCB(sector >= 0), bufferEmptyCB(sector >= 0, isRead),
		isRead, phys, nphys, txBytes, data, sectorSize, 0);

    currSector = sector;

    op = (sector >= 0) ? (isRead ? "RD" : "WR") : "SYS";

    DPRINTF(ScsiDisk, "ScsiDisk %s size %d sector %d dma=[", op, size, sector);

    for (i = 0; i < 8 && i < nphys; i++)
	DPRINTFR(ScsiDisk, "%#x", (uint64_t)phys[i].addr);
    DPRINTFR(ScsiDisk, "%s]\n", (i < nphys) ? "...." : "");

    if (sector < 0) {
	// The callback used to be assigned here, until I consolidated
	// everything in the init() function.  Just checking that I
	// didn't mess it up.  We can remove this after a few weeks if
	// we don't see any problens.  Steve 12/5/03
	assert(dmareq.dmaDoneCB == &diskDoneCB);
	dmareq.start = curTick;
	ctrl->dma->doTransfer(&dmareq);
    } else
	doFixedLatency();
}

void
ScsiDisk::diskDone()
{
    DPRINTF(ScsiDisk, "diskDone!\n");

    if (done)
	panic("Bogus dma state");

    done = true; // operation done

    // raise interrupt to signal operation done
    if (ctrl)
	ctrl->intr_post();
}

// Model the DMA of a fixed latency disk.
void
ScsiDisk::doFixedLatency()
{
    DPRINTF(ScsiDisk, "doFixedLatency!\n");

    Tick sector_time = diskDelay / size;

    if (inFixedLatency) {
	// this must be DMA done routine callback from 0 latency DMA
	inFixedLatency++;
	return;
    }

    while (true) {
	if (dmareq.totalMoved() == size * sectorSize) {
	    // DMA'ed the whole thing. Signal that we are done.
	    diskDone();
	    return;
	}

	if (dmareq.totalMoved() >= size * sectorSize)
	    panic("somehow dma moved more data than possible");

	inFixedLatency = 1;

	// The callback used to be assigned here, until I consolidated
	// everything in the init() function.  Just checking that I
	// didn't mess it up.  We can remove this after a few weeks if
	// we don't see any problens.  Steve 12/5/03
	assert(dmareq.dmaDoneCB == &fixedDoneCB);
	dmareq.start = curTick + sector_time;
	ctrl->dma->doTransfer(&dmareq);

	if (inFixedLatency == 1) {
	    inFixedLatency = 0;
	    break;
	}

	// we must be in 0 latency DMA model, and the DMA scheduled
	// abovehas finished, so loop around to do next sector
	inFixedLatency = 0;
    }
}

// I/O complete. Tells disk device that OS is done with this i/o.
void
ScsiDisk::ioDone()
{
    if (!done)
	panic("Trying to complete an IO event when we're not done!");

    // clear interrupt for this disk
    if (ctrl)
	ctrl->intr_clear();

    done = false;
}

// Do SCSI command
//
// Side effect:
//  - for commands requiring a data transfer: sets up parameters
//  - for other commands: performs them, sets results and done field.
//
void
ScsiDisk::doCommand()
{
    if (done)
	panic("trying to do a completed command");

#if TRACING_ON
    const char *opcode = ScsiOpcode(command[0]);
#endif

    DPRINTF(ScsiDisk, "SCSI Command %s executing\n", opcode);

    switch (command[0]) {
      case CMD_SEEK:
      case CMD_RDEFECTS:
      case CMD_ADD_DEFECTS:
      case CMD_REQ_SENSE:
      case CMD_SEND_DIAG:
      case CMD_MODE_SELECT:
      case CMD_PREVREM:
      case CMD_FORMAT:
      case CMD_STARTUNIT:
      case CMD_ERASE:
	DPRINTF(ScsiDisk, "Unsupported SCSI command %s\n", opcode);
	error = -1;
	done = true;
	txBytes = 0;
	return;
	
      case CMD_READ:
	isRead = 1;
	sector = (command[2] << 24) + (command[3] << 16) + (command[4] << 8) +
	    command[5];
	size = (command[7] << 8) + command[8];
	error = 0; // will succeed
	txBytes = size * sectorSize;
	return;
	
      case CMD_READ_6:
	isRead = 1;
	sector = ((command[1] & 0x1f) << 16) + (command[2] << 8) + command[3];
	size = command[4];
	if (size == 0) {
	    DPRINTF(ScsiDisk, "Suspicious sizeInSectors in ScsiDisk\n");
	    size = 256;
	}
	  
	error = 0; // will succeed
	txBytes = size * sectorSize;
	return;
	  
      case CMD_WRITE:
	isRead = 0;
	sector = (command[2] << 24) + (command[3] << 16) + (command[4] << 8) +
	    command[5];
	size = (command[7] << 8) + command[8];
	error = 0; // will succeed
	txBytes = size * sectorSize;
	return;

      case CMD_WRITE_6:
	isRead = 0;
	sector = (((command[1] << 3) >> 3) << 16) + (command[2] << 8) +
	    command[3];
	size = command[4];
	if (size == 0) {
	    DPRINTF(ScsiDisk, "Suspicious sizeInSectors in ScsiDisk\n");
	    size = 256;
	}

	error = 0; // will succeed
	txBytes = size * sectorSize;
	return;

      case CMD_READCAPACITY:
        {
	    // Use DMA --- alpha port
	    uint64_t size = image->size();
	    DirReadCapData cap;
	    error = 0;
	    txBytes = sizeof(DirReadCapData);
	    sector = -1; // signal this is sense
	    bzero((char *)&cap, sizeof(cap));
	    cap.lbn3 = (size >> 24) & 0xff;
	    cap.lbn2 = (size >> 16) & 0xff;
	    cap.lbn1 = (size >> 8) & 0xff;
	    cap.lbn0 = (size >> 0) & 0xff;
	    cap.block_len3 = 0;
	    cap.block_len2 = 0;
	    cap.block_len1 = (sectorSize >> 8)  & 0xff;
	    cap.block_len0 = (sectorSize >> 0)  & 0xff;
	    memcpy(data, (char *)&cap, sizeof(cap));
	    DPRINTF(ScsiDisk,
		    "ScsiDisk READCAPACITY size=%d (%#x,%#x,%#x,%#x)\n",
		    size, cap.lbn3, cap.lbn2, cap.lbn1, cap.lbn0);
	    return;
	}
	 
      case CMD_TST_UNIT_RDY:
	// XXX When does this return an error?
	// XXX What errors does it return?
	error = 0;
	done = true;
	txBytes = 0;
	return;

      case CMD_MODE_SENSE:
	// Return that disk is not write protected and block size is
	// 512 bytes
	// WARNING: Watch this. I omitted the last section of the
	// ModeSenseData data structure so that we could easily use
	// this structure when running on other platforms. SAH
	struct ModeSenseData msd;
	bzero((char *)&msd, sizeof(msd));
	msd.wprot = 0;
	msd.bd_len = 8;
	msd.block_descrip[5] = 0;
	msd.block_descrip[6] = sectorSize >> 8;
	msd.block_descrip[7] = 0;
	// this information needs to be DMA'ed to memory. Set up correct
	// transfer size and place info in buffer.
	isRead = 1;
	txBytes = sizeof(struct ModeSenseData);
	error = 0;
	sector = -1; // signal this is sense
	memcpy(data, (char *)&msd,
	       sizeof(struct ModeSenseData));
	return;

      case CMD_INQUIRY:
	// XXX When does this return an error?
	// XXX What errors does it return?
	error = 0;
	if ((command[1] & 0x1) == 0x0) {
	    AllInqData inq;

	    isRead = 1;
	    txBytes = sizeof(inq);
	    sector = -1; // signal this is sense

	    memset(&inq, 0, sizeof(inq));

	    inq.dtype = 0x0;
	    inq.pqual = (image && lun == 0) ? 0x0 : 0x3;
	    inq.dmodify = 0x0;
	    inq.rmb = 0x0;
	    inq.ansi = 0x2;
	    inq.ecma = 0x0;
	    inq.iso = 0x0;
	    inq.rdf = 0x2;
	    inq.trmiop = 0x0;
	    inq.aenc = 0x0;
	    inq.addlen = 0x1f;
	    char vid[8]  = { 'D', 'E', 'C',   0,   0,   0,   0,   0 };
	    char pid[16] = { 'R', 'Z', '1', 'B', 'A', '-', 'B', 'S',
			     ' ', '(', 'C', ')', ' ', 'D', 'E', 'C' };
	    char revlev[4] = { '3', '0', '0', '1' };

	    memcpy(inq.vid, vid, sizeof(inq.vid));
	    memcpy(inq.pid, pid, sizeof(inq.pid));
	    memcpy(inq.revlevel, revlev, sizeof(inq.revlevel));
	    memcpy(data, &inq, sizeof(AllInqData));
	} else {
	    EvpdInqData inq;

	    isRead = 1;
	    txBytes = sizeof(inq);
	    sector = -1; // signal this is sense

	    memset(&inq, 0, sizeof(inq));

	    switch (command[2]) {
	      case 0x00:
		inq.pqual = (image && lun == 0) ? 0 : 3;
		inq.page_code = 0x00;
		inq.page_length = 2;
		inq.page_codes[0] = 0x00;
		inq.page_codes[1] = 0x80;
		break;

	      case 0x80:
		inq.pqual = (image && lun == 0) ? 0 : 3;
		inq.page_code = 0x80;
		inq.page_length = serial.size();
		memcpy(inq.page_codes, serial.c_str(), serial.size());
		break;

	      default:
		DPRINTF(ScsiDisk, "Unsupported INQUIRY request %#x",
			command[2]);
		error = -1;
		done = true;
		txBytes = 0;
		return;
		break;
	    }

	    memcpy(data, &inq, sizeof(EvpdInqData));
	}
	return;

      default:
	panic("ScsiDisk: scsi command not supported (%#x)\n", command[0]);
	return;
    }
}

void
ScsiDisk::serialize(ostream &os)
{
    int i;
    for (i=0; i<SCSI_COMMAND_LEN; ++i) {
	paramOut(os, csprintf("comm%d",i), command[i]);
    }
    SERIALIZE_SCALAR(isRead);
    SERIALIZE_SCALAR(size);
    SERIALIZE_SCALAR(sector);
    SERIALIZE_SCALAR(done);
    SERIALIZE_SCALAR(txBytes);
    SERIALIZE_SCALAR(error);
    for (i=0; i < SCSI_MAXDMA_LEN; ++i) {
	paramOut(os, csprintf("phys_addr%d",i), phys[i].addr);
	paramOut(os, csprintf("phys_len%d",i), phys[i].length);
    }
    SERIALIZE_ARRAY(data, sectorSize);
    SERIALIZE_SCALAR(currSector);
    SERIALIZE_SCALAR(lun);
    SERIALIZE_SCALAR(inFixedLatency);
    
    int doneCB = -1;
    int bufferCB = -1;
    if (dmareq.dmaDoneCB == &readDoneCB) {
	doneCB = 0;
    } else if (dmareq.dmaDoneCB == &writeDoneCB) {
	doneCB = 1;
    } else if (dmareq.dmaDoneCB == &diskDoneCB) {
	doneCB = 2;
    } else if (dmareq.dmaDoneCB == &fixedDoneCB) {
	doneCB = 3;
    } else if (dmareq.dmaDoneCB == NULL) {
	doneCB = 4;
    }
    if (dmareq.bufferCB == &readDoneCB) {
	bufferCB = 0;
    } else if (dmareq.bufferCB == &writeDoneCB) {
	bufferCB = 1;
    } else if (dmareq.bufferCB == &diskDoneCB) {
	bufferCB = 2;
    } else if (dmareq.bufferCB == &fixedDoneCB) {
	bufferCB = 3;
    } else if (dmareq.bufferCB == NULL) {
	bufferCB = 4;
    }    
    assert(doneCB != -1);
    assert(bufferCB != -1);
    SERIALIZE_SCALAR(doneCB);
    SERIALIZE_SCALAR(bufferCB);

    // output the name of the dma request
    paramOut(os, csprintf("dmaReqName"), dmareq.name());

    // create a section for the dma request
    nameOut(os, dmareq.name());
    paramOut(os, csprintf("parent"), name());
    paramOut(os, csprintf("id"), 0);
    dmareq.serialize(os);
}

void 
ScsiDisk::unserialize(Checkpoint *cp, const string &section)
{
    int i;
    for (i=0; i<SCSI_COMMAND_LEN; ++i) {
	paramIn(cp, section, csprintf("comm%d",i), command[i]);
    }
    UNSERIALIZE_SCALAR(isRead);
    UNSERIALIZE_SCALAR(size);
    UNSERIALIZE_SCALAR(sector);
    UNSERIALIZE_SCALAR(done);
    UNSERIALIZE_SCALAR(txBytes);
    UNSERIALIZE_SCALAR(error);
    for (i=0; i < SCSI_MAXDMA_LEN; ++i) {
	paramIn(cp, section, csprintf("phys_addr%d",i), phys[i].addr);
	paramIn(cp, section, csprintf("phys_len%d",i), phys[i].length);
    }
    UNSERIALIZE_ARRAY(data, sectorSize);
    UNSERIALIZE_SCALAR(currSector);
    UNSERIALIZE_SCALAR(lun);
    UNSERIALIZE_SCALAR(inFixedLatency);

    std::string dmaReqName;
    UNSERIALIZE_SCALAR(dmaReqName);
    dmareq.unserialize(cp, dmaReqName);
    
    // Fixup the DmaRequest
    int doneCB;
    int bufferCB;
    UNSERIALIZE_SCALAR(doneCB);
    UNSERIALIZE_SCALAR(bufferCB);
    switch (doneCB) {
      case 0:
	dmareq.dmaDoneCB = &readDoneCB;
	break;
      case 1:
	dmareq.dmaDoneCB = &writeDoneCB;
	break;
      case 2:
	dmareq.dmaDoneCB = &diskDoneCB;
	break;
      case 3:
	dmareq.dmaDoneCB = &fixedDoneCB;
	break;
      case 4:
	dmareq.dmaDoneCB = NULL;
    }
    switch (bufferCB) {
      case 0:
	dmareq.bufferCB = &readDoneCB;
	break;
      case 1:
	dmareq.bufferCB = &writeDoneCB;
	break;
      case 2:
	dmareq.bufferCB = &diskDoneCB;
	break;
      case 3:
	dmareq.bufferCB = &fixedDoneCB;
	break;
      case 4:
	dmareq.bufferCB = NULL;
    }
    dmareq.data = data;
    dmareq.phys = phys;
}

#ifndef DOXYGEN_SHOULD_SKIP_THIS

DEFINE_SIM_OBJECT_CLASS_NAME("ScsiDevice", ScsiDevice)

BEGIN_DECLARE_SIM_OBJECT_PARAMS(ScsiDisk)
    
    SimObjectParam<DiskImage *> image;
    Param<int> sector_size;
    Param<int> disk_delay;
    Param<int> target;
    Param<string> serial;
    Param<bool> use_interface;

END_DECLARE_SIM_OBJECT_PARAMS(ScsiDisk)

BEGIN_INIT_SIM_OBJECT_PARAMS(ScsiDisk)

    INIT_PARAM(image, "disk image"),
    INIT_PARAM_DFLT(sector_size, "disk sector size", 512),
    INIT_PARAM_DFLT(disk_delay, "fixed disk delay in milliseconds", 0),
    INIT_PARAM(target, "target scsi ID"),
    INIT_PARAM(serial, "disk serial number"),
    INIT_PARAM_DFLT(use_interface, "use dma interface", true)

END_INIT_SIM_OBJECT_PARAMS(ScsiDisk)


CREATE_SIM_OBJECT(ScsiDisk)
{
    return new ScsiDisk(getInstanceName(), image, disk_delay,
			sector_size, target, serial, use_interface);
}

REGISTER_SIM_OBJECT("ScsiDisk", ScsiDisk)

#endif //DOXYGEN_SHOULD_SKIP_THIS
