/* $Id: s.scsi_disk.hh 1.38 03/12/06 09:57:59-08:00 stever@isabel.reinhardt.house $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/** @file
 * Device model for a scsi disk
 */

#ifndef __SCSI_DISK_HH__
#define __SCSI_DISK_HH__

#include "sim/host.hh"

#include "dev/dma.hh"
#include "dev/scsi_device.hh"
#include "sim/sim_object.hh"

class DiskImage;

/**
 * SCSI Disk device model
 */
class ScsiDisk : public ScsiDevice, public DmaHolder
{
  protected:
    /** The current SCSI command being executed. */
    ScsiCommand command;

    /** True for a read request. */
    bool isRead;
    /** The transfer size in disk sectors. */
    uint64_t size;
    /** The disk sector to access on the current I/O request */
    int64_t sector;
    /** True if the current request is done. */
    bool done;
    /** The return value of the I/O access. */
    uint32_t txBytes;
    /** The error code of the accesses, 0 if none. */
    int error;
    /** The physical addresses and lengths for the current I/O request. */
    DmaPhys phys[SCSI_MAXDMA_LEN];
    /** Data buffer for a sector. */
    uint8_t *data;
    /** The current sector on the disk. */
    int64_t currSector;
    /** The storage for all parameters to initiate a DMA request. */
    DmaRequest dmareq;
    /** The LUN to access. */
    int lun;

  protected:
    /** Count of the current nummer of sectors being transmitted. */
    int inFixedLatency;
    /** The image that contains the data of this disk. */
    DiskImage *image;
    /** The serial number of this disk. */
    std::string serial;

  protected:
    /**
     * Read DMA complete callback
     */
    class ReadDone : public DmaCallback
    {
      public:
	/**
	 * Construct this callback.
	 * @param d The disk to call back.
	 */
	ReadDone(ScsiDisk *d) : disk(d) {}
	/** The disk to call back. */
	ScsiDisk *disk;
	/** Call ScsiDisk::readDone */
	virtual void process() { disk->readDone(); }
    };

    /**
     * Write DMA Complete callback
     */
    class WriteDone : public DmaCallback
    {
      public:
	/**
	 * Construct this callback.
	 * @param d The disk to call back.
	 */
	WriteDone(ScsiDisk *d) : disk(d) {}
	/** The disk to call back. */
	ScsiDisk *disk;
	/** Call ScsiDisk::writeDone */
	virtual void process() { disk->writeDone(); }
    };

    /**
     * DMA complete
     */
    class DiskDone : public DmaCallback
    {
      public:
	/**
	 * Construct this callback.
	 * @param d The disk to call back.
	 */
	DiskDone(ScsiDisk *d) : disk(d) {}
	/** The disk to call back. */
	ScsiDisk *disk;
	/** Call ScsiDisk:diskDone */
	virtual void process() { disk->diskDone(); }
    };

    /**
     * Fixed disk latency callback
     */
    class FixedDone : public DmaCallback
    {
      public:
	/**
	 * Construct this callback.
	 * @param d The disk to call back.
	 */
	FixedDone(ScsiDisk *d) : disk(d) {}
	/** The disk to call back. */
	ScsiDisk *disk;
	/** Call ScsiDisk::doFixedLatency */
	virtual void process() { disk->doFixedLatency(); }
    };

    friend class ScsiDisk::ReadDone;
    friend class ScsiDisk::WriteDone;
    friend class ScsiDisk::DiskDone;
    friend class ScsiDisk::FixedDone;

  protected:
    /** The disk delay in XXX. @todo What is the unit? */
    int diskDelay;
    /** The size of a sector in bytes. */
    int sectorSize;
    
    /** The read DMA callback for when its buffer is empty. */
    ReadDone readDoneCB;
    /** The write DMA callback for when its buffer is empty. */
    WriteDone writeDoneCB;
    /** The DMA callback for when the I/O is finished. */
    DiskDone diskDoneCB;
    /** The general callback for a fixed latency disk. */
    FixedDone fixedDoneCB;

    /**
     * Read the next Sector from the disk.
     */
    void readDone();
    /**
     * Write the data to the current sector.
     */
    void writeDone();
    /**
     * Mark the current request as done, post an interrupt.
     */
    void diskDone();
    /**
     * Model the DMA for a fixed latency disk.
     */
    void doFixedLatency();

    /**
     * Setup the DMA request for the current command.
     */
    void doCommand();

    DmaCallback *doneCB(bool isFixed)
    {
	return isFixed ?
	    (DmaCallback *)&fixedDoneCB : (DmaCallback *)&diskDoneCB;
    }

    DmaCallback *bufferEmptyCB(bool isFixed, bool isRead)
    {
	return isFixed ?
	    (isRead ? (DmaCallback *)&readDoneCB : (DmaCallback *)&writeDoneCB)
	    : (DmaCallback *)0;
    }

  public:
    /**
     * Create and initialize this Disk.
     * @param name The name of this disk.
     * @param img The disk image of this disk.
     * @param disk_delay The disk delay in ???
     * @param sector_size The sector size in bytes.
     * @param target The SCSI target number.
     * @param ser The serial number of this disk.
     * @param use_interface Use the timing memory interface for DMA.
     */
    ScsiDisk(const std::string &name, DiskImage *img, int disk_delay,
	     int sector_size, int target, const std::string &ser,
	     bool use_interface);
    /**
     * Delete the data buffer.
     */
    ~ScsiDisk();

    // OS interface
    bool isDone() const { return done; }
    int bytesTX() const { return txBytes; }
    int getError() const { return error; }

    // Schedule transfers to take place at the correct simulated time
    void startIO(ScsiCommand cmd, DmaPhys *pages, int lun);

    // Tell "controller" that an i/o operation is complete.
    void ioDone();

    virtual DmaRequest *find_dmareq(uint32_t &id) {
        if (id == 0)
            return(&dmareq);
        else
            return(NULL);
    }

    /**
     * Serialize this object to the given output stream.
     * @param os The stream to serialize to.
     */
    void serialize(std::ostream &os);

    /**
     * Reconstruct the state of this object from a checkpoint.
     * @param cp The checkpoint to use.
     * @param section The section name describing this object.
     */
    void unserialize(Checkpoint *cp, const std::string &section);
};


#endif // __SCSI_DISK_HH__
