/* $Id: s.scsi_none.cc 1.18 03/12/06 09:57:59-08:00 stever@isabel.reinhardt.house $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

#include <errno.h>
#include <string.h>

#include <deque>
#include <string>

#include "cpu/full_cpu/smt.hh"

#include "dev/scsi.hh"
#include "dev/scsi_ctrl.hh"
#include "dev/scsi_device.hh"
#include "dev/scsi_none.hh"
#include "base/trace.hh"

#include "sim/sim_object.hh"

using namespace std;

ScsiNone::ScsiNone(const string &name, int tgt)
    : ScsiDevice(name, tgt), dmareq(true), done(false), dmaDoneCB(this)
{
}

ScsiNone::~ScsiNone()
{}

void
ScsiNone::startIO(ScsiCommand c, DmaPhys *pg, int l)
{
    DPRINTF(ScsiNone, "ScsiNone request\n");

    if (done)
	panic("starting a finished request");

    command = c;
    lun = l;
    done = false;
    txBytes = 0;
    doCommand();

    if (done) {
	DPRINTF(ScsiNone, "ScsiNone IO done\n");
	if (ctrl)
	    ctrl->intr_post();
    }

    assert(dmareq.data);

    int nphys;
    for (nphys = 0; nphys < SCSI_MAXDMA_LEN && pg[nphys].addr; nphys++ ) {
	DPRINTF(ScsiNone, "paddr[%d]=%#x len=%d\n", nphys,
		pg[nphys].addr, pg[nphys].length);
	phys[nphys] = pg[nphys];
    }

    if (nphys >= SCSI_MAXDMA_LEN)
	panic("Trying to dma too many pages");

    dmareq.init(&dmaDoneCB, 0, true, phys, nphys, txBytes,
		dmareq.data, dmareq.datalen, curTick);

    ctrl->dma->doTransfer(&dmareq);
}

void
ScsiNone::dmaDone()
{
    DPRINTF(ScsiNone, "dmaDone!\n");

    if (done)
	panic("DMA already done!?");

    done = true;

    if (dmareq.data) {
	delete [] dmareq.data;
	dmareq.data = NULL;
    }

    if (ctrl)
	ctrl->intr_post();
}

void
ScsiNone::ioDone()
{
    if (ctrl)
	ctrl->intr_clear();

    done = false;
}

void
ScsiNone::doCommand()
{
    DPRINTF(ScsiNone, "SCSI Command %s executing\n", ScsiOpcode(command[0]));

    switch (command[0]) {
      case CMD_INQUIRY:
	if ((command[1] & 0x1) != 0x0) {
	    done = true;
	    error = -1;
	    txBytes = 0;
	} else {
	    assert(dmareq.data == NULL);
	    dmareq.datalen = sizeof(AllInqData);
	    dmareq.data = new uint8_t[sizeof(AllInqData)];
	    AllInqData *inq = (AllInqData *)dmareq.data;
	    error = 0;
	    txBytes = sizeof(*inq);

	    memset(inq, 0, sizeof(*inq));

	    inq->dtype = 0x0;
	    inq->pqual = 0x3;
	    inq->dmodify = 0x0;
	    inq->rmb = 0x0;
	    inq->ansi = 0x2;
	    inq->ecma = 0x0;
	    inq->iso = 0x0;
	    inq->rdf = 0x2;
	    inq->trmiop = 0x0;
	    inq->aenc = 0x0;
	    inq->addlen = 0x1f;
	}
	return;

      case CMD_READ:
      case CMD_READ_6:
      case CMD_WRITE:
      case CMD_WRITE_6:
      case CMD_READCAPACITY:
      case CMD_MODE_SENSE:
      case CMD_SEEK:
      case CMD_RDEFECTS:
      case CMD_ADD_DEFECTS:
      case CMD_REQ_SENSE:
      case CMD_SEND_DIAG:
      case CMD_MODE_SELECT:
      case CMD_PREVREM:
      case CMD_FORMAT:
      case CMD_STARTUNIT:
      case CMD_ERASE:
      case CMD_TST_UNIT_RDY:
	done = true;
	error = -1;
	txBytes = 0;
	return;

      default:
	panic("ScsiNone: scsi command not supported (0x%x)\n", command[0]);
	return;
    }
}
