/* $Id: s.tlaser_clock.cc 1.43 03/12/06 09:57:59-08:00 stever@isabel.reinhardt.house $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* @file
 * Clock interrupts
 * Clock tick frequency is assumed to be 10millseconds
 * (should confirm this somewhere eventually)
 *
 * The SimOS console does not manage the TLINTRMASK register.
 * Therefore, we start the clock simply at the start of the simulation
 * Since IPL==max at that point anyway, it does not matter.
 */

#include <cstddef>
#include <cstdlib>
#include <string>
#include <vector>

#include "base/inifile.hh"
#include "base/str.hh"	// for to_number
#include "base/trace.hh"
#include "cpu/exec_context.hh"
#include "cpu/intr_control.hh"
#include "dev/tlaser_clock.hh"
#include "dev/turbolaser.hh"
#include "sim/builder.hh"
#include "sim/sim_object.hh"
#include "sim/system.hh"
#include "sim/universe.hh"

using namespace std;

TlaserClock::ClockEvent::ClockEvent(TlaserClock *clk)
    : Event(&mainEventQueue), clock(clk)
{
    interval = ticksPerSecond / clk->frequency();
    DPRINTF(Clock, "clock interval = %d cycles\n", interval);
}

void
TlaserClock::ClockEvent::process()
{
    DPRINTF(Clock, "event!\n");

    clock->interrupt();
    schedule(curTick + interval);
}

const char *
TlaserClock::ClockEvent::description()
{
    return "turbolaser clock";
}

///////////////////////////////////////////////////////////////////////
//
// Turbolaser

TlaserClock::TlaserClock(const string &name, IntrControl *intctrl, Tick f,
			 Tick d)
    : SimObject(name), freq(f), delay(d), event(this), intrctrl(intctrl)
{
    event.schedule(delay);

    for (int i = 0; i < Turbolaser::Max_CPUs; i++)
	interrupting[i] = false;
}

void
TlaserClock::ack(int cpu_id)
{
    interrupting[cpu_id] = false;
    intrctrl->clear(cpu_id, TheISA::INTLEVEL_IRQ2, 0);
}

void
TlaserClock::interrupt()
{
    int size = intrctrl->cpu->system->execContexts.size();

    for (int i = 0; i < size; i++){
	if (!interrupting[i]) {
	    interrupting[i] = true;
	    intrctrl->post(i, TheISA::INTLEVEL_IRQ2, 0);
	}
    }
}


void
TlaserClock::serialize(ostream &os)
{
    SERIALIZE_ARRAY(interrupting, 2);
    Tick when = event.when();
    SERIALIZE_SCALAR(when);
}

void
TlaserClock::unserialize(Checkpoint *cp, const std::string &section)
{
    UNSERIALIZE_ARRAY(interrupting, 2);
    Tick when;
    UNSERIALIZE_SCALAR(when);
    event.reschedule(when);
}

BEGIN_DECLARE_SIM_OBJECT_PARAMS(TlaserClock)

  SimObjectParam<IntrControl *> intr_control;
  Param<Tick> frequency;
  Param<Tick> delay;

END_DECLARE_SIM_OBJECT_PARAMS(TlaserClock)

BEGIN_INIT_SIM_OBJECT_PARAMS(TlaserClock)


    INIT_PARAM(intr_control, "interrupt controller"),
    INIT_PARAM(frequency, "clock interrupt frequency"),
    INIT_PARAM_DFLT(delay, "number of cycles to delay clock start", 40000000)

END_INIT_SIM_OBJECT_PARAMS(TlaserClock)


CREATE_SIM_OBJECT(TlaserClock)
{
    return new TlaserClock(getInstanceName(), intr_control, frequency, delay);
}

REGISTER_SIM_OBJECT("TlaserClock", TlaserClock)










