/* $Id: s.tlaser_ipi.cc 1.35 04/02/29 16:18:39-05:00 ehallnor@zazzer.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* @file
 * Inter-Processor Interrupt
 */

#include <string>
#include <vector>

#include "base/trace.hh"
#include "cpu/exec_context.hh"
#include "cpu/intr_control.hh"
#include "dev/tlaser_ipi.hh"
#include "dev/tlaserreg.h"
#include "dev/turbolaser.hh"
#include "mem/bus/bus.hh"
#include "mem/bus/pio_interface.hh"
#include "mem/bus/pio_interface_impl.hh"
#include "mem/functional_mem/memory_control.hh"
#include "sim/builder.hh"
#include "sim/system.hh"
#include "targetarch/ev5.hh"

using namespace std;

TlaserIpi::TlaserIpi(const string &name, MemoryController *mmu, Addr a,
		     HierParams *hier, Bus *bus)
    : PioDevice(name), addr(a), tlaser(NULL)
{
    mmu->add_child(this, Range<Addr>(addr, addr + size));

    if (bus) {
	pioInterface = newPioInterface(name, hier, bus, this,
				       &TlaserIpi::cacheAccess);
	pioInterface->addAddrRange(addr, addr + size - 1);
    }
}

/* @todo
 * Add intr_cntrl as a sim object parameter to signal interrupts
 * with less pointer chasing
 */

void
TlaserIpi::setTurbolaser(Turbolaser *t)
{
    if (tlaser != NULL)
	panic("turbolaser already set!");

    tlaser = t;
    for (int i = 0; i < Turbolaser::Max_CPUs; i++)
	tlaser->ipi_pending[i] = 0; //Initialize pending IPI's
}

Fault
TlaserIpi::write(MemReqPtr &req, const uint8_t *data)
{
    Addr daddr = req->paddr - addr;

    if (daddr == TLIPINTR_REG) {
	int target = *data;

	vector<ExecContext *> &xcvec = req->xc->system->execContexts;
	int size = xcvec.size();

	for (int i = 0; i < size; i++) {
	    if ((ULL(1) << i) & target) {
		DPRINTF(IPI, "send cpu=%d pending=%d from=%d\n", i,
			tlaser->ipi_pending[i] + 1, req->cpu_num);

		if (!tlaser->ipi_pending[i]) {
		    BaseCPU *cpu = xcvec[i]->cpu;
		    cpu->post_interrupt(TheISA::INTLEVEL_IRQ2, 1);
		}

		tlaser->ipi_pending[i]++;
	    }
	}

	return No_Fault;
    }

    DPRINTF(IPI, "write daddr=%#x val=%d\n", daddr, *data);

    return No_Fault;
}

Fault
TlaserIpi::read(MemReqPtr &req, uint8_t *data)
{
    DPRINTF(IPI, "read  daddr=%#x val=%d\n", req->paddr - addr, *data);

    return No_Fault;
}

Tick
TlaserIpi::cacheAccess(MemReqPtr &req)
{
    return curTick + 1000;
}

BEGIN_DECLARE_SIM_OBJECT_PARAMS(TlaserIpi)

    SimObjectParam<MemoryController *> mmu;
    Param<Addr> addr;
    SimObjectParam<Bus*> io_bus;
    SimObjectParam<HierParams *> hier;

END_DECLARE_SIM_OBJECT_PARAMS(TlaserIpi)

BEGIN_INIT_SIM_OBJECT_PARAMS(TlaserIpi)

    INIT_PARAM(mmu, "Memory Controller"),
    INIT_PARAM(addr, "Device Address"),
    INIT_PARAM_DFLT(io_bus, "The IO Bus to attach to", NULL),
    INIT_PARAM_DFLT(hier, "Hierarchy global variables", &defaultHierParams)

END_INIT_SIM_OBJECT_PARAMS(TlaserIpi)


CREATE_SIM_OBJECT(TlaserIpi)
{
    return new TlaserIpi(getInstanceName(), mmu, addr, hier, io_bus);
}

REGISTER_SIM_OBJECT("TlaserIpi", TlaserIpi)

