/* $Id: s.tlaser_mbox.cc 1.23 04/02/02 10:47:19-08:00 stever@isabel.reinhardt.house $ */

/* @file
 * Turbolaser Mailbox Register
 */

/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */

#include <cstddef>
#include <cstdlib>
#include <string>
#include <vector>

#include "base/misc.hh"
#include "base/trace.hh"
#include "cpu/full_cpu/smt.hh"
#include "dev/tlaser_mbox.hh"
#include "sim/builder.hh"
#include "sim/sim_object.hh"

using namespace std;

// from io/dec/mbox/mbox.h

struct MBox {
    // hardware portion
    // NB: the hardware portion of the mailbox must be on a 64 byte boundary
    uint32_t cmd;
    uint8_t mask;
    uint8_t rsv1;
    uint8_t hose;
    uint8_t rsv2;
    uint64_t rbadr;
    uint64_t wdata;
    uint32_t unpredictable0[2];
    uint32_t rdata;
    uint32_t unpredictable1;
    union {
#define mb_status status_un.stat
#define mb_don_bit status_un.status_64.don
#define mb_err_bit status_un.status_64.err
	struct status_64 {
	    uint64_t don:1,
		err:1,
		stat0:62;
	} status_64;
	uint64_t stat;
    } status_un;
    uint64_t unpredictable2[2];
    uint64_t unpredictable3[2];

    // software portion of the mailbox
    uint64_t bus_timeout;	// number of usec's to wait for comp
    uint64_t mbox_inuse;	// debug
    Addr mbox_reg;		// vaddr of mail box reg for this mbox
    Addr mbox_pa;		// mbox phys addr, set by mbox_pool_init
    void (*mbox_cmd)();		// adapter specific cmd field rtn
    int (*err_rtn)();		// adapter specific err rtn
				// if != 0, then bus_ctlr_ptr must be
				// filled in

#ifdef KERNEL
    bus_ctlr_common_t bus_ctlr_ptr; // used by err_rtn
#else
    void *bus_ctlr_ptr;
#endif
    MBox *next; // used to link mbox structures when in a list
#define PAD 8
    // pad out to next 64 byte boundary
    uint8_t pad[PAD];
};

#define BRDG_CSR  0x40000000  // b(ridge) bit <30> 1-cmd to remote adpt
#define WHOAREYOU BRDG_CSR    // who's on the other end of this hose?

TlaserMBox::TlaserMBox(const string &name, PhysicalMemory *m)
    : SimObject(name), pmem(m)
{}

void
TlaserMBox::access(MemReqPtr &req, uint64_t val)
{
    MBox *msgbox = (MBox *)pmem->dma_addr(val, sizeof(MBox));

    DPRINTF(Mbox, "access io_addr=%#x mem_addr=%#x\n",
	    req->paddr, val);

    DPRINTF(Mbox, "cmd=%#x mask=%#x hose=%#x rbadr=%#x\n",
	    msgbox->cmd, msgbox->mask, msgbox->hose, msgbox->rbadr);

    DPRINTF(Mbox, "wdata=%#x rdata=%#x\n",
	    msgbox->wdata, msgbox->rdata);

    switch (msgbox->cmd) {
      case WHOAREYOU:
	msgbox->rdata = 0xef00; // PCIA
	break;

      default:
	panic("TlaserMBox::access\n");
    }

    msgbox->status_un.status_64.err = 0;
    msgbox->status_un.status_64.don = 1;
}

BEGIN_DECLARE_SIM_OBJECT_PARAMS(TlaserMBox)

  SimObjectParam<PhysicalMemory *> physmem;

END_DECLARE_SIM_OBJECT_PARAMS(TlaserMBox)

BEGIN_INIT_SIM_OBJECT_PARAMS(TlaserMBox)

  INIT_PARAM(physmem, "physical memory")

END_INIT_SIM_OBJECT_PARAMS(TlaserMBox)

CREATE_SIM_OBJECT(TlaserMBox)
{
    return new TlaserMBox(getInstanceName(), physmem);
}

REGISTER_SIM_OBJECT("TlaserMBox", TlaserMBox)
