/* $Id: s.tlaser_mc146818.cc 1.42 04/02/29 16:18:40-05:00 ehallnor@zazzer.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* @file
 * Real Time Clock device
 */

#include <sys/time.h>

#include <string>
#include <vector>

#include "base/inifile.hh"
#include "base/str.hh"	// for to_number
#include "base/trace.hh"
#include "cpu/exec_context.hh"
#include "cpu/full_cpu/smt.hh"
#include "dev/tlaser_clock.hh"
#include "dev/tlaser_mc146818.hh"
#include "dev/tlaserreg.h"
#include "mem/bus/bus.hh"
#include "mem/bus/pio_interface.hh"
#include "mem/bus/pio_interface_impl.hh"
#include "mem/functional_mem/memory_control.hh"
#include "sim/builder.hh"

using namespace std;

#define UNIX_YEAR_OFFSET 52

TlaserMC146818::TlaserMC146818(const string &name, TlaserClock *c,
			       time_t init_time, MemoryController *mmu, Addr a,
			       HierParams *hier, Bus *bus)
    : PioDevice(name), clock(c), addr(a)
{
    set_time(init_time == 0 ? time(NULL) : init_time);
    mmu->add_child(this, Range<Addr>(addr, addr + size));

    if (bus) {
	pioInterface = newPioInterface(name, hier, bus, this,
				       &TlaserMC146818::cacheAccess);
	pioInterface->addAddrRange(addr, addr + size - 1);
    }
}

void
TlaserMC146818::set_time(time_t t)
{
    gmtime_r(&t, &tm);
    DPRINTFN("Real-time clock set to %s", asctime(&tm));
}

Fault
TlaserMC146818::read(MemReqPtr &req, uint8_t *data)
{
    Addr daddr = req->paddr - addr;
    Addr gbus_addr = daddr >> GBUS_BIT_SHIFT;

    DPRINTF(MC146818, "read  daddr=%#x size=%d\n", daddr, req->size);

    switch (req->size) {
      case sizeof(uint64_t):
	switch (gbus_addr) {
	  case RTC_SECOND:
	    *(uint64_t *)data = tm.tm_sec;
	    return No_Fault;
	  case RTC_MINUTE:
	    *(uint64_t *)data = tm.tm_min;
	    return No_Fault;
	  case RTC_HOUR:
	    *(uint64_t *)data = tm.tm_hour;
	    return No_Fault;
	  case RTC_DAY_OF_WEEK:
	    *(uint64_t *)data = tm.tm_mday;
	    return No_Fault;
	  case RTC_MONTH:
	    *(uint64_t *)data = tm.tm_mon + 1;
	    return No_Fault;
	  case RTC_YEAR:
	    *(uint64_t *)data = tm.tm_year - UNIX_YEAR_OFFSET;
	    return No_Fault;

	  case RTC_CONTROL_REGISTERA:
	    *(uint64_t *)data = 0x20;
	    return No_Fault;
	  case RTC_CONTROL_REGISTERB:
	    *(uint64_t *)data = 0x6;
	    return No_Fault;
	  case RTC_CONTROL_REGISTERC:
	    *(uint64_t *)data = 0x0;
	    clock->ack(req->xc->cpu_id);
	    return No_Fault;
	  case RTC_CONTROL_REGISTERD:
	    *(uint64_t *)data = 0x80;
	    return No_Fault;

	  default:
	    *(uint64_t *)data = 0;
	    return No_Fault;
	}
	break;

      case sizeof(uint32_t):
	switch (gbus_addr) {
	  case RTC_CONTROL_REGISTERC:
	    *(uint32_t *)data = 0;
	    clock->ack(req->xc->cpu_id);
	    return No_Fault;

	  default:
	    *(uint32_t *)data = 0;
	    return No_Fault;
	}
	break;

      default:
	return Machine_Check_Fault;
    }

    return No_Fault;
}

Fault
TlaserMC146818::write(MemReqPtr &req, const uint8_t *data)
{
    DPRINTF(MC146818, "write daddr=%#x val=%#x\n", req->paddr - addr,
	    *(uint64_t *)data);
    return No_Fault;
}

Tick
TlaserMC146818::cacheAccess(MemReqPtr &req)
{
    return curTick + 1000;
}

void
TlaserMC146818::serialize(ostream &os)
{
    SERIALIZE_SCALAR(tm.tm_sec);
    SERIALIZE_SCALAR(tm.tm_min);
    SERIALIZE_SCALAR(tm.tm_hour);
    SERIALIZE_SCALAR(tm.tm_mday);
    SERIALIZE_SCALAR(tm.tm_mon);
    SERIALIZE_SCALAR(tm.tm_year);
    SERIALIZE_SCALAR(tm.tm_wday);
    SERIALIZE_SCALAR(tm.tm_yday);
    SERIALIZE_SCALAR(tm.tm_isdst);
}

void
TlaserMC146818::unserialize(Checkpoint *cp, const std::string &section)
{
    UNSERIALIZE_SCALAR(tm.tm_sec);
    UNSERIALIZE_SCALAR(tm.tm_min);
    UNSERIALIZE_SCALAR(tm.tm_hour);
    UNSERIALIZE_SCALAR(tm.tm_mday);
    UNSERIALIZE_SCALAR(tm.tm_mon);
    UNSERIALIZE_SCALAR(tm.tm_year);
    UNSERIALIZE_SCALAR(tm.tm_wday);
    UNSERIALIZE_SCALAR(tm.tm_yday);
    UNSERIALIZE_SCALAR(tm.tm_isdst);
}

BEGIN_DECLARE_SIM_OBJECT_PARAMS(TlaserMC146818)

    SimObjectParam<TlaserClock *> clock;
    Param<time_t> time;
    SimObjectParam<MemoryController *> mmu;
    Param<Addr> addr;
    SimObjectParam<Bus*> io_bus;
    SimObjectParam<HierParams *> hier;

END_DECLARE_SIM_OBJECT_PARAMS(TlaserMC146818)

BEGIN_INIT_SIM_OBJECT_PARAMS(TlaserMC146818)

    INIT_PARAM(clock, "turbolaser clock"),
    // Time defaults to Jan 1, 2006 (midnight GMT) if not set.
    INIT_PARAM_DFLT(time, "System time to use "
		    "(0 for actual time, default is 1/1/06", ULL(1136073600)),
    INIT_PARAM(mmu, "Memory Controller"),
    INIT_PARAM(addr, "Device Address"),
    INIT_PARAM_DFLT(io_bus, "The IO Bus to attach to", NULL),
    INIT_PARAM_DFLT(hier, "Hierarchy global variables", &defaultHierParams)

END_INIT_SIM_OBJECT_PARAMS(TlaserMC146818)


CREATE_SIM_OBJECT(TlaserMC146818)
{
    return new TlaserMC146818(getInstanceName(), clock, time, mmu, addr,
			      hier, io_bus);
}

REGISTER_SIM_OBJECT("TlaserMC146818", TlaserMC146818)
