/* $Id: s.tlaser_node.cc 1.43 04/02/29 16:18:40-05:00 ehallnor@zazzer.eecs.umich.edu $ */

/* @file
 * Turbolaser system bus node (processor, memory, or IO)
 */

#include <deque>
#include <string>
#include <vector>

#include "base/trace.hh"
#include "cpu/exec_context.hh"
#include "dev/console.hh"
#include "dev/etherdev.hh"
#include "dev/scsi_ctrl.hh"
#include "dev/tlaser_clock.hh"
#include "dev/tlaser_ipi.hh"
#include "dev/tlaser_mbox.hh"
#include "dev/tlaser_node.hh"
#include "dev/tlaserreg.h"
#include "dev/turbolaser.hh"
#include "mem/bus/bus.hh"
#include "mem/bus/pio_interface.hh"
#include "mem/bus/pio_interface_impl.hh"
#include "mem/functional_mem/memory_control.hh"
#include "sim/builder.hh"
#include "sim/system.hh"

using namespace std;

TlaserNode::TlaserNode(const string &name, Turbolaser *t, int num, Type _type,
		       MemoryController *mmu, Addr a,
		       HierParams *hier, Bus *bus)
    : PioDevice(name), addr(a), tlaser(t), number(num),
      tl_cpu_mask(-1), type(_type)
{
    mmu->add_child(this, Range<Addr>(addr, addr + size));

    if (bus) {
	pioInterface = newPioInterface(name, hier, bus, this,
				       &TlaserNode::cacheAccess);
	pioInterface->addAddrRange(addr, addr + size - 1);
    }
}

Fault
TlaserNode::read(MemReqPtr &req, uint8_t *data)
{
    Addr daddr = req->paddr - addr;
    DPRINTF(Regs, "node%d read  da=%#x pa=%#x va=%#x size=%d\n",
	    number, daddr, req->paddr, req->vaddr, req->size);

    uint32_t reg = 0;
    ExecContext *xc = req->xc;
    int cpuid = xc->cpu_id;

    switch (req->size) {
      case sizeof(uint32_t):
	switch(daddr) {
	  case TLDEV_REG: // 0x00
	    switch (type) {
	      case KFTHA:
		*(uint32_t *)data = 0x2020;
		break;
	      case KFTIA:
		*(uint32_t *)data = 0x2020;
		break;
	      case MS7CC:
		*(uint32_t *)data = 0x5000;
		break;
	      case SimgleProc4M:
		*(uint32_t *)data = 0x8011;
		break;
	      case SimgleProc16M:
		*(uint32_t *)data = 0x8012;
		break;
	      case DualProc4M:
		*(uint32_t *)data = 0x8014;
		break;
	      case DualProc16M:
		*(uint32_t *)data = 0x8015;
		break;
	    }
	    // from tlsb_init
	    return No_Fault;

	  case TLBER_REG: // 0x40
	    *(uint32_t *)data = 0;
	    return No_Fault;

	  case TLCNR_REG: // 0x80
	    // tlsb_conf: nxv->lsb_lcnr
	    *(uint32_t *)data = 0; // ok
	    return No_Fault;

	  case TLILID1_REG:  // 0xa40
	    // return the interrupt vector.
	    // Accessed for I/O nodes only
	    //
	    if (tlaser->ethernet->cpuIntrPending()) {
		tlaser->ethernet->cpuIntrAck();

		// For now, hack in 0x830 (which corresponds to the entry
		// used by if_es.c
		//
		// XXX should have OS notify device of interrupt vector
		*(uint32_t *)data = 0x830;
	    } else if (tlaser->scsi->intr_pending()) {
		// For now, hack in 0x820 (which corresponds to the
		// entry used by cam_simos.c
		//
		// XXX should have OS notify device of interrupt vector
		*(uint32_t *)data = 0x820;
	    } else
		panic("No pending ethernet or scsi interrupt!");

	    return No_Fault;

	  case TLDIAG_REG: // 0x1000
	  case TLMODCONFIG_REG: // 0x10c0
	  case TLEPAERR_REG: // 0x1500
	    *(uint32_t *)data = 0;
	    return No_Fault;

	  case TLINTRSUM0_REG: // 0x1180
	  case TLINTRSUM1_REG: // 0x11c0
	    //CPUMASK needed here??? RGD
	    // TLINTRSUMx interrupt source register read
	    if (cpuid == 0 && tlaser->cons->intStatus()) {
		// first CPU only
		reg |= (1 << 0);
	    }

	    if (tlaser->clock->pending(cpuid))
		reg |= (1 << 6);

	    if (tlaser->ipi_pending[cpuid])
		reg |= (1 << 5);

	    if ((tlaser->ethernet->cpuIntrPending() ||
		 tlaser->scsi->intr_pending()) && cpuid == 0) {
		reg |= (1 << 2);   // summary bit set
		reg |= (1 << 12);  // node4   bit
	    }

	    if (!reg) {
		// XXX
		// XXX bugnion. known possible bug. System will hang.
		// XXX under certain race conditions, the system will
		// XXX get into the state where it issues sprious
		// XXX interrupts at infinitum
		// XXX
		// XXX Mail me if you can reproduce it but can't fix
		// XXX it. I have not seen it in a long long time.
		panic("Turbolaser: spurious interrupt\n");
	    }

	    *(uint32_t *)data = reg;
	    return No_Fault;

	  case IDPNSE0_REG: // 0x2a40
	    DPRINTF(Regs, "TlaserNode(%d): EMPTY READ daddr=%#x\n",
		    number, daddr);
	    *(uint32_t *)data = 0;
	    return No_Fault;

	  case IDPNSE1_REG: // 0x2140
	    // tlsb_iop.c::tiopconfl1  iopreg->idpnse1;
	    *(uint32_t *)data = 0x6;  // ((IDPNSE_CBLOK | IDPNSE_PWROKC))
	    return No_Fault;

	}

      case sizeof(uint64_t):
	switch(daddr) {
	  case TLINTRSUM0_REG:
	    if ((cpuid % 2) != 0)
		panic("Wrong cpu accessing TLINTRSUM\n");
	    switch (tlaser->intr_sum_type[cpuid]) {
	      case TLASER_INTRSUM_UART:
		// at sys_int_20, nothing else there
		*(uint64_t *)data = 1;
		return No_Fault;
	      case TLASER_INTRSUM_INTIM:
		*(uint64_t *)data = 0;
		tlaser->clock->ack(cpuid);
		return No_Fault;

	      case TLASER_INTRSUM_IPI:
		if (tlaser->ipi_pending[cpuid] <= 0)
		    panic("No Pending IPI\n");

		DPRINTF(Regs, "IPI ACK pending=%d \n",
			tlaser->ipi_pending[cpuid]);

		DPRINTF(IPI, "IPI ACK pending=%d \n",
			tlaser->ipi_pending[cpuid]);

		tlaser->ipi_pending[cpuid]--;

		if (!tlaser->ipi_pending[cpuid]) {
		    BaseCPU *cpu = xc->system->execContexts[cpuid]->cpu;
		    cpu->clear_interrupt(TheISA::INTLEVEL_IRQ2, 1);
		}
		return No_Fault;

	    }
	  case TLINTRSUM1_REG:
	    if ((cpuid % 2) != 1)
		panic("Wrong cpu accessing TLINTRSUM\n");
	    switch (tlaser->intr_sum_type[cpuid]) {
	      case TLASER_INTRSUM_UART:
		// at sys_int_20, nothing else there
		*(uint64_t *)data = 1;
		return No_Fault;
	      case TLASER_INTRSUM_INTIM:
		*(uint64_t *)data = 0;
		tlaser->clock->ack(cpuid);
		return No_Fault;

	      case TLASER_INTRSUM_IPI:
		if (tlaser->ipi_pending[cpuid] <= 0)
		    panic("No Pending IPI\n");

		DPRINTF(Regs, "IPI ACK pending=%d \n",
			tlaser->ipi_pending[cpuid]);
		DPRINTF(IPI, "IPI ACK pending=%d \n",
			tlaser->ipi_pending[cpuid]);
		tlaser->ipi_pending[cpuid]--;
		if (!tlaser->ipi_pending[cpuid]) {
		    BaseCPU *cpu = xc->system->execContexts[cpuid]->cpu;
		    cpu->clear_interrupt(TheISA::INTLEVEL_IRQ2, 1);
		}
		return No_Fault;
	    }
	}
	*(uint64_t *)data = 0;
	break;

      case sizeof(uint16_t):
      case sizeof(uint8_t):
      default:
	panic("invalid access size for turbolaser register!");
    }

    DPRINTFN("error read  node=%d daddr=%#x size=%d\n", number, daddr,
	     req->size);

    return No_Fault;
}

Fault
TlaserNode::write(MemReqPtr &req, const uint8_t *data)
{
    Addr daddr = req->paddr - addr;
    DPRINTF(Regs, "node%d write da=%#x pa=%#x va=%#x size=%d\n",
	    number, daddr, req->paddr, req->vaddr, req->size);

    // This hack comes from simos!!! always do word reads!
    req->size = sizeof(uint32_t);
    ExecContext *xc = req->xc;
    int cpuid = xc->cpu_id;

    switch (req->size) {
      case sizeof(uint32_t):
	switch (daddr) {
	  case TLBER_REG: // 0x40
	    *(uint32_t *)data = 0;
	    return No_Fault;

	  case TLMBPR_REG: // 0xc00
	    // Mailbox (store conditional)
	    tlaser->mbox->access(req, *(uint64_t *)data);
	    return No_Fault;

	  case TLCPUMASK_REG: //0xb00
	    DPRINTF(IPI, "TLCPUMASK_REG set to %d \n", *(uint32_t *)data);
	    tl_cpu_mask = *(uint32_t *)data;
	    return No_Fault;

	  case TLDIAG_REG: // 0x1000
	  case TLMODCONFIG_REG: // 0x10c0
	  case TLEPAERR_REG: // 0x1500
	  case ICCNSE_REG: // 0x2040
	  case IDPVR_REG: // 0x2b40
	  case IDPNSE0_REG: // 0x2a40
	  case IDPNSE1_REG: // 0x2140
	  case IDPNSE2_REG: // 0x2240
	  case IDPNSE3_REG: // 0x2340
	    DPRINTF(Regs, "TlaserNode(%d): EMPTY WRITE daddr=%#x size=%d\n",
		    number, daddr, req->size);
	    return No_Fault;

	    // This hack goes with the above hack in making all requests word
	    // requests.
#if 0
	}
      case sizeof(uint64_t):
	switch (daddr) {
#endif
	  case TLINTRSUM0_REG:
	    if ((cpuid % 2) != 0)
		panic("wrong cpu!\n");
	    tlaser->intr_sum_type[cpuid] = *(uint64_t *)data;
	    return No_Fault;
	  case TLINTRSUM1_REG:
	    if ((cpuid % 2) != 1)
		panic("wrong cpu!\n");
	    tlaser->intr_sum_type[cpuid] = *(uint64_t *)data;
	    return No_Fault;
#if 0
	  case TLMBPR_REG:
	    DPRINTF(Regs, "TlaserNode(%d): EMPTY WRITE daddr=%#x size=%d\n",
		    number, daddr, req->size);
	    return No_Fault;
#endif
	}
    }

    DPRINTFN("error write node=%d daddr=%#x size=%d\n", number, daddr,
	     req->size);

    return No_Fault;
}

Tick
TlaserNode::cacheAccess(MemReqPtr &req)
{
    return curTick + 1000;
}

void
TlaserNode::serialize(std::ostream &os)
{
    SERIALIZE_SCALAR(tl_cpu_mask);
}

void
TlaserNode::unserialize(Checkpoint *cp, const std::string &section)
{
    UNSERIALIZE_SCALAR(tl_cpu_mask);
}

const char *nodeTypeStrings[] = {
    "KFTHA", "KFTIA", "MS7CC", "SingleProc4M", "SingleProc16M",
    "DualProc4M", "DualProc16M" };

BEGIN_DECLARE_SIM_OBJECT_PARAMS(TlaserNode)

    Param<int> number;
    SimpleEnumParam<TlaserNode::Type> node_type;
    SimObjectParam<Turbolaser *> tlaser;
    SimObjectParam<MemoryController *> mmu;
    Param<Addr> addr;
    SimObjectParam<Bus*> io_bus;
    SimObjectParam<HierParams *> hier;

END_DECLARE_SIM_OBJECT_PARAMS(TlaserNode)

BEGIN_INIT_SIM_OBJECT_PARAMS(TlaserNode)

    INIT_PARAM_DFLT(number, "Node Number", 0),
    INIT_ENUM_PARAM(node_type, "Node Type", nodeTypeStrings),
    INIT_PARAM(tlaser, "Turbolaser"),
    INIT_PARAM(mmu, "Memory Controller"),
    INIT_PARAM(addr, "Device Address"),
    INIT_PARAM_DFLT(io_bus, "The IO Bus to attach to", NULL),
    INIT_PARAM_DFLT(hier, "Hierarchy global variables", &defaultHierParams)

END_INIT_SIM_OBJECT_PARAMS(TlaserNode)

CREATE_SIM_OBJECT(TlaserNode)
{
    return new TlaserNode(getInstanceName(), tlaser, number, node_type,
			  mmu, addr, hier, io_bus);
}

REGISTER_SIM_OBJECT("TlaserNode", TlaserNode)
