/* $Id: s.tlaser_pcia.cc 1.33 04/02/29 16:18:41-05:00 ehallnor@zazzer.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* @file
 * Turbolaser PCI interface adapter
 */

#include <cstddef>
#include <cstdlib>
#include <list>
#include <string>
#include <vector>

#include "base/misc.hh"
#include "base/trace.hh"
#include "dev/pciareg.h"
#include "dev/tlaser_pcia.hh"
#include "mem/bus/bus.hh"
#include "mem/bus/pio_interface.hh"
#include "mem/bus/pio_interface_impl.hh"
#include "mem/functional_mem/memory_control.hh"
#include "sim/builder.hh"
#include "sim/param.hh"
#include "sim/universe.hh"

TlaserPcia::TlaserPcia(const std::string &name, MemoryController *mmu, Addr a,
		       HierParams *hier, Bus *bus)
    : PioDevice(name), addr(a)
{
    mmu->add_child(this, Range<Addr>(addr, addr + size));

    if (bus) {
	pioInterface = newPioInterface(name, hier, bus, this,
				       &TlaserPcia::cacheAccess);
	pioInterface->addAddrRange(addr, addr + size - 1);
    }
}

bool
TlaserPcia::badaddr(Addr paddr) const
{
    Addr offset = paddr & PCIA_MASK;
    if (offset >= PCIA_PCI1 && offset < PCIA_PRESENT)
	return true;
    else
	return false;
}

Fault
TlaserPcia::read(MemReqPtr &req, uint8_t *data)
{
    Addr offset = req->paddr - addr;
    uint64_t res = 0;

    switch (offset) {
      case PCIA_PRESENT:
	res = 0x0000000f;
	break;

      default:
	res = 0;
    }

    switch (req->size) {
      case sizeof(uint8_t):
	*(uint8_t *)data = *(uint8_t *)&res;
	break;

      case sizeof(uint16_t):
	*(uint16_t *)data = *(uint16_t *)&res;
	break;

      case sizeof(uint32_t):
	*(uint32_t *)data = *(uint32_t *)&res;
	break;

      default:
	panic("invalid access size");
    }

    DPRINTF(PCIA, "read  va=%#x off=%#x data=%#x\n",
	    req->vaddr, offset,
	    *(uint32_t *)data & ((1 << (req->size * 8 - 1)) |
			       ((1 << (req->size * 8 - 1)) - 1)));

    return No_Fault;
}

Fault
TlaserPcia::write(MemReqPtr &req, const uint8_t *data)
{
    DPRINTF(PCIA, "write va=%#x off=%#x data=%#x\n",
	    req->vaddr, req->paddr - addr,
	    *(uint32_t *)data & ((1 << (req->size * 8 - 1)) |
			       ((1 << (req->size * 8 - 1)) - 1)));

    return No_Fault;
}

Tick
TlaserPcia::cacheAccess(MemReqPtr &req)
{
    return curTick + 1000;
}


BEGIN_DECLARE_SIM_OBJECT_PARAMS(TlaserPcia)

    SimObjectParam<MemoryController *> mmu;
    Param<Addr> addr;
    SimObjectParam<Bus*> io_bus;
    SimObjectParam<HierParams *> hier;

END_DECLARE_SIM_OBJECT_PARAMS(TlaserPcia)

BEGIN_INIT_SIM_OBJECT_PARAMS(TlaserPcia)

    INIT_PARAM(mmu, "Memory Controller"),
    INIT_PARAM(addr, "Device Address"),
    INIT_PARAM_DFLT(io_bus, "The IO Bus to attach to", NULL),
    INIT_PARAM_DFLT(hier, "Hierarchy global variables", &defaultHierParams)

END_INIT_SIM_OBJECT_PARAMS(TlaserPcia)

CREATE_SIM_OBJECT(TlaserPcia)
{
    return new TlaserPcia(getInstanceName(), mmu, addr, hier, io_bus);
}

REGISTER_SIM_OBJECT("TlaserPcia", TlaserPcia)
