/* $Id: s.tlaser_pcidev.cc 1.39 04/02/29 16:18:41-05:00 ehallnor@zazzer.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* @file
 * Turbolaser PCI device configuration space
 */

#include <list>
#include <sstream>
#include <string>
#include <vector>

#include "base/inifile.hh"
#include "base/misc.hh"
#include "base/str.hh"	// for to_number
#include "base/trace.hh"
#include "dev/pciareg.h"
#include "dev/scsi_ctrl.hh"
#include "dev/tlaser_pcidev.hh"
#include "mem/bus/bus.hh"
#include "mem/bus/pio_interface.hh"
#include "mem/bus/pio_interface_impl.hh"
#include "mem/functional_mem/memory_control.hh"
#include "sim/builder.hh"
#include "sim/param.hh"
#include "sim/universe.hh"

using namespace std;

TlaserPciDev::TlaserPciDev(const string &name, MemoryController *mmu, Addr a,
			   HierParams *hier, Bus *bus)
    : PioDevice(name), addr(a)
{
    mmu->add_child(this, Range<Addr>(addr, addr + size));

    if (bus) {
	pioInterface = newPioInterface(name, hier, bus, this,
				       &TlaserPciDev::cacheAccess);
	pioInterface->addAddrRange(addr, addr + size - 1);
    }

    memset(config.data, 0, sizeof(config.data));
}

uint32_t
byte_enables(uint32_t pci_ben, const uint8_t *data)
{
    union {
	uint32_t word_value;
	uint8_t byte_value[4];
    };

    word_value = 0;

    switch (pci_ben) {
      case 0x0:
	byte_value[0] = data[0];
	break;
      case 0x4:
	byte_value[1] = data[0];
	break;
      case 0x8:
	byte_value[2] = data[0];
	break;
      case 0xc:
	byte_value[3] = data[0];
	break;
      case 0x1:
	byte_value[0] = data[0];
	byte_value[1] = data[1];
	break;
      case 0x5:
	byte_value[1] = data[0];
	byte_value[2] = data[1];
	break;
      case 0x9:
	byte_value[2] = data[0];
	byte_value[3] = data[1];
	break;
      case 0x2:
	byte_value[0] = data[0];
	byte_value[1] = data[1];
	byte_value[2] = data[2];
	break;
      case 0x6:
	byte_value[1] = data[0];
	byte_value[2] = data[1];
	byte_value[3] = data[2];
	break;
      case 0x3:
	byte_value[0] = data[0];
	byte_value[1] = data[1];
	byte_value[2] = data[2];
	byte_value[3] = data[3];
	break;
      default:
	panic("invalid byte enables");
    }

    return word_value;
}

Fault
TlaserPciDev::read(MemReqPtr &req, uint8_t *data)
{
    Addr base = req->paddr - addr; // clear top bits
    uint64_t pci_addr = (base >> PCICONF_REG_SHIFT);
    uint64_t offset  = pci_addr & 0xff;
#if TRACING_ON
    //uint64_t width = (base >> WIDTH_SHIFT) & 0x3;
    uint64_t pci_dev = (base >> PCICONF_DEV_SHIFT) & PCICONF_DEV_MASK;
    uint64_t pci_fun = (base >> PCICONF_FUN_SHIFT) & PCICONF_FUN_MASK;
    uint64_t pci_reg = (base >> PCICONF_REG_SHIFT) & PCICONF_REG_MASK;
#endif
    uint64_t pci_ben = (base >> PCICONF_BEN_SHIFT) & PCICONF_BEN_MASK;

    if (req->size < (pci_ben & 0x3) + 1)
	panic("request size is smaller that indicated by PCI");

    union {
	uint8_t byte_value;
	uint16_t half_value;
	uint32_t word_value;
    };

    word_value = byte_enables(pci_ben, config.data + offset);

    switch(req->size) {
      case sizeof(uint8_t):
	*(uint8_t *)data = byte_value;
	break;

      case sizeof(uint16_t):
	*(uint16_t *)data = half_value;
	break;

      case sizeof(uint32_t):
	*(uint32_t *)data = word_value;
	break;

      default:
	panic("invalid access size");
    }

    DPRINTF(PCIDEV, "read  va=%#x dev=%x fun=%x reg=%x ben=%x data=%x\n",
	    req->vaddr, pci_dev, pci_fun, pci_reg, pci_ben,
	    *(uint32_t *)data & ((1 << (req->size * 8 - 1)) |
			       ((1 << (req->size * 8 - 1)) - 1)));

    return No_Fault;
}

Fault
TlaserPciDev::write(MemReqPtr &req, const uint8_t *data)
{
    Addr base = req->paddr - addr;
    uint64_t pci_addr = (base >> PCICONF_REG_SHIFT);
    uint64_t offset  = pci_addr & 0xff;
#if TRACING_ON
    //uint64_t width = (base >> WIDTH_SHIFT) & 0x3;
    uint64_t pci_dev = (base >> PCICONF_DEV_SHIFT) & PCICONF_DEV_MASK;
    uint64_t pci_fun = (base >> PCICONF_FUN_SHIFT) & PCICONF_FUN_MASK;
    uint64_t pci_reg = (base >> PCICONF_REG_SHIFT) & PCICONF_REG_MASK;
#endif
    uint64_t pci_ben = (base >> PCICONF_BEN_SHIFT) & PCICONF_BEN_MASK;

    DPRINTF(PCIDEV, "write va=%#x dev=%x fun=%x reg=%x ben=%x data=%x\n",
	    req->vaddr, pci_dev, pci_fun, pci_reg, pci_ben,
	    *(uint32_t *)data & ((1 << (req->size * 8 - 1)) |
			       ((1 << (req->size * 8 - 1)) - 1)));

    if (req->size < (pci_ben & 0x3) + 1)
	panic("request size is smaller that indicated by PCI");

    union {
	uint8_t byte_value;
	uint16_t half_value;
	uint32_t word_value;
    };

    word_value = byte_enables(pci_ben, data);

    switch (pci_ben & 0x3) {
      case 0: // 1-byte access
	switch (offset) {
	  case PCI0_INTERRUPT_LINE:
	    *(uint8_t *)&config.data[offset] = byte_value;
	    break;

	  default:
	    panic("writing to a read only register");
	}
	break;

      case 1: // 2-byte access
	switch (offset) {
	  case PCI_COMMAND:
	  case PCI_STATUS:
	    *(uint16_t *)&config.data[offset] = half_value;
	    break;

	  default:
	    panic("writing to a read only register");
	}
	break;

      case 2: // 3-byte access
	panic("invalid access size");

      case 3: // 4-byte access
	switch (offset) {
	  case PCI0_BASE_ADDR0:
	  case PCI0_BASE_ADDR1:
	  case PCI0_BASE_ADDR2:
	  case PCI0_BASE_ADDR3:
	  case PCI0_BASE_ADDR4:
	  case PCI0_BASE_ADDR5:
	  case PCI0_ROM_BASE_ADDR:
	    *(uint32_t *)&config.data[offset] = word_value;
	    break;

	  default:
	    panic("writing to a read only register");
	}
	break;
    }

    return No_Fault;
}

Tick
TlaserPciDev::cacheAccess(MemReqPtr &req)
{
    return curTick + 1000;
}

void
TlaserPciDev::serialize(ostream &os)
{
    SERIALIZE_ARRAY(config.data, 64);
}

void
TlaserPciDev::unserialize(Checkpoint *cp, const std::string &section)
{
    UNSERIALIZE_ARRAY(config.data, 64);
}


BEGIN_DECLARE_SIM_OBJECT_PARAMS(TlaserPciDev)

    Param<int> VendorID;
    Param<int> DeviceID;
    Param<int> Command;
    Param<int> Status;
    Param<int> Revision;
    Param<int> ProgIF;
    Param<int> SubClassCode;
    Param<int> ClassCode;
    Param<int> CacheLineSize;
    Param<int> LatencyTimer;
    Param<int> HeaderType;
    Param<int> BIST;
    Param<uint32_t> BAR0;
    Param<uint32_t> BAR1;
    Param<uint32_t> BAR2;
    Param<uint32_t> BAR3;
    Param<uint32_t> BAR4;
    Param<uint32_t> BAR5;
    Param<uint32_t> CardbusCIS;
    Param<int> SubsystemVendorID;
    Param<int> SubsystemID;
    Param<uint32_t> ExpansionROM;
    Param<int> InterruptLine;
    Param<int> InterruptPin;
    Param<int> MinimumGrant;
    Param<int> MaximumLatency;

    SimObjectParam<MemoryController *> mmu;
    Param<Addr> addr;
    SimObjectParam<Bus*> io_bus;
    SimObjectParam<HierParams *> hier;

END_DECLARE_SIM_OBJECT_PARAMS(TlaserPciDev)

BEGIN_INIT_SIM_OBJECT_PARAMS(TlaserPciDev)

    INIT_PARAM(VendorID, "Vendor ID"),
    INIT_PARAM(DeviceID, "Device ID"),
    INIT_PARAM_DFLT(Command, "Command Register", 0x00),
    INIT_PARAM_DFLT(Status, "Status Register", 0x00),
    INIT_PARAM_DFLT(Revision, "Device Revision", 0x00),
    INIT_PARAM_DFLT(ProgIF, "Programming Interface", 0x00),
    INIT_PARAM(SubClassCode, "Sub-Class Code"),
    INIT_PARAM(ClassCode, "Class Code"),
    INIT_PARAM_DFLT(CacheLineSize, "System Cacheline Size", 0x00),
    INIT_PARAM_DFLT(LatencyTimer, "PCI Latency Timer", 0x00),
    INIT_PARAM_DFLT(HeaderType, "PCI Header Type", 0x00),
    INIT_PARAM_DFLT(BIST, "Built In Self Test", 0x00),
    INIT_PARAM_DFLT(BAR0, "Base Address Register 0", 0x00),
    INIT_PARAM_DFLT(BAR1, "Base Address Register 1", 0x00),
    INIT_PARAM_DFLT(BAR2, "Base Address Register 2", 0x00),
    INIT_PARAM_DFLT(BAR3, "Base Address Register 3", 0x00),
    INIT_PARAM_DFLT(BAR4, "Base Address Register 4", 0x00),
    INIT_PARAM_DFLT(BAR5, "Base Address Register 5", 0x00),
    INIT_PARAM_DFLT(CardbusCIS, "Cardbus Card Information Structure", 0x00),
    INIT_PARAM_DFLT(SubsystemVendorID, "Subsystem Vendor ID", 0x00),
    INIT_PARAM_DFLT(SubsystemID, "Subsystem ID", 0x00),
    INIT_PARAM_DFLT(ExpansionROM, "Expansion ROM Base Address Register", 0x00),
    INIT_PARAM(InterruptLine, "Interrupt Line Register"),
    INIT_PARAM(InterruptPin, "Interrupt Pin Register"),
    INIT_PARAM_DFLT(MinimumGrant, "Minimum Grant", 0x00),
    INIT_PARAM_DFLT(MaximumLatency, "Maximum Latency", 0x00),

    INIT_PARAM(mmu, "Memory Controller"),
    INIT_PARAM(addr, "Device Address"),
    INIT_PARAM_DFLT(io_bus, "The IO Bus to attach to", NULL),
    INIT_PARAM_DFLT(hier, "Hierarchy global variables", &defaultHierParams)

END_INIT_SIM_OBJECT_PARAMS(TlaserPciDev)

CREATE_SIM_OBJECT(TlaserPciDev)
{
    TlaserPciDev *dev = new TlaserPciDev(getInstanceName(), mmu, addr,
					 hier, io_bus);

    dev->config.hdr.vendor = VendorID;
    dev->config.hdr.device = DeviceID;
    dev->config.hdr.command = Command;
    dev->config.hdr.status = Status;
    dev->config.hdr.revision = Revision;
    dev->config.hdr.progIF = ProgIF;
    dev->config.hdr.subClassCode = SubClassCode;
    dev->config.hdr.classCode = ClassCode;
    dev->config.hdr.cacheLineSize = CacheLineSize;
    dev->config.hdr.latencyTimer = LatencyTimer;
    dev->config.hdr.headerType = HeaderType;
    dev->config.hdr.bist = BIST;

    dev->config.hdr.pci0.baseAddr0 = BAR0;
    dev->config.hdr.pci0.baseAddr1 = BAR1;
    dev->config.hdr.pci0.baseAddr2 = BAR2;
    dev->config.hdr.pci0.baseAddr3 = BAR3;
    dev->config.hdr.pci0.baseAddr4 = BAR4;
    dev->config.hdr.pci0.baseAddr5 = BAR5;
    dev->config.hdr.pci0.cardbusCIS = CardbusCIS;
    dev->config.hdr.pci0.subsystemVendorID = SubsystemVendorID;
    dev->config.hdr.pci0.subsystemID = SubsystemVendorID;
    dev->config.hdr.pci0.expansionROM = ExpansionROM;
    dev->config.hdr.pci0.interruptLine = InterruptLine;
    dev->config.hdr.pci0.interruptPin = InterruptPin;
    dev->config.hdr.pci0.minimumGrant = MinimumGrant;
    dev->config.hdr.pci0.maximumLatency = MaximumLatency;

    return dev;
}

REGISTER_SIM_OBJECT("TlaserPciDev", TlaserPciDev)
