/* $Id: s.tlaser_serial.cc 1.32 04/02/29 16:18:41-05:00 ehallnor@zazzer.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* @file
 * Turbolaser serial number
 */

#include <string>
#include <vector>

#include "base/inifile.hh"
#include "base/misc.hh"
#include "base/str.hh"	// for to_number
#include "base/trace.hh"
#include "dev/tlaser_serial.hh"
#include "mem/bus/bus.hh"
#include "mem/bus/pio_interface.hh"
#include "mem/bus/pio_interface_impl.hh"
#include "mem/functional_mem/memory_control.hh"
#include "sim/builder.hh"

using namespace std;

TlaserSerial::TlaserSerial(const string &name, uint32_t sernum,
			   MemoryController *mmu, Addr a,
			   HierParams *hier, Bus *bus)
    : PioDevice(name), addr(a), serialNumber(sernum)
{
    mmu->add_child(this, Range<Addr>(addr, addr + size));

    if (bus) {
	pioInterface = newPioInterface(name, hier, bus, this,
				       &TlaserSerial::cacheAccess);
	pioInterface->addAddrRange(addr, addr + size - 1);
    }
}

Fault
TlaserSerial::read(MemReqPtr &req, uint8_t *data)
{
    Addr daddr = req->paddr - addr;

    if (daddr != 0)
	panic("%s: invalid read addr (%#x)\n", name(), daddr);

    switch (req->size) {
      case sizeof(uint32_t):
	*(uint32_t *)data = serialNumber;
	break;
      case sizeof(uint64_t):
	*(uint64_t *)data = serialNumber;
	break;
      default:
	panic("%s: invalid read size (%d)\n", name(), req->size);
    }

    return No_Fault;
}

Fault
TlaserSerial::write(MemReqPtr &req, const uint8_t *data)
{
    Addr daddr = req->paddr - addr;

    if (daddr != 0)
	panic("%s: invalid write addr (%#x)\n", name(), daddr);

    if (req->size != sizeof(uint32_t) &&
	req->size != sizeof(uint64_t))
	panic("%s: invalid write size (%d)\n", name(), req->size);

    serialNumber = *(uint32_t *)data;

    return No_Fault;
}

Tick
TlaserSerial::cacheAccess(MemReqPtr &req)
{
    return curTick + 1000;
}

void
TlaserSerial::serialize(ostream &os)
{
    SERIALIZE_SCALAR(serialNumber);
}

void
TlaserSerial::unserialize(Checkpoint *cp, const std::string &section)
{
    UNSERIALIZE_SCALAR(serialNumber);
}


#define SERIAL_NUMBER 0xFAFAFAFA

BEGIN_DECLARE_SIM_OBJECT_PARAMS(TlaserSerial)

    Param<uint32_t> sernum;
    SimObjectParam<MemoryController *> mmu;
    Param<Addr> addr;
    SimObjectParam<Bus*> io_bus;
    SimObjectParam<HierParams *> hier;

END_DECLARE_SIM_OBJECT_PARAMS(TlaserSerial)

BEGIN_INIT_SIM_OBJECT_PARAMS(TlaserSerial)

    INIT_PARAM_DFLT(sernum, "Serial Number", SERIAL_NUMBER),
    INIT_PARAM(mmu, "Memory Controller"),
    INIT_PARAM(addr, "Device Address"),
    INIT_PARAM_DFLT(io_bus, "The IO Bus to attach to", NULL),
    INIT_PARAM_DFLT(hier, "Hierarchy global variables", &defaultHierParams)

END_INIT_SIM_OBJECT_PARAMS(TlaserSerial)


CREATE_SIM_OBJECT(TlaserSerial)
{
    return new TlaserSerial(getInstanceName(), sernum, mmu, addr, hier,
			    io_bus);
}

REGISTER_SIM_OBJECT("TlaserSerial", TlaserSerial)
