/* $Id: s.tlaser_uart.cc 1.47 04/02/29 16:18:42-05:00 ehallnor@zazzer.eecs.umich.edu $ */

/* @file
 * Turbolaser UART
 */

/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */

#include <string>
#include <vector>

#include "base/inifile.hh"
#include "base/str.hh"	// for to_number
#include "base/trace.hh"
#include "dev/console.hh"
#include "dev/tlaser_uart.hh"
#include "mem/bus/bus.hh"
#include "mem/bus/pio_interface.hh"
#include "mem/bus/pio_interface_impl.hh"
#include "mem/functional_mem/memory_control.hh"
#include "sim/builder.hh"
#include "targetarch/ev5.hh"

using namespace std;

#define CONS_INT_TX   0x01  // interrupt enable / state bits
#define CONS_INT_RX   0x02

TlaserUart::TlaserUart(const string &name, SimConsole *c,
		       MemoryController *mmu, Addr a,
		       HierParams *hier, Bus *bus)
    : PioDevice(name), addr(a), cons(c), status_store(0), valid_char(false)

{
    mmu->add_child(this, Range<Addr>(addr, addr + size));

    if (bus) {
	pioInterface = newPioInterface(name, hier, bus, this,
				       &TlaserUart::cacheAccess);
	pioInterface->addAddrRange(addr, addr + size - 1);
    }
}

Fault
TlaserUart::read(MemReqPtr &req, uint8_t *data)
{
    Addr daddr = req->paddr - addr;

    switch (req->size) {
      case sizeof(uint64_t):
	*(uint64_t *)data = 0;
	break;
      case sizeof(uint32_t):
	*(uint32_t *)data = 0;
	break;
      case sizeof(uint16_t):
	*(uint16_t *)data = 0;
	break;
      case sizeof(uint8_t):
	*(uint8_t *)data = 0;
	break;
    }

    switch (daddr) {
      case 0x80: // Status Register
	{
	    int status = cons->intStatus();
	    if (!valid_char) {
		valid_char = cons->in(next_char);
		if (!valid_char)
		    status &= ~CONS_INT_RX;
	    } else {
		status |= CONS_INT_RX;
	    }

	    if (status_store == 3) {
		// RR3 stuff? Don't really understand it, btw
		status_store = 0;
		if (status & CONS_INT_TX) {
		    *data = (1 << 4);
		    return No_Fault;
		} else if (status & CONS_INT_RX) {
		    *data = (1 << 5);
		    return No_Fault;
		} else {
		    DPRINTF(TlaserUart, "spurious read\n");
		    return No_Fault;
		}
	    } else {
		int reg = (1 << 2);
		if (status & CONS_INT_RX)
		    reg |= (1 << 0);
		*data = reg;
		return No_Fault;
	    }
	    break;
	}

      case 0xc0: // Data register (RX)
	if (!valid_char)
	    panic("Invalid character");

	DPRINTF(TlaserUart, "read data register \'%c\' %#02x\n",
		isprint(next_char) ? next_char : ' ', next_char);

	*data = next_char;
	valid_char = false;

	return No_Fault;
    }

    panic("%s: read daddr=%#x type=read *data=%#x\n", name(), daddr, *data);

    return No_Fault;
}

Fault
TlaserUart::write(MemReqPtr &req, const uint8_t *data)
{
    Addr daddr = req->paddr - addr;

    switch (daddr) {
      case 0x80:
	status_store = *data;
	switch (*data) {
	  case 0x03: // going to read RR3
	    return No_Fault;

	  case 0x28: // Ack of TX
	    {
		if ((cons->intStatus() & CONS_INT_TX) == 0)
		    panic("Ack of transmit, though there was no interrupt");

		cons->clearInt(CONS_INT_TX);
		return No_Fault;
	    }

	  case 0x00:
	  case 0x01:
	  case 0x12:
	    // going to write data???
	    return No_Fault;

	  default:
	    DPRINTF(TlaserUart, "writing status register %#x \n",
		    *(uint64_t *)data);
	    return No_Fault;
	}

      case 0xc0: // Data register (TX)
	cons->out(*(uint64_t *)data);
	return No_Fault;
    }

    return No_Fault;
}

Tick
TlaserUart::cacheAccess(MemReqPtr &req)
{
    return curTick + 1000;
}

void
TlaserUart::serialize(ostream &os)
{
    SERIALIZE_SCALAR(status_store);
    SERIALIZE_SCALAR(next_char);
    SERIALIZE_SCALAR(valid_char);
}

void
TlaserUart::unserialize(Checkpoint *cp, const std::string &section)
{
    UNSERIALIZE_SCALAR(status_store);
    UNSERIALIZE_SCALAR(next_char);
    UNSERIALIZE_SCALAR(valid_char);
}

BEGIN_DECLARE_SIM_OBJECT_PARAMS(TlaserUart)

    SimObjectParam<SimConsole *> console;
    SimObjectParam<MemoryController *> mmu;
    Param<Addr> addr;
    SimObjectParam<Bus*> io_bus;
    SimObjectParam<HierParams *> hier;

END_DECLARE_SIM_OBJECT_PARAMS(TlaserUart)

BEGIN_INIT_SIM_OBJECT_PARAMS(TlaserUart)

    INIT_PARAM(console, "The console"),
    INIT_PARAM(mmu, "Memory Controller"),
    INIT_PARAM(addr, "Device Address"),
    INIT_PARAM_DFLT(io_bus, "The IO Bus to attach to", NULL),
    INIT_PARAM_DFLT(hier, "Hierarchy global variables", &defaultHierParams)

END_INIT_SIM_OBJECT_PARAMS(TlaserUart)

CREATE_SIM_OBJECT(TlaserUart)
{
    return new TlaserUart(getInstanceName(), console, mmu, addr, hier, io_bus);
}

REGISTER_SIM_OBJECT("TlaserUart", TlaserUart)
