/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/** @file
 * Definitions of a unidirectional bus bridge.
 */

#include "mem/bus/bus_bridge.hh"

#include "mem/bus/bus_bridge_slave.hh"
#include "mem/bus/bus_bridge_master.hh"
#include "mem/bus/bus.hh"
#include "sim/builder.hh"

using namespace std;

BusBridge::BusBridge(const string &name, BusBridge::Params params)
    : BaseHier(name, params.hier),  latency(params.latency), 
      max(params.max), inWidth(params.inWidth), inRatio(params.inRatio), 
      outWidth(params.outWidth), outRatio(params.outRatio)
{
    slaveCount = 0;
    slaveBlocked = false;
    masterCount = 0;
    masterBlocked = false;
}

MemAccessResult 
BusBridge::slaveAccess(MemReqPtr &req) {
    if (++slaveCount == max) {
	slaveBlocked = true;
	si->setBlocked();
    }
    slaveBuffer.push_back(req);
    Tick ready = curTick + latency;
    if (req->cmd.isWrite()) {
	// need to wait for the data to arrive.
	ready += ((req->size > inWidth) ? (req->size - inWidth)/inWidth : 0)
	    * inRatio;
    }
    slaveTimes.push_back(ready);
    mi->request(ready);
    requests[(intptr_t)req.get()] = req->busId;
    /** @todo Should change this from cache miss to something else. */
    return MA_CACHE_MISS;
}

MemAccessResult 
BusBridge::masterAccess(MemReqPtr &req) {
    if (++masterCount == max) {
	masterBlocked = true;
	mi->setBlocked();
    }
    masterBuffer.push_back(req);
    Tick ready = curTick + latency;
    if (req->cmd.isWrite()) {
	// need to wait for the data to arrive.
	ready += ((req->size > inWidth) ? (req->size - inWidth)/inWidth : 0)
	    * inRatio;
    }
    masterTimes.push_back(ready);
    si->request(ready);
    requests[(intptr_t)req.get()] = req->busId;
    /** @todo Should change this from cache miss to something else. */
    return MA_CACHE_MISS;
}

void 
BusBridge::sendSlaveResult(MemReqPtr &req, bool success)
{
    if (success) {
	assert(req == masterBuffer.front());
	if (masterTimes.front() >curTick) {
	    warn("BusBridge:: Forwarding request %d cycles before its ready",
		 masterTimes.front() - curTick);
	}
	masterBuffer.pop_front();
	masterTimes.pop_front();
	if (masterCount-- == max) {
	    masterBlocked = false;
	    mi->clearBlocked();
	}
	if (!masterTimes.empty() && masterTimes.front() > curTick) {
	    // Need to rerequest the bus.
	    si->request(masterTimes.front());
	}
    }
}

void 
BusBridge::sendMasterResult(MemReqPtr &req, bool success)
{
    if (success) {
	assert(req == slaveBuffer.front());
	if (slaveTimes.front() >curTick) {
	    warn("BusBridge:: Forwarding request %d cycles before its ready",
		 slaveTimes.front() - curTick);
	}
	slaveBuffer.pop_front();
	slaveTimes.pop_front();
	if (slaveCount-- == max) {
	    slaveBlocked = false;
	    si->clearBlocked();
	}
	if (!slaveTimes.empty() && slaveTimes.front() > curTick) {
	    // Need to rerequest the bus.
	    mi->request(slaveTimes.front());
	}
    }
}

void
BusBridge::handleSlaveResponse(MemReqPtr &req)
{
    intptr_t key = (intptr_t)req.get();
    assert(requests.count(key) == 1);
    req->busId = requests[key];
    requests.erase(key);
    Tick ready = curTick + latency;
    if (req->cmd.isRead()) {
	// need to wait for the data to arrive.
	ready += ((req->size > outWidth) ? (req->size - outWidth)/outWidth : 0)
	    * outRatio;
    }
    mi->respond(req, ready);
}

void
BusBridge::handleMasterResponse(MemReqPtr &req)
{
    intptr_t key = (intptr_t)req.get();
    assert(requests.count(key) == 1);
    req->busId = requests[key];
    requests.erase(key);
    Tick ready = curTick + latency;
    if (req->cmd.isRead()) {
	// need to wait for the data to arrive.
	ready += ((req->size > outWidth) ? (req->size - outWidth)/outWidth : 0)
	    * outRatio;
    }
    si->respond(req, ready);
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS

BEGIN_DECLARE_SIM_OBJECT_PARAMS(BusBridge)

    SimObjectParam<Bus*> in_bus;
    SimObjectParam<Bus*> out_bus;
    Param<int> max_buffer;
    Param<int> latency;
    SimObjectParam<HierParams *> hier;

END_DECLARE_SIM_OBJECT_PARAMS(BusBridge)


BEGIN_INIT_SIM_OBJECT_PARAMS(BusBridge)

    INIT_PARAM(in_bus, "The bus to forward from"),
    INIT_PARAM(out_bus, "The bus to forward to"),
    INIT_PARAM_DFLT(max_buffer, "The number of requests to buffer", 8),
    INIT_PARAM_DFLT(latency, "The latency of this bridge", 0),
    INIT_PARAM_DFLT(hier, "Hierarchy global variables", &defaultHierParams)

END_INIT_SIM_OBJECT_PARAMS(BusBridge)


CREATE_SIM_OBJECT(BusBridge)
{
    BusBridge::Params params;
    params.max = max_buffer;
    params.latency = latency;
    params.hier = hier;
    params.inWidth = in_bus->width;
    params.inRatio = in_bus->clockRatio;
    params.outWidth = out_bus->width;
    params.outRatio = out_bus->clockRatio;

    string name = getInstanceName();
    
    BusBridge *retval = new BusBridge(name, params);
    
    retval->setInterfaces(new BusBridgeSlave(name, hier, in_bus, retval),
			  new BusBridgeMaster(name, hier, out_bus, retval));
    return retval;
}

REGISTER_SIM_OBJECT("BusBridge", BusBridge)

#endif //DOXYGEN_SHOULD_SKIP_THIS
