/* $Id: s.master_interface_impl.hh 1.24 04/04/15 23:47:15-04:00 ehallnor@zizzer.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/**
 * @file
 * Definitions for the bus master interface.
 */
#include <algorithm>
#include <string>

#include "base/intmath.hh"
#include "mem/bus/master_interface.hh"

using namespace std;

template<class Mem, class Bus>
MasterInterface<Mem, Bus>::MasterInterface(const string &name,
					   HierParams *hier, Mem *_mem,
					   Bus *bus)
    : BusInterface<Bus>(name, hier, bus, true), mem(_mem)
{
}

template<class Mem, class Bus>
bool
MasterInterface<Mem, Bus>::grantAddr()
{
    MemReqPtr req = mem->getMemReq();
    Tick req_time = curTick;
    if (req) {
	req->busId = id;
	req_time = req->time;
	if (req->cmd.isWrite()) {
	    responseQueue.push_back(DataResponseEntry(req->size, curTick));
	    bus->requestDataBus(id, curTick);
	}
    }
    bool successful = bus->sendAddr(req, req_time);
    mem->sendResult(req, successful);
    return mem->doMasterRequest();
}

template<class Mem, class Bus>
void
MasterInterface<Mem, Bus>::deliver(MemReqPtr &req)
{
    mem->handleResponse(req);
}


template<class Mem, class Bus>
MemAccessResult
MasterInterface<Mem, Bus>::access(MemReqPtr &req)
{
    int satisfied_before = req->flags & SATISFIED;
    // Cache Coherence call goes here
    mem->snoop(req);
    if (satisfied_before != (req->flags & SATISFIED)) {
	return BA_SUCCESS;
    }
    return BA_NO_RESULT;
}


/*
 * Called when granted the data bus, forwards a response to the Bus
 */
template<class Mem, class Bus>
bool
MasterInterface<Mem, Bus>::grantData()
{	
    typename BusInterface<Bus>::DataResponseEntry entry = 
	responseQueue.front();

    if (entry.size > 0) {
	bus->delayData(entry.size);
    } else {
	if (entry.req->cmd.isWrite()){
	    bus->sendAck(entry.req, entry.time);
	} else {
	    bus->sendData(entry.req, entry.time);
	}
    }
    responseQueue.pop_front();
    if (!responseQueue.empty()) {
	bus->requestDataBus(id, max(curTick, responseQueue.front().time));
    }
    
    return false;
}

/*
 * Called by attached TimingMemObj when a request is satisfied
 */
template<class Mem, class Bus>
void
MasterInterface<Mem, Bus>::respond(MemReqPtr &req, Tick time)
{
    assert(req->cmd == Read || req->cmd == ReadEx);
    responseQueue.push_back(DataResponseEntry(req,time));
    bus->requestDataBus(id, time);
}
