/* $Id: s.slave_interface_impl.hh 1.27 04/06/08 17:31:02-04:00 ehallnor@zizzer.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/**
 * @file
 * Definition of a templatize slave interface.
 */

#include <algorithm>
#include <string>

#include "base/intmath.hh"
#include "mem/bus/slave_interface.hh"
#include "mem/trace/mem_trace_writer.hh"

/** @todo remove stats include once nate fixes bug */
#include "base/statistics.hh"

using namespace std;

template <class Mem, class Bus>
SlaveInterface<Mem, Bus>::SlaveInterface(const string &name,
					 HierParams *hier, Mem *_mem,
					 Bus *bus, MemTraceWriter *mem_trace)
    : BusInterface<Bus>(name, hier, bus, false), mem(_mem), memTrace(mem_trace)
{
}

/*
 * Calls access on the associated TimingMemObj
 */
template <class Mem, class Bus>
MemAccessResult
SlaveInterface<Mem, Bus>::access(MemReqPtr &req)
{
    int retval;
    bool already_satisfied = req->isSatisfied();
    if (inRange(req->paddr)) {
	if (memTrace) {
	    memTrace->writeReq(req);
	}

        if (isBlocked()) {
	    return BA_BLOCKED;
	}

	retval = mem->access(req);

	assert(!isBlocked() || mem->isBlocked());

	if (already_satisfied) {
	    // a backside bus interface supplying the data
	    return BA_NO_RESULT;
	}
	return BA_SUCCESS;  // This transaction went through ok
    }
    return BA_NO_RESULT;
}

/*
 * Called when granted the data bus, forwards a response to the Bus
 */
template <class Mem, class Bus>
bool
SlaveInterface<Mem, Bus>::grantData()
{	
    typename BusInterface<Bus>::DataResponseEntry entry = 
	responseQueue.front();

    if (entry.size > 0) {
	bus->delayData(entry.size);
    } else {
	if (entry.req->cmd.isWrite()){
	    bus->sendAck(entry.req, entry.time);
	} else {
	    bus->sendData(entry.req, entry.time);
	}
    }
    responseQueue.pop_front();
    if (!responseQueue.empty()) {
	bus->requestDataBus(id, max(curTick, responseQueue.front().time));
    }
    
    return false;
}

/*
 * Called by attached TimingMemObj when a request is satisfied
 */
template <class Mem, class Bus>
void
SlaveInterface<Mem, Bus>::respond(MemReqPtr &req, Tick time)
{
    if (!req->cmd.isNoResponse()) {
	responseQueue.push_back(BusInterface<Bus>::DataResponseEntry(req, 
								     time));
	bus->requestDataBus(id, time);
    }
}

template <class Mem, class Bus>
bool
SlaveInterface<Mem, Bus>::grantAddr()
{
    // get coherence message here
    MemReqPtr req = mem->getCoherenceReq();
    if (req) {
	req->busId = id;
    }
    bus->sendAddr(req,req->time);
    return mem->doSlaveRequest();
}
