/* $Id: s.cache_builder.cc 1.35 04/06/08 17:31:02-04:00 ehallnor@zizzer.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/**
 * @file
 * Simobject instatiation of caches.
 */
#include <vector>

#include "mem/cache/base_cache.hh"
#include "mem/cache/cache.hh"
#include "mem/bus/bus.hh"
#include "mem/cache/coherence/coherence_protocol.hh"
#include "sim/builder.hh"

// Tag Templates
#include "mem/cache/tags/lru.hh"
#include "mem/cache/tags/fa_lru.hh"
#include "mem/cache/tags/iic.hh"

// Compression Templates
#include "base/compression/null_compression.hh"
#include "base/compression/lzss_compression.hh"

// CacheTags Templates
#include "mem/cache/tags/cache_tags.hh"

// MissQueue Templates
#include "mem/cache/miss/miss_queue.hh"
#include "mem/cache/miss/blocking_buffer.hh"

// Coherence Templates
#include "mem/cache/coherence/uni_coherence.hh"
#include "mem/cache/coherence/simple_coherence.hh"

// Bus Interfaces
#include "mem/bus/slave_interface.hh"
#include "mem/bus/master_interface.hh"
#include "mem/memory_interface.hh"

#include "mem/trace/mem_trace_writer.hh"

using namespace std;

#ifndef DOXYGEN_SHOULD_SKIP_THIS

BEGIN_DECLARE_SIM_OBJECT_PARAMS(BaseCache)

    Param<int> size;
    Param<int> assoc;
    Param<int> block_size;
    Param<int> hit_latency;
    Param<int> mshrs;
    Param<int> tgts_per_mshr;
    Param<int> write_buffers;
    Param<bool> prioritizeRequests;
    SimObjectParam<Bus *> in_bus;
    SimObjectParam<Bus *> out_bus;
    Param<bool> do_copy;
    SimObjectParam<CoherenceProtocol *> protocol;
    Param<Addr> trace_addr;
    Param<int> hash_delay;
    SimObjectParam<Repl *> repl;
    Param<bool> compressed_bus;
    Param<bool> store_compressed;
    Param<bool> adaptive_compression;
    Param<int> compression_latency;
    Param<int> subblock_size;
    Param<Counter> max_miss_count;
    SimObjectParam<HierParams *> hier;
    VectorParam<Addr> start_addr;
    VectorParam<Addr> addr_range;
    SimObjectParam<MemTraceWriter *> mem_trace;

END_DECLARE_SIM_OBJECT_PARAMS(BaseCache)


BEGIN_INIT_SIM_OBJECT_PARAMS(BaseCache)

    INIT_PARAM(size, "capacity in bytes"),
    INIT_PARAM(assoc, "associativity"),
    INIT_PARAM(block_size, "block size in bytes"),
    INIT_PARAM(hit_latency, "hit latency in CPU cycles"),
    INIT_PARAM(mshrs, "number of MSHRs (max outstanding requests)"),
    INIT_PARAM(tgts_per_mshr, "max number of accesses per MSHR"),
    INIT_PARAM_DFLT(write_buffers, "number of write buffers", 8),
    INIT_PARAM_DFLT(prioritizeRequests, "always service demand misses first",
		    false),
    INIT_PARAM_DFLT(in_bus, "incoming bus object", NULL),
    INIT_PARAM(out_bus, "outgoing bus object"),
    INIT_PARAM_DFLT(do_copy, "perform fast copies in the cache", false),
    INIT_PARAM_DFLT(protocol, "coherence protocol to use in the cache", NULL),
    INIT_PARAM_DFLT(trace_addr, "address to trace", 0),

    INIT_PARAM_DFLT(hash_delay, "time in cycles of hash access",1),
    INIT_PARAM_DFLT(repl, "replacement policy",NULL),
    INIT_PARAM_DFLT(compressed_bus,
                    "This cache connects to a compressed memory",
                    false),
    INIT_PARAM_DFLT(store_compressed, "Store compressed data in the cache",
		    false),
    INIT_PARAM_DFLT(adaptive_compression, "Use an adaptive compression scheme",
		    false),
    INIT_PARAM_DFLT(compression_latency,
		    "Latency in cycles of compression algorithm",
		    0),
    INIT_PARAM_DFLT(subblock_size,
		    "Size of subblock in IIC used for compression",
		    0),
    INIT_PARAM_DFLT(max_miss_count,
		    "The number of misses to handle before calling exit",
                    0),
    INIT_PARAM_DFLT(hier,
		    "Hierarchy global variables",
		    &defaultHierParams),
    INIT_PARAM_DFLT(start_addr, "The first address this bridge forwards", 
		    vector<Addr>(1,0)),
    INIT_PARAM_DFLT(addr_range, "The address range in bytes", 
		    vector<Addr>(1,MaxAddr)),
    INIT_PARAM_DFLT(mem_trace, "Memory trace to write accesses to", NULL)
END_INIT_SIM_OBJECT_PARAMS(BaseCache)


#define BUILD_CACHE(t, comp, b, c) {\
    Cache<CacheTags<t, comp>, b, c>::Params params(tagStore, mq, coh, do_copy, base_params, in_bus, out_bus); \
    Cache<CacheTags<t, comp>, b, c> *retval = new Cache<CacheTags<t, comp>, b, c>(getInstanceName(), hier, params); \
    if (in_bus == NULL) { \
        retval->setSlaveInterface(new MemoryInterface<Cache<CacheTags<t, comp>, b, c> >(getInstanceName(), hier, retval, mem_trace)); \
    } else { \
        retval->setSlaveInterface(new SlaveInterface<Cache<CacheTags<t, comp>, b, c>, Bus>(getInstanceName(), hier, retval, in_bus, mem_trace)); \
    } \
    retval->setMasterInterface(new MasterInterface<Cache<CacheTags<t, comp>, b, c>, Bus>(getInstanceName(), hier, retval, out_bus)); \
    return retval; \
}

#define BUILD_FALRU_CACHE(b,c) { \
    FALRU *tags = new FALRU(block_size, size, hit_latency); \
    if (compressed_bus || store_compressed){ \
	CacheTags<FALRU, LZSSCompression> *tagStore = new CacheTags<FALRU, LZSSCompression>(tags, compression_latency, true, store_compressed, adaptive_compression); \
        BUILD_CACHE(FALRU, LZSSCompression, b, c); \
    } else { \
	CacheTags<FALRU, NullCompression> *tagStore = new CacheTags<FALRU, NullCompression>(tags, compression_latency, true, store_compressed, adaptive_compression); \
        BUILD_CACHE(FALRU, NullCompression, b, c); \
    } \
}

#define BUILD_LRU_CACHE(b, c) { \
    LRU *tags = new LRU(nsets, block_size, assoc, hit_latency); \
    if (compressed_bus || store_compressed){ \
	CacheTags<LRU, LZSSCompression> *tagStore = new CacheTags<LRU,LZSSCompression>(tags, compression_latency, true, store_compressed, adaptive_compression); \
        BUILD_CACHE(LRU, LZSSCompression, b, c); \
    } else { \
	CacheTags<LRU, NullCompression> *tagStore = new CacheTags<LRU,NullCompression>(tags, compression_latency, true, store_compressed, adaptive_compression); \
        BUILD_CACHE(LRU, NullCompression, b, c); \
    } \
}

#define BUILD_IIC_CACHE(b ,c) { \
    IIC *tags = new IIC(iic_params); \
    if (compressed_bus || store_compressed){ \
	CacheTags<IIC, LZSSCompression> *tagStore = new CacheTags<IIC,LZSSCompression>(tags, compression_latency, true, store_compressed, adaptive_compression); \
        BUILD_CACHE(IIC, LZSSCompression, b, c); \
    } else { \
	CacheTags<IIC, NullCompression> *tagStore = new CacheTags<IIC,NullCompression>(tags, compression_latency, true, store_compressed, adaptive_compression); \
        BUILD_CACHE(IIC, NullCompression, b, c); \
    } \
}

#define BUILD_CACHES(b, c) { \
    if (repl == NULL) { \
        if (nsets == 1) { \
            BUILD_FALRU_CACHE(b, c); \
        } else { \
            BUILD_LRU_CACHE(b, c); \
        } \
    } else { \
       BUILD_IIC_CACHE(b, c); \
    } \
}

#define BUILD_COHERENCE(b) {\
    if (protocol == NULL) { \
        UniCoherence *coh = new UniCoherence(); \
        BUILD_CACHES(b, UniCoherence); \
    } else { \
        SimpleCoherence *coh = new SimpleCoherence(protocol); \
        BUILD_CACHES(b, SimpleCoherence); \
    }\
}

CREATE_SIM_OBJECT(BaseCache)
{
    string name = getInstanceName();
    int nsets = size / (assoc * block_size);
    if (subblock_size == 0) {
	subblock_size = block_size;
    }

    // Build BaseCache param object
    BaseCache::Params base_params(start_addr, addr_range, hit_latency, block_size, max_miss_count);
    
    // Build IIC params
    IIC::Params iic_params;
    iic_params.size = size;
    iic_params.nsets = nsets;
    iic_params.blkSize = block_size;
    iic_params.assoc = assoc;
    iic_params.hashDelay = hash_delay;
    iic_params.hitLatency = hit_latency;
    iic_params.rp = repl;
    iic_params.subblockSize = subblock_size;


    if (mshrs == 1 || out_bus->doEvents() == false) {
	BlockingBuffer *mq = new BlockingBuffer(true);
	BUILD_COHERENCE(BlockingBuffer);
    } else {
	MissQueue *mq = new MissQueue(mshrs, tgts_per_mshr, write_buffers, true);
	BUILD_COHERENCE(MissQueue);
    }
    return NULL;
}

REGISTER_SIM_OBJECT("BaseCache", BaseCache)


#endif //DOXYGEN_SHOULD_SKIP_THIS
