/* $Id: s.cache_impl.hh 1.110 04/06/14 00:45:29-04:00 ehallnor@zizzer.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/**
 * @file
 * Cache definitions.
 */

#include <assert.h>
#include <math.h>

#include <cassert>
#include <iostream>
#include <sstream>
#include <string>

#include "sim/host.hh"
#include "base/misc.hh"
#include "cpu/full_cpu/smt.hh"

#include "mem/cache/cache.hh"
#include "mem/cache/cache_blk.hh"
#include "mem/cache/miss/mshr.hh"

#include "mem/bus/bus.hh"

#include "mem/bus/slave_interface.hh"
#include "mem/memory_interface.hh"
#include "mem/bus/master_interface.hh"

#include "mem/mem_debug.hh"

#include "sim/sim_events.hh" // for SimExitEvent

using namespace std;

template<class TagStore, class Buffering, class Coherence>
Cache<TagStore,Buffering,Coherence>::
Cache(const std::string &_name, HierParams *hier_params, 
      Cache<TagStore,Buffering,Coherence>::Params &params)
    : BaseCache(_name, hier_params, params.baseParams), 
      tags(params.tags), missQueue(params.missQueue),
      coherence(params.coherence), doCopy(params.doCopy),
      blockOnCopy(params.blockOnCopy)
{
    if (params.in == NULL) {
	topLevelCache = true;
    }
    tags->setCache(this, params.out->width, params.out->clockRatio);
    missQueue->setCache(this);
    coherence->setCache(this);
}

template<class TagStore, class Buffering, class Coherence>
void
Cache<TagStore,Buffering,Coherence>::regStats()
{
    BaseCache::regStats();
    tags->regStats(name());
    missQueue->regStats(name());
    coherence->regStats(name());
}

template<class TagStore, class Buffering, class Coherence>
MemAccessResult
Cache<TagStore,Buffering,Coherence>::access(MemReqPtr &req)
{
    MemDebug::cacheAccess(req);
    BlkType *blk = NULL;
    MemReqList writebacks;
    int size = blkSize;
    int lat = hitLatency;
    if (req->cmd == Copy) {
	startCopy(req);
	/**
	 * @todo What return value makes sense on a copy.
	 */
	return MA_CACHE_MISS;
    }
    if (!req->isUncacheable()) {
	blk = tags->handleAccess(req, lat, writebacks);
    } else {
	size = req->size;
    }
    // If this is a block size write/hint (WH64) allocate the block here
    // if the coherence protocol allows it.
    /** @todo make the fast write alloc (wh64) work with coherence. */
    /** @todo Do we want to do fast writes for writebacks as well? */
    if (!blk && req->cmd == Write && req->size >= blkSize && 
	coherence->allowFastWrites() && 
	missQueue->findMSHR(req->paddr, req->asid) == NULL) {
	// not outstanding misses, can do this
	blk = tags->handleFill(NULL, req, CACHE_BLK_VALID | CACHE_BLK_WRITABLE,
			       writebacks);
	++fastWrites;
    }
    while (!writebacks.empty()) {
	missQueue->doWriteback(writebacks.front());
	writebacks.pop_front();
    }
    DPRINTF(Cache, "%s %d %x %d blk_addr: %x\n", req->cmd.toString(),
	    req->asid, req->paddr & (((ULL(1))<<48)-1), (blk) ? 1 : 0,
	    req->paddr & ~((Addr)blkSize - 1));
    if (blk) {
	// Hit
	hits[req->cmd.toIndex()][req->thread_num]++;
	// clear dirty bit if write through
	respond(req, curTick+lat);
	return MA_HIT;
	}

    // Miss
    misses[req->cmd.toIndex()][req->thread_num]++;
    /** @todo Move miss count code into BaseCache */
    if (missCount) {
	--missCount;
	if (missCount == 0)
	    new SimExitEvent("A cache reached the maximum miss count");
    }
    missQueue->handleMiss(req, size, curTick + hitLatency);
    return MA_CACHE_MISS;
}


template<class TagStore, class Buffering, class Coherence>
MemReqPtr
Cache<TagStore,Buffering,Coherence>::getMemReq()
{
    MemReqPtr req = missQueue->getMemReq();
    if (req) {
	if (!req->isUncacheable()) {
	    BlkType *blk = tags->findBlock(req);
	    MemCmd cmd = coherence->getBusCmd(req->cmd,
	    				      (blk)? blk->status : 0);
	    missQueue->setBusCmd(req, cmd);
	} 
    }

    assert(!doMasterRequest() || missQueue->havePending());
    assert(!req || req->time <= curTick);
    return req;
}

template<class TagStore, class Buffering, class Coherence>
void
Cache<TagStore,Buffering,Coherence>::sendResult(MemReqPtr &req, bool success)
{
    if (success) {
	missQueue->markInService(req);
    } else if (req && !req->isUncacheable()) {
	missQueue->restoreOrigCmd(req);
    }
}

template<class TagStore, class Buffering, class Coherence>
void
Cache<TagStore,Buffering,Coherence>::handleResponse(MemReqPtr &req)
{
    MemReqPtr copy_request;
    BlkType *blk = NULL;
    if (req->mshr) {
	MemDebug::cacheResponse(req);
	DPRINTF(Cache, "Handling reponse to %x\n",req->paddr);
	
	if (req->isCacheFill()) {
	    blk = tags->findBlock(req);
	    CacheBlk::State old_state = (blk) ? blk->status : 0;
	    MemReqList writebacks;
	    blk = tags->handleFill(blk, req->mshr, 
				   coherence->getNewState(req,old_state),
				   writebacks);
	    while (!writebacks.empty()) {
		if (writebacks.front()->cmd == Copy) {
		    copy_request = writebacks.front();
		} else {
		    missQueue->doWriteback(writebacks.front());
		}
		writebacks.pop_front();
	    }
	}
	if (copy_request) {
	    // The mshr is handled in handleCopy
	    handleCopy(copy_request, req->paddr, blk, req->mshr);
	} else {
	    missQueue->handleResponse(req, curTick + hitLatency);
	}
    }
}

template<class TagStore, class Buffering, class Coherence>
void
Cache<TagStore,Buffering,Coherence>::pseudoFill(Addr addr, int asid)
{
    // Need to temporarily move this blk into MSHRs
    MSHR *mshr = missQueue->allocateTargetList(addr, asid);
    int lat;
    MemReqList dummy;
    // Read the data into the mshr
    BlkType *blk = tags->handleAccess(mshr->req, lat, dummy, false);
    assert(dummy.empty());
    assert(mshr->req->isSatisfied());
    // can overload order since it isn't used on non pending blocks
    mshr->order = blk->status;
    // temporarily remove the block from the cache.
    tags->invalidateBlk(addr, asid);
}

template<class TagStore, class Buffering, class Coherence>
void
Cache<TagStore,Buffering,Coherence>::pseudoFill(MSHR *mshr)
{
    // Need to temporarily move this blk into MSHRs
    assert(mshr->req->cmd == Read);
    int lat;
    MemReqList dummy;
    // Read the data into the mshr
    BlkType *blk = tags->handleAccess(mshr->req, lat, dummy, false);
    assert(dummy.empty());
    assert(mshr->req->isSatisfied());
    // can overload order since it isn't used on non pending blocks
    mshr->order = blk->status;
    // temporarily remove the block from the cache.
    tags->invalidateBlk(mshr->req->paddr, mshr->req->asid);
}

template<class TagStore, class Buffering, class Coherence>
void
Cache<TagStore,Buffering,Coherence>::startCopy(MemReqPtr &req)
{
    MemDebug::cacheStartCopy(req);
    bool delayed = false;
    Addr source = req->paddr;
    Addr dest = req->dest;
    int asid = req->asid;
    
    bool source_unaligned = source & (blkSize - 1);
    bool dest_unaligned = dest & (blkSize - 1);

    // Block addresses for unaligned accesses.
    Addr source1 = source &  ~(Addr)(blkSize - 1);
    Addr source2 = source1 + blkSize;
    Addr dest1 = dest & ~(Addr)(blkSize - 1);
    Addr dest2 = dest1 + blkSize;

    MSHR *source1_mshr = missQueue->findMSHR(source1, asid);
    MSHR *source2_mshr = missQueue->findMSHR(source2, asid);
    MSHR *dest1_mshr = missQueue->findMSHR(dest1, asid);
    MSHR *dest2_mshr = missQueue->findMSHR(dest2, asid);

    if (source1_mshr) {
	req->flags |= COPY_SOURCE1;
	missQueue->addTarget(source1_mshr, req);
    }
    if (source2_mshr && source_unaligned) {
	req->flags |= COPY_SOURCE2;
	missQueue->addTarget(source2_mshr, req);
    }
    if (dest1_mshr) {
	req->flags |= COPY_DEST1;
	missQueue->addTarget(dest1_mshr, req);
    }
    if (dest2_mshr && dest_unaligned) {
	req->flags |= COPY_DEST2;
	missQueue->addTarget(dest2_mshr, req);
    }

    BlkType *source1_blk = tags->findBlock(source1, asid);
    BlkType *source2_blk = tags->findBlock(source2, asid);
    BlkType *dest1_blk = tags->findBlock(dest1, asid);
    BlkType *dest2_blk = tags->findBlock(dest2, asid);

    if (!doCopy) {
	// Writeback sources if dirty.
	if (source1_blk && source1_blk->isModified()) {
	    MemReqPtr writeback = tags->writebackBlk(source1_blk);
	    missQueue->doWriteback(writeback);
	}
	if (source_unaligned && source2_blk && source2_blk->isModified()) {
	    MemReqPtr writeback = tags->writebackBlk(source2_blk);
	    missQueue->doWriteback(writeback);
	}
	
	if (dest_unaligned) {
	    // Need to writeback dirty destinations
	    if (dest1_blk && dest1_blk->isModified()) {
		MemReqPtr writeback = tags->writebackBlk(dest1_blk);
		missQueue->doWriteback(writeback);
	    }
	    if (dest2_blk && dest2_blk->isModified()) {
		MemReqPtr writeback = tags->writebackBlk(dest2_blk);
		missQueue->doWriteback(writeback);
	    }
	    // Need to invalidate both dests
	    tags->invalidateBlk(dest2, asid);
	}
	// always need to invalidate this block.
	tags->invalidateBlk(dest1, asid);
	
	if (source1_mshr || (source2_mshr && source_unaligned) ||
	    dest1_mshr || (dest2_mshr && dest_unaligned)) {
	    
	    // Need to delay the copy until the outstanding requests
	    // finish.
	    delayed = true;
	    if (blockOnCopy) {
		setBlocked(Blocked_Copy);
	    }
	    if (!dest1_mshr) {
		dest1_mshr = missQueue->allocateTargetList(dest1, asid);
		dest1_mshr->req->xc = req->xc;
	    }
	    if (!dest2_mshr && dest_unaligned) {
		dest2_mshr = missQueue->allocateTargetList(dest2, asid);
		dest2_mshr->req->xc = req->xc;
	    }
	    if (!source1_mshr) {
		if (source1_blk) {
		    pseudoFill(source1, asid);
		} else {
		    source1_mshr = 
			missQueue->allocateTargetList(source1, asid);
		    source1_mshr->req->xc = req->xc;
		}
	    }
	    if (!source2_mshr && source_unaligned) {
		if (source2_blk) {
		    pseudoFill(source2, asid);
		} else {
		    source2_mshr = 
			missQueue->allocateTargetList(source2, asid);
		    source2_mshr->req->xc = req->xc;
		}
	    }
	} else {
	    // Forward the copy to the next level.
	    missQueue->doWriteback(req);
	}
    } else { // doCopy
	// Need to fetch sources if they aren't present.
	if (!source1_mshr && !source1_blk) {
	    req->flags |= COPY_SOURCE1;
	    source1_mshr = missQueue->fetchBlock(source1, asid, blkSize, 
						 curTick, req);
	}
	if (source_unaligned && !source2_mshr && !source2_blk) {
	    req->flags |= COPY_SOURCE2;
	    source1_mshr = missQueue->fetchBlock(source2, asid, blkSize, 
						 curTick, req);
	}
	
	if (dest_unaligned) {
	    if (!dest1_mshr && !dest1_blk) {
		req->flags |= COPY_DEST1;
		dest1_mshr = missQueue->fetchBlock(dest1, asid, blkSize,
						   curTick, req);
	    }
	    if (!dest2_mshr && !dest2_blk) {
		req->flags |= COPY_DEST2;
		dest2_mshr = missQueue->fetchBlock(dest2, asid, blkSize,
						   curTick, req);
	    }
	}
	
	if (source1_mshr || (source2_mshr && source_unaligned) ||
	    dest1_mshr || (dest2_mshr && dest_unaligned)) {

	    delayed = true;
	    if (blockOnCopy) {
		setBlocked(Blocked_Copy);
	    }
	    if (source1_blk) {
		pseudoFill(source1, asid);
	    }
	    if (source2_blk && source_unaligned) {
		pseudoFill(source2, asid);
	    }
	    if (dest_unaligned) {
		if (dest1_blk) {
		    pseudoFill(dest1, asid);
		}
		if (dest2_blk) {
		    pseudoFill(dest2, asid);
		}
	    } else {
		if (!dest1_mshr) {
		    tags->invalidateBlk(dest1, asid);
		    dest1_mshr = missQueue->allocateTargetList(dest1, asid);
		    dest1_mshr->req->xc = req->xc;
		}
	    }
	} else {
	    MemReqList writebacks;
	    if (source_unaligned || dest_unaligned) {
		tags->doUnalignedCopy(source, dest, asid, writebacks);
	    } else {
		tags->doCopy(source, dest, asid, writebacks);
	    }
	    while(!writebacks.empty()) {
		missQueue->doWriteback(writebacks.front());
		writebacks.pop_front();
	    }
	}   
    }
    DPRINTF(Cache, "Starting copy from %x (%x)  to %x (%x) %s\n", source, 
	    source1, dest, dest1,
	    (delayed)?"delayed":"successful");
}

template<class TagStore, class Buffering, class Coherence>
void
Cache<TagStore,Buffering,Coherence>::handleCopy(MemReqPtr &req, Addr addr,
						BlkType *blk, MSHR *mshr)
{
    MemDebug::cacheHandleCopy(req);
    Addr source = req->paddr;
    Addr dest = req->dest;
    int asid = req->asid;
    MemReqList writebacks;
    
    assert(mshr->getTarget()->cmd == Copy);

    bool source_unaligned = source & (blkSize - 1);
    bool dest_unaligned = dest & (blkSize - 1);

    // Block addresses for unaligned accesses.
    Addr source1 = source &  ~(Addr)(blkSize - 1);
    Addr source2 = source1 + blkSize;
    Addr dest1 = dest & ~(Addr)(blkSize - 1);
    Addr dest2 = dest1 + blkSize;

    DPRINTF(Cache,"Handling delayed copy from %x (%x) to %x (%x)\n\t" 
	    "pending = %x, addr = %x\n", source, source1, dest, dest1, 
	    req->flags & COPY_PENDING_MASK, 
	    addr);

    if (!doCopy) {
	if (addr == source1) {
	    req->flags &= ~COPY_SOURCE1;
	    // pop the copy off.
	    mshr->popTarget();
	    if (blk) {
		if (blk->isModified()) {
		    // writeback block.
		    MemReqPtr writeback = tags->writebackBlk(blk);
		    missQueue->doWriteback(writeback);
		}
		pseudoFill(mshr);
	    }			    
	}
	if (addr == source2) {
	    req->flags &= ~COPY_SOURCE2;
	    // pop the copy off.
	    mshr->popTarget();
	    if (blk) {
		if (blk->isModified()) {
		    // writeback block.
		    MemReqPtr writeback = tags->writebackBlk(blk);
		    missQueue->doWriteback(writeback);
		}
		pseudoFill(mshr);
	    }
	} 
	if (addr == dest1) {
	    req->flags &= ~COPY_DEST1;
	    if (dest_unaligned && blk && blk->isModified()) {
		// writeback block.
		MemReqPtr writeback = tags->writebackBlk(blk);
		missQueue->doWriteback(writeback);
	    }
	    tags->invalidateBlk(dest1, asid);
	    // Pop the copy request off of the list.
	    mshr->popTarget();
	    // Mark the mshr request as not statisfied to make it a targetlist
	    mshr->req->flags & ~SATISFIED;
	}
	if (addr == dest2) {
	    assert(dest_unaligned);
	    req->flags &= ~COPY_DEST2;
	    if (blk && blk->isModified()) {
		// writeback block.
		MemReqPtr writeback = tags->writebackBlk(blk);
		missQueue->doWriteback(writeback);
	    }
	    tags->invalidateBlk(dest2, asid);  
	    // Pop the copy request off of the list.
	    mshr->popTarget();
	    // Mark the mshr request as not statisfied to make it a targetlist
	    mshr->req->flags & ~SATISFIED;
	}
	if (!req->pendingCopy()) {
	    // Need to clear the satisfied flag before we forward
	    req->flags &= ~SATISFIED;
	    missQueue->doWriteback(req);
	    if (blockOnCopy) {
		clearBlocked(Blocked_Copy);
	    }
	    
	    // handle the first source block.
	    MSHR* source1_mshr = missQueue->findMSHR(source1, asid);
	    assert(source1_mshr);
	    if (source1_mshr->req->isSatisfied()) {
		CacheBlk::State state = source1_mshr->order;
		assert(state < 32);
		assert(req->xc || !doData());
		source1_mshr->req->xc = req->xc;
		tags->pseudoFill(source1_mshr, state, writebacks);
	    }
	    if (source1_mshr->hasTargets() && 
		source1_mshr->getTarget()->cmd == Copy) {
		
		MemReqPtr copy_request = source1_mshr->getTarget();
		handleCopy(copy_request, source1,
			   tags->findBlock(source1, asid), source1_mshr);
	    } else {
		// Refetch if necessary, or free the MSHR.
		missQueue->handleResponse(source1_mshr->req, 
					  curTick + hitLatency);
	    }

	    // Handle the second source if necessary
	    if (source_unaligned) {
		MSHR* source2_mshr = missQueue->findMSHR(source2, asid);
		assert(source2_mshr);
		if (source2_mshr->req->isSatisfied()) {
		    CacheBlk::State state = source2_mshr->order;
		    assert(state < 32);
		    assert(req->xc || !doData());
		    source2_mshr->req->xc = req->xc;
		    tags->pseudoFill(source2_mshr, state, writebacks);
		}
		if (source2_mshr->hasTargets() && 
		    source2_mshr->getTarget()->cmd == Copy) {
		    
		    MemReqPtr copy_request = source2_mshr->getTarget();
		    handleCopy(copy_request, source2,
			       tags->findBlock(source2, asid), source2_mshr);
		} else {
		    // Refetch if necessary, or free the MSHR.
		    missQueue->handleResponse(source2_mshr->req, 
					      curTick + hitLatency);
		}
	    }
	    
	    // Restart the first destination
	    MSHR* dest1_mshr = missQueue->findMSHR(dest1, asid);
	    // Restart the target list.
	    missQueue->handleResponse(dest1_mshr->req, 

				      curTick + hitLatency);

	    // Restart the second destination, if necessary
	    if (dest_unaligned) {
		MSHR* dest2_mshr = missQueue->findMSHR(dest2, asid);
		// Restart the target list.
		missQueue->handleResponse(dest2_mshr->req, 
					  curTick + hitLatency);
	    }
	}
    } else { // doCopy
	if (addr == source1) {
	    req->flags &= ~COPY_SOURCE1;
	    mshr->popTarget();
	    pseudoFill(mshr);
	}
	if (addr == source2) {
	    req->flags &= ~COPY_SOURCE2;
	    mshr->popTarget();
	    pseudoFill(mshr);
	}
	if (addr == dest1) {
	    req->flags &= ~COPY_DEST1;
	    mshr->popTarget();
	    if (dest_unaligned) {
		// Store the block for later
		pseudoFill(mshr);
	    } else {
		// Mark the MSHR as a target list
		mshr->req->flags &= ~SATISFIED;
		tags->invalidateBlk(dest1, asid);
	    }
	}
	if (addr == dest2) {
	    req->flags &= ~COPY_DEST2;
	    mshr->popTarget();
	    // Store the block for later
	    pseudoFill(mshr);
	}
	    
	if (!req->pendingCopy()) {
	    MSHR* source1_mshr = missQueue->findMSHR(source1, asid);
	    MSHR* source2_mshr = missQueue->findMSHR(source2, asid);
	    MSHR* dest1_mshr = missQueue->findMSHR(dest1, asid);
	    MSHR* dest2_mshr = missQueue->findMSHR(dest2, asid);
	 
	    // Refill the first Source
	    CacheBlk::State state = source1_mshr->order;
	    source1_mshr->req->xc = req->xc;
	    tags->handleFill(NULL, source1_mshr->req, state, writebacks, NULL);
	    
	    // Refill the second source, if necessary
	    if (source_unaligned) {
		state = source2_mshr->order;
		source2_mshr->req->xc = req->xc;
		tags->handleFill(NULL, source2_mshr->req, state, writebacks, 
				 NULL);
	    }

	    // Refill destination blocks, if necessary
	    if (dest_unaligned) {
		state = dest1_mshr->order;
		dest1_mshr->req->xc = req->xc;
		tags->handleFill(NULL, dest1_mshr->req, state, writebacks,
				 NULL);
		state = dest2_mshr->order;
		dest2_mshr->req->xc = req->xc;
		tags->handleFill(NULL, dest2_mshr->req, state, writebacks,
				 NULL);
	    }
	    
	    // Perform the copy
	    if (source_unaligned || dest_unaligned) {
		tags->doUnalignedCopy(source, dest, asid, writebacks);
	    } else {
		tags->doCopy(source, dest, asid, writebacks);
	    }
	    
	    if (blockOnCopy) {
		clearBlocked(Blocked_Copy);
	    }
	    
	    // Handle targets, if any
	    tags->handleTargets(source1_mshr, writebacks);
	    // If there is another outstanding copy, don't free
	    // the MSHR here.
	    if (source1_mshr->hasTargets() &&
		source1_mshr->getTarget()->cmd == Copy) {
			
		MemReqPtr copy_request = source1_mshr->getTarget();
		handleCopy(copy_request, source1,
			   tags->findBlock(source1, asid), source1_mshr);
	    } else {
		// Refetch if necessary, or free the MSHR
		missQueue->handleResponse(source1_mshr->req, 
					  curTick + hitLatency);
	    }
	    
	    if (source_unaligned) {
		tags->handleTargets(source2_mshr, writebacks);
		// If there is another outstanding copy, don't free
		// the MSHR here.
		if (source2_mshr->hasTargets() &&
		    source2_mshr->getTarget()->cmd == Copy) {
		    
		    MemReqPtr copy_request = source2_mshr->getTarget();
		    handleCopy(copy_request, source2,
			       tags->findBlock(source2, asid), source2_mshr);
		} else {
		    // Refetch if necessary, or free the MSHR
		    missQueue->handleResponse(source2_mshr->req, 
					      curTick + hitLatency);
		}
	    }
	    
	    tags->handleTargets(dest1_mshr, writebacks);
	    // If there is another outstanding copy, don't free
	    // the MSHR here.
	    if (dest1_mshr->hasTargets() && 
		dest1_mshr->getTarget()->cmd == Copy) {
		
		MemReqPtr copy_request = dest1_mshr->getTarget();
		handleCopy(copy_request, dest1,
			   tags->findBlock(dest1, asid), dest1_mshr);
	    } else {
		// Refetch if necessary, or free the MSHR
		missQueue->handleResponse(dest1_mshr->req, 
					  curTick + hitLatency);
	    }
	    
	    if (dest_unaligned) {
		tags->handleTargets(dest2_mshr, writebacks);
		// If there is another outstanding copy, don't free
		// the MSHR here.
		if (dest2_mshr->hasTargets() && 
		    dest2_mshr->getTarget()->cmd == Copy) {
		    
		    MemReqPtr copy_request = dest2_mshr->getTarget();
		    handleCopy(copy_request, dest2,
			       tags->findBlock(dest2, asid), dest2_mshr);
		} else {
		    // Refetch if necessary, or free the MSHR
		    missQueue->handleResponse(dest2_mshr->req, 
					      curTick + hitLatency);
		}
	    }
		
	    
	}
    }
    
    // Send off any writebacks
    while (!writebacks.empty()) {
	if (writebacks.front()->cmd != Copy) {
	    missQueue->doWriteback(writebacks.front());
	}
	writebacks.pop_front();
    }
}

template<class TagStore, class Buffering, class Coherence>
MemReqPtr
Cache<TagStore,Buffering,Coherence>::getCoherenceReq()
{
    return coherence->getMemReq();
}


template<class TagStore, class Buffering, class Coherence>
void
Cache<TagStore,Buffering,Coherence>::snoop(MemReqPtr &req)
{
    Addr blk_addr = req->paddr & ~(Addr(blkSize-1));
    BlkType *blk = tags->findBlock(req);
    MSHR *mshr = missQueue->findMSHR(blk_addr, req->asid);
    CacheBlk::State new_state;
    bool satisfy = coherence->handleBusRequest(req,blk,mshr, new_state);
    if (satisfy) {
	tags->handleSnoop(blk, new_state, req);
	respondToSnoop(req);
	return;
    }
    tags->handleSnoop(blk, new_state);
}

template<class TagStore, class Buffering, class Coherence>
void
Cache<TagStore,Buffering,Coherence>::invalidateBlk(Addr addr, int asid)
{
    tags->invalidateBlk(addr,asid);
}


/**
 * @todo Fix to not assume write allocate
 */
template<class TagStore, class Buffering, class Coherence>
Tick
Cache<TagStore,Buffering,Coherence>::probe(MemReqPtr &req, bool update)
{
    MemDebug::cacheProbe(req);
    if (!update && !doData()) {
	// Nothing to do here
	return mi->sendProbe(req,update);
    }

    MemReqList writebacks;
    int lat;
    BlkType *blk = tags->handleAccess(req, lat, writebacks, update);

    if (!blk) {
	// Need to check for outstanding misses and writes
	Addr blk_addr = req->paddr & ~(blkSize - 1);
	
	// There can only be one matching outstanding miss.
	MSHR* mshr = missQueue->findMSHR(blk_addr, req->asid);
	
	// There can be many matching outstanding writes.
	vector<MSHR*> writes;
	missQueue->findWrites(blk_addr, req->asid, writes);

	if (!update) {
	    mi->sendProbe(req, update);
	    // Check for data in MSHR and writebuffer.
	    if (mshr) {
		warn("Found outstanding miss on an non-update probe");
		MSHR::TargetList *targets = mshr->getTargetList();
		MSHR::TargetList::iterator i = targets->begin();
		MSHR::TargetList::iterator end = targets->end();
		for (; i != end; ++i) {
		    MemReqPtr target = *i;
		    // If the target contains data, and it overlaps the
		    // probed request, need to update data
		    if (target->cmd.isWrite() && target->overlaps(req)) {
			uint8_t* req_data;
			uint8_t* write_data;
			int data_size;
			if (target->paddr < req->paddr) {
			    int offset = req->paddr - target->paddr;
			    req_data = req->data;
			    write_data = target->data + offset;
			    data_size = target->size - offset;
			    assert(data_size > 0);
			    if (data_size > req->size) 
				data_size = req->size;
			} else {
			    int offset = target->paddr - req->paddr;
			    req_data = req->data + offset;
			    write_data = target->data;
			    data_size = req->size - offset;
			    assert(data_size > req->size);
			    if (data_size > target->size)
				data_size = target->size;
			}
			
			if (req->cmd.isWrite()) {
			    memcpy(req_data, write_data, data_size);
			} else {
			    memcpy(write_data, req_data, data_size);
			}
		    }
		}
	    }
	    for (int i = 0; i < writes.size(); ++i) {
		MemReqPtr write = writes[i]->req;
		if (write->overlaps(req)) {
		    warn("Found outstanding write on an non-update probe");
		    uint8_t* req_data;
		    uint8_t* write_data;
		    int data_size;
		    if (write->paddr < req->paddr) {
			int offset = req->paddr - write->paddr;
			req_data = req->data;
			write_data = write->data + offset;
			data_size = write->size - offset;
			assert(data_size > 0);
			if (data_size > req->size) 
			    data_size = req->size;
		    } else {
			int offset = write->paddr - req->paddr;
			req_data = req->data + offset;
			write_data = write->data;
			data_size = req->size - offset;
			assert(data_size > req->size);
			if (data_size > write->size)
			    data_size = write->size;
		    }
			
		    if (req->cmd.isWrite()) {
			memcpy(req_data, write_data, data_size);
		    } else {
			memcpy(write_data, req_data, data_size);
		    }
		    
		}
	    }
	    return 0;
	} else { 
	    // update the cache state and statistics
	    if (mshr || !writes.empty()){
		// Can't handle it, return request unsatisfied.
		return 0;
	    }
	    if (!req->isUncacheable()) {
		// Fetch the cache block to fill
		MemReqPtr busReq = new MemReq();
		busReq->paddr = blk_addr;
		busReq->size = blkSize;
		busReq->data = new uint8_t[blkSize];
		
		BlkType *blk = tags->findBlock(req);
		busReq->cmd = coherence->getBusCmd(req->cmd,
						   (blk)? blk->status : 0);
		
		busReq->asid = req->asid;
		busReq->xc = req->xc;
		busReq->thread_num = req->thread_num;
		busReq->time = curTick;
		
		lat = mi->sendProbe(busReq, update);
		
		if (!busReq->isSatisfied()) {
		    // blocked at a higher level, just return
		    return 0;
		}
		
		misses[req->cmd.toIndex()][req->thread_num]++;

		CacheBlk::State old_state = (blk) ? blk->status : 0;
		tags->handleFill(blk, busReq,
				 coherence->getNewState(busReq, old_state),
				 writebacks, req);
		// Handle writebacks if needed
		while (!writebacks.empty()){
		    mi->sendProbe(writebacks.front(), update);
		    writebacks.pop_front();
		}
		return lat + hitLatency;
	    } else {
		return mi->sendProbe(req,update);
	    }
	}
    } else {
	// There was a cache hit.
	// Handle writebacks if needed
	while (!writebacks.empty()){
	    mi->sendProbe(writebacks.front(), update);
	    writebacks.pop_front();
	}
	
	if (update) {
	    hits[req->cmd.toIndex()][req->thread_num]++;
	} else if (req->cmd.isWrite()) {
	    // Still need to change data in all locations.
	    return mi->sendProbe(req, update);
	}
	return curTick + lat;
    }
    fatal("Probe not handled.\n");
    return 0;
}

template<class TagStore, class Buffering, class Coherence>
Tick
Cache<TagStore,Buffering,Coherence>::snoopProbe(MemReqPtr &req, bool update)
{
    Addr blk_addr = req->paddr & ~(Addr(blkSize-1));
    BlkType *blk = tags->findBlock(req);
    MSHR *mshr = missQueue->findMSHR(blk_addr, req->asid);
    CacheBlk::State new_state = 0;
    bool satisfy = coherence->handleBusRequest(req,blk,mshr, new_state);
    if (satisfy) {
	tags->handleSnoop(blk, new_state, req);
	return hitLatency;
    }
    tags->handleSnoop(blk, new_state);
    return 0;
}

