/* $Id: s.coherence_protocol.cc 1.23 04/02/02 10:47:19-08:00 stever@isabel.reinhardt.house $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/**
 * @file
 * Definitions of CoherenceProtocol.
 */

#include <string>

#include "base/misc.hh"
#include "mem/cache/cache.hh"
#include "mem/cache/coherence/coherence_protocol.hh"
#include "sim/builder.hh"

using namespace std;


CoherenceProtocol::StateTransition::StateTransition()
    : busCmd(InvalidCmd), newState(-1), snoopFunc(invalidTransition)
{
}


void
CoherenceProtocol::regStats()
{
    // Even though we count all the possible transitions in the
    // requestCount and snoopCount arrays, most of these are invalid,
    // so we just select the interesting ones to print here.

    requestCount[Invalid][Read]
	.name(name() + ".read_invalid")
	.desc("read misses to invalid blocks")
	;

    requestCount[Invalid][Write]
	.name(name() +".write_invalid")
	.desc("write misses to invalid blocks")
	;

    requestCount[Shared][Write]
	.name(name() + ".write_shared")
	.desc("write misses to shared blocks")
	;

    requestCount[Owned][Write]
	.name(name() + ".write_owned")
	.desc("write misses to owned blocks")
	;

    snoopCount[Shared][Read]
	.name(name() + ".snoop_read_shared")
	.desc("read snoops on shared blocks")
	;

    snoopCount[Shared][ReadEx]
	.name(name() + ".snoop_readex_shared")
	.desc("readEx snoops on shared blocks")
	;

    snoopCount[Shared][Upgrade]
	.name(name() + ".snoop_upgrade_shared")
	.desc("upgradee snoops on shared blocks")
	;

    snoopCount[Modified][Read]
	.name(name() + ".snoop_read_modified")
	.desc("read snoops on modified blocks")
	;

    snoopCount[Modified][ReadEx]
	.name(name() + ".snoop_readex_modified")
	.desc("readEx snoops on modified blocks")
	;

    snoopCount[Owned][Read]
	.name(name() + ".snoop_read_owned")
	.desc("read snoops on owned blocks")
	;

    snoopCount[Owned][ReadEx]
	.name(name() + ".snoop_readex_owned")
	.desc("readEx snoops on owned blocks")
	;

    snoopCount[Owned][Upgrade]
	.name(name() + ".snoop_upgrade_owned")
	.desc("upgrade snoops on owned blocks")
	;

    snoopCount[Exclusive][Read]
	.name(name() + ".snoop_read_exclusive")
	.desc("read snoops on exclusive blocks")
	;

    snoopCount[Exclusive][ReadEx]
	.name(name() + ".snoop_readex_exclusive")
	.desc("readEx snoops on exclusive blocks")
	;
}


bool
CoherenceProtocol::invalidateTrans(BaseCache *cache, MemReqPtr &req,
				   CacheBlk *blk, MSHR *mshr, 
				   CacheBlk::State & new_state)
{
    // invalidate the block
    new_state = Invalid;
    return false;
}


bool
CoherenceProtocol::supplyTrans(BaseCache *cache, MemReqPtr &req, 
			       CacheBlk *blk,
			       MSHR *mshr, 
			       CacheBlk::State & new_state
			       )
{
    return true;
}


bool
CoherenceProtocol::supplyAndGotoSharedTrans(BaseCache *cache, MemReqPtr &req,
					    CacheBlk *blk,
					    MSHR *mshr, 
					    CacheBlk::State & new_state)
{
    new_state = Shared;
    return supplyTrans(cache, req, blk, mshr, new_state);
}


bool
CoherenceProtocol::supplyAndGotoOwnedTrans(BaseCache *cache, MemReqPtr &req,
					   CacheBlk *blk,
					   MSHR *mshr, 
					   CacheBlk::State & new_state)
{
    new_state = Owned;
    return supplyTrans(cache, req, blk, mshr, new_state);
}


bool
CoherenceProtocol::supplyAndInvalidateTrans(BaseCache *cache, MemReqPtr &req,
					    CacheBlk *blk,
					    MSHR *mshr, 
					    CacheBlk::State & new_state)
{
    new_state = Invalid;
    return supplyTrans(cache, req, blk, mshr, new_state);
}

CoherenceProtocol::CoherenceProtocol(const string &name,
				     const string &protocol,
				     const bool doUpgrades)
    : SimObject(name)
{
    if ((protocol == "mosi" || protocol == "moesi") && !doUpgrades) {
	cerr << "CoherenceProtocol: ownership protocols require upgrade transactions"
	     << "(write miss on owned block generates ReadExcl, which will clobber dirty block)"
	     << endl;
	fatal("");
    }

    MemCmdEnum writeToSharedCmd = doUpgrades ? Upgrade : ReadEx;

    // === common transitions ===

    // incoming requests: specify outgoing bus request
    transitionTable[Invalid][Read].onRequest(Read);
    transitionTable[Invalid][Write].onRequest(ReadEx);
    transitionTable[Shared][Write].onRequest(writeToSharedCmd);

    // on response to given request: specify new state
    transitionTable[Invalid][ReadEx].onResponse(Modified);
    transitionTable[Shared][writeToSharedCmd].onResponse(Modified);

    // bus snoop transition functions
    transitionTable[Invalid][Read].onSnoop(nullTransition);
    transitionTable[Invalid][ReadEx].onSnoop(nullTransition);
    transitionTable[Shared][Read].onSnoop(nullTransition);
    transitionTable[Shared][ReadEx].onSnoop(invalidateTrans);
    transitionTable[Modified][ReadEx].onSnoop(supplyAndInvalidateTrans);

    if (doUpgrades) {
	transitionTable[Invalid][Upgrade].onSnoop(nullTransition);
	transitionTable[Shared][Upgrade].onSnoop(invalidateTrans);
    }


    if (protocol == "msi" || protocol == "mosi") {
	// No Exclusive state: always go to Shared after read request
	transitionTable[Invalid][Read].onResponse(Shared);
    }
    else if (protocol == "mesi" || protocol == "moesi") {
	// Exclusive state
	cerr << "CoherenceProtocol: exclusive state not supported yet" << endl;
	fatal("");
    }
    else {
	cerr << "CoherenceProtocol: unrecognized protocol " << protocol
	     << " (exclusive state?)" << endl;
	fatal("");
    }

    if (protocol == "msi" || protocol == "mesi") {
	// No Owned state: go to Shared on read of Modified block.
	// Assumes main memory will self-update on cache-to-cache transfer.
	transitionTable[Modified][Read].onSnoop(supplyAndGotoSharedTrans);
    }
    else if (protocol == "mosi" || protocol == "moesi") {
	// Configure Owned state

	// incoming requests: specify outgoing bus request
	transitionTable[Owned][Write].onRequest(writeToSharedCmd);

	// on response to given request: specify new state
	transitionTable[Owned][writeToSharedCmd].onResponse(Modified);

	// bus snoop transition functions
	transitionTable[Owned][Read].onSnoop(supplyTrans);
	transitionTable[Owned][ReadEx].onSnoop(supplyAndInvalidateTrans);
	transitionTable[Modified][Read].onSnoop(supplyAndGotoOwnedTrans);

	if (doUpgrades) {
	    transitionTable[Owned][Upgrade].onSnoop(invalidateTrans);
	}
    }
    else {
	cerr << "CoherenceProtocol: unrecognized protocol " << protocol
	     << " (ownership?)" << endl;
	fatal("");
    }

    // reset statistics counters
    for (int s = 0; s <= stateMax; ++s) {
	for (int c = 0; c < NUM_MEM_CMDS; ++c) {
	    requestCount[s][c] = 0;
	    snoopCount[s][c] = 0;
	}
    }
}


MemCmd
CoherenceProtocol::getBusCmd(MemCmd cmdIn, CacheBlk::State state,
			     MSHR *mshr)
{
    state &= stateMask;
    int cmd_idx = cmdIn.toIndex();

    assert(0 <= state && state <= stateMax);
    assert(0 <= cmd_idx && cmd_idx < NUM_MEM_CMDS);

    MemCmd cmdOut = transitionTable[state][cmd_idx].busCmd;

    assert(cmdOut != InvalidCmd);

    ++requestCount[state][cmd_idx];

    return cmdOut;
}


CacheBlk::State
CoherenceProtocol::getNewState(const MemReqPtr &req, CacheBlk::State oldState)
{
    CacheBlk::State state = oldState & stateMask;
    int cmd_idx = req->cmd.toIndex();

    assert(0 <= state && state <= stateMax);
    assert(0 <= cmd_idx && cmd_idx < NUM_MEM_CMDS);

    CacheBlk::State newState = transitionTable[state][cmd_idx].newState;

    assert(newState != -1);

    return newState;
}


bool
CoherenceProtocol::handleBusRequest(BaseCache *cache, MemReqPtr &req,
				    CacheBlk *blk,
				    MSHR *mshr, 
				    CacheBlk::State & new_state)
{
    if (blk == NULL) {
	// nothing to do if we don't have a block
	return false;
    }

    CacheBlk::State state = blk->status & stateMask;
    int cmd_idx = req->cmd.toIndex();

    assert(0 <= state && state <= stateMax);
    assert(0 <= cmd_idx && cmd_idx < NUM_MEM_CMDS);

    assert(mshr == NULL); // can't currently handle outstanding requests

    ++snoopCount[state][cmd_idx];

    return transitionTable[state][cmd_idx].snoopFunc(cache, req, blk, mshr,
						     new_state);
}

bool
CoherenceProtocol::nullTransition(BaseCache *cache, MemReqPtr &req,
				  CacheBlk *blk, MSHR *mshr, 
				  CacheBlk::State & new_state)
{
    // do nothing
    if (blk)
	new_state = blk->status;
    return false;
}


bool
CoherenceProtocol::invalidTransition(BaseCache *cache, MemReqPtr &req,
				     CacheBlk *blk, MSHR *mshr, 
				     CacheBlk::State & new_state)
{
    panic("Invalid transition");
    return false;
}

#ifndef DOXYGEN_SHOULD_SKIP_THIS

BEGIN_DECLARE_SIM_OBJECT_PARAMS(CoherenceProtocol)

    Param<string> protocol;
    Param<bool> do_upgrades;

END_DECLARE_SIM_OBJECT_PARAMS(CoherenceProtocol)


BEGIN_INIT_SIM_OBJECT_PARAMS(CoherenceProtocol)

    INIT_PARAM(protocol, "name of coherence protocol"),
    INIT_PARAM_DFLT(do_upgrades, "use upgrade transactions?", true)

END_INIT_SIM_OBJECT_PARAMS(CoherenceProtocol)


CREATE_SIM_OBJECT(CoherenceProtocol)
{
    return new CoherenceProtocol(getInstanceName(), protocol,
				 do_upgrades);
}

REGISTER_SIM_OBJECT("CoherenceProtocol", CoherenceProtocol)

#endif // DOXYGEN_SHOULD_SKIP_THIS
