/* $Id */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/**
 * @file
 * Miss and writeback queue definitions.
 */

#include "mem/cache/miss/miss_queue.hh"
#include "cpu/full_cpu/smt.hh" //for maxThreadsPerCPU
#include "mem/cache/base_cache.hh"
#include "cpu/exec_context.hh"

using namespace std;

// simple constructor
/**
 * @todo Remove the +16 from the write buffer constructor once we handle
 * stalling on writebacks do to compression writes.
 */
MissQueue::MissQueue(int numMSHRs, int numTargets, int write_buffers,
		     bool write_allocate)
    : mq(numMSHRs, 4), wb(write_buffers,numMSHRs+1000), numMSHR(numMSHRs), 
      numTarget(numTargets), writeBuffers(write_buffers),
      writeAllocate(write_allocate), order(0)
{
    noTargetMSHR = NULL;
}

void
MissQueue::regStats(const string &name)
{
    using namespace Stats;

    writebacks
	.init(maxThreadsPerCPU)
	.name(name + ".writebacks")
	.desc("number of writebacks")
	.flags(total)
	;

    // MSHR hit statistics
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	mshr_hits[access_idx]
	    .init(maxThreadsPerCPU)
	    .name(name + "." + cstr + "_mshr_hits")
	    .desc("number of " + cstr + " MSHR hits")
	    .flags(total | nozero | nonan)
	    ;
    }

    demandMshrHits
	.name(name + ".demand_mshr_hits")
	.desc("number of demand (read+write) MSHR hits")
	.flags(total)
	;
    demandMshrHits = mshr_hits[Read] + mshr_hits[Write];

    overallMshrHits
	.name(name + ".overall_mshr_hits")
	.desc("number of overall MSHR hits")
	.flags(total)
	;
    overallMshrHits = demandMshrHits + mshr_hits[Soft_Prefetch] +
	mshr_hits[Hard_Prefetch];

    // MSHR miss statistics
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	mshr_misses[access_idx]
	    .init(maxThreadsPerCPU)
	    .name(name + "." + cstr + "_mshr_misses")
	    .desc("number of " + cstr + " MSHR misses")
	    .flags(total | nozero | nonan)
	    ;
    }

    demandMshrMisses
	.name(name + ".demand_mshr_misses")
	.desc("number of demand (read+write) MSHR misses")
	.flags(total)
	;
    demandMshrMisses = mshr_misses[Read] + mshr_misses[Write];

    overallMshrMisses
	.name(name + ".overall_mshr_misses")
	.desc("number of overall MSHR misses")
	.flags(total)
	;
    overallMshrMisses = demandMshrMisses + mshr_misses[Soft_Prefetch] +
	mshr_misses[Hard_Prefetch];

    // MSHR miss latency statistics
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	mshr_miss_latency[access_idx]
	    .init(maxThreadsPerCPU)
	    .name(name + "." + cstr + "_mshr_miss_latency")
	    .desc("number of " + cstr + " MSHR miss cycles")
	    .flags(total | nozero | nonan)
	    ;
    }

    demandMshrMissLatency
	.name(name + ".demand_mshr_miss_latency")
	.desc("number of demand (read+write) MSHR miss cycles")
	.flags(total)
	;
    demandMshrMissLatency = mshr_miss_latency[Read] + mshr_miss_latency[Write];

    overallMshrMissLatency
	.name(name + ".overall_mshr_miss_latency")
	.desc("number of overall MSHR miss cycles")
	.flags(total)
	;
    overallMshrMissLatency = demandMshrMissLatency +
	mshr_miss_latency[Soft_Prefetch] + mshr_miss_latency[Hard_Prefetch];


#if 0
    // MSHR access formulas
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	mshrAccesses[access_idx]
	    .name(name + "." + cstr + "_mshr_accesses")
	    .desc("number of " + cstr + " mshr accesses(hits+misses)")
	    .flags(total | nozero | nonan)
	    ;
	mshrAccesses[access_idx] =
	    mshr_hits[access_idx] + mshr_misses[access_idx];
    }

    demandMshrAccesses
	.name(name + ".demand_mshr_accesses")
	.desc("number of demand (read+write) mshr accesses")
	.flags(total | nozero | nonan)
	;
    demandMshrAccesses = demandMshrHits + demandMshrMisses;

    overallMshrAccesses
	.name(name + ".overall_mshr_accesses")
	.desc("number of overall (read+write) mshr accesses")
	.flags(total | nozero | nonan)
	;
    overallMshrAccesses = overallMshrHits + overallMshrMisses;
#endif

    // MSHR miss rate formulas
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	mshrMissRate[access_idx]
	    .name(name + "." + cstr + "_mshr_miss_rate")
	    .desc("mshr miss rate for " + cstr + " accesses")
	    .flags(total | nozero | nonan)
	    ;

	mshrMissRate[access_idx] =
	    mshr_misses[access_idx] / cache->accesses[access_idx];
    }

    demandMshrMissRate
	.name(name + ".demand_mshr_miss_rate")
	.desc("mshr miss rate for demand accesses")
	.flags(total)
	;
    demandMshrMissRate = demandMshrMisses / cache->demandAccesses;

    overallMshrMissRate
	.name(name + ".overall_mshr_miss_rate")
	.desc("mshr miss rate for overall accesses")
	.flags(total)
	;
    overallMshrMissRate = overallMshrMisses / cache->overallAccesses;

    // mshrMiss latency formulas
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	avgMshrMissLatency[access_idx]
	    .name(name + "." + cstr + "_avg_mshr_miss_latency")
	    .desc("average " + cstr + " mshr miss latency")
	    .flags(total | nozero | nonan)
	    ;

	avgMshrMissLatency[access_idx] =
	    mshr_miss_latency[access_idx] / mshr_misses[access_idx];
    }

    demandAvgMshrMissLatency
	.name(name + ".demand_avg_mshr_miss_latency")
	.desc("average overall mshr miss latency")
	.flags(total)
	;
    demandAvgMshrMissLatency = demandMshrMissLatency / demandMshrMisses;

    overallAvgMshrMissLatency
	.name(name + ".overall_avg_mshr_miss_latency")
	.desc("average overall mshr miss latency")
	.flags(total)
	;
    overallAvgMshrMissLatency = overallMshrMissLatency / overallMshrMisses;

    mshr_cap_events
	.init(maxThreadsPerCPU)
	.name(name + ".mshr_cap_events")
	.desc("number of times MSHR cap was activated")
	.flags(total)
	;

    //software prefetching stats
    soft_prefetch_mshr_full
	.init(maxThreadsPerCPU)
	.name(name + ".soft_prefetch_mshr_full")
	.desc("number of mshr full events for SW prefetching instrutions")
	.flags(total)
	;


}

void
MissQueue::setCache(BaseCache *_cache)
{
    cache = _cache;
    blkSize = cache->getBlockSize();
}

MSHR*
MissQueue::allocateMiss(MemReqPtr &req, int size, Tick time)
{
    MSHR* mshr = mq.allocate(req, size);
    mshr->order = order++;
    if (!req->isUncacheable()) {
	// Mark this as a cache line fill
	mshr->req->flags |= CACHE_LINE_FILL;
    }
    if (mq.isFull()) {
	cache->setBlocked(Blocked_NoMSHRs);
    }
    cache->setMasterRequest(Request_MSHR, time);
    return mshr;
}


MSHR*
MissQueue::allocateWrite(MemReqPtr &req, int size, Tick time)
{
    MSHR* mshr = wb.allocate(req,req->size);
    mshr->order = order++;
    if (cache->doData()){
	if (req->isCompressed()) {
	    delete [] mshr->req->data;
	    mshr->req->actualSize = req->actualSize;
	    mshr->req->data = new uint8_t[req->actualSize];
	    memcpy(mshr->req->data, req->data, req->actualSize);
	} else {
	    memcpy(mshr->req->data, req->data, req->size);
	}   
    }
    if (wb.isFull()) {
	cache->setBlocked(Blocked_NoWBBuffers);
    }
    cache->setMasterRequest(Request_WB, time);
    return mshr;
}


/**
 * @todo Remove SW prefetches on mshr hits.
 */
void
MissQueue::handleMiss(MemReqPtr &req, int blkSize, Tick time)
{
    int size = blkSize;
    Addr blkAddr = req->paddr & ~(Addr)(blkSize-1);
    MSHR* mshr = NULL;
    if (!req->isUncacheable()) {
	mshr = mq.findMatch(blkAddr, req->asid);
	if (mshr) {
	    mshr_hits[req->cmd.toIndex()][req->thread_num]++;
	    if (mshr->threadNum != req->thread_num) {
		mshr->threadNum = -1;
	    }
	    mq.allocateTarget(mshr, req);
	    if (mshr->getNumTargets() == numTarget) {
		noTargetMSHR = mshr;
		cache->setBlocked(Blocked_NoTargets);
		mq.moveToFront(mshr);
	    }
	    return;
	}
    } else {
	size = req->size;
    }
    mshr_misses[req->cmd.toIndex()][req->thread_num]++;
    if (req->cmd.isWrite() && (req->isUncacheable() || !writeAllocate ||
			       req->cmd.isNoResponse())) {
	/**
	 * @todo Add write merging here.
	 */
	mshr = allocateWrite(req, req->size, time);
	return;
    }

    mshr = allocateMiss(req, size, time);
}

MSHR* 
MissQueue::fetchBlock(Addr addr, int asid, int blk_size, Tick time,
		      MemReqPtr &target)
{
    Addr blkAddr = addr & ~(Addr)(blk_size - 1);
    assert(mq.findMatch(addr, asid) == NULL);
    MSHR *mshr = mq.allocateFetch(blkAddr, asid, blk_size, target);
    mshr->order = order++;
    mshr->req->flags |= CACHE_LINE_FILL;
    if (mq.isFull()) {
	cache->setBlocked(Blocked_NoMSHRs);
    }
    cache->setMasterRequest(Request_MSHR, time);
    return mshr;
}

MemReqPtr
MissQueue::getMemReq(){
    MemReqPtr req = mq.getReq();
    if (((wb.isFull() && wb.inServiceMSHRs == 0) || !req || 
	 req->time > curTick) && wb.havePending()) {
	req = wb.getReq();
	// Need to search for earlier miss.
	MSHR *mshr = mq.findPending(req);
	if (mshr && mshr->order < req->mshr->order) {
	    // Service misses in order until conflict is cleared.
	    return mq.getReq();
	}
    }
    if (req) {
	MSHR* mshr = wb.findPending(req);
	if (mshr && mshr->order < req->mshr->order) {
	    // Earlier write, need to service it first
	    // To perserve operation order, service writes in order
	    return wb.getReq();
	}
    }
    return req;
}

void
MissQueue::setBusCmd(MemReqPtr &req, MemCmd cmd)
{
    assert(req->mshr != 0);
    MSHR * mshr = req->mshr;
    mshr->originalCmd = req->cmd;
    if (req->isCacheFill())
	req->cmd = cmd;
}

void
MissQueue::restoreOrigCmd(MemReqPtr &req)
{
    req->cmd = req->mshr->originalCmd;
}

void
MissQueue::markInService(MemReqPtr &req)
{
    assert(req->mshr != 0);
    bool unblock = false;
    BlockedCause cause = NUM_BLOCKED_CAUSES;

    /**
     * @todo Should include MSHRQueue pointer in MSHR to select the correct
     * one.
     */
    if ((!req->isCacheFill() && req->cmd.isWrite()) || req->cmd == Copy) {
	// Forwarding a write/ writeback, don't need to change
	// the command
	unblock = wb.isFull();
	wb.markInService(req->mshr);
	if (!wb.havePending()){
	    cache->clearMasterRequest(Request_WB);
	}
	if (unblock) {
	    // Do we really unblock?
	    unblock = !wb.isFull();
	    cause = Blocked_NoWBBuffers;
	}
    } else {
	unblock = mq.isFull();
	mq.markInService(req->mshr);
	if (!mq.havePending()){
	    cache->clearMasterRequest(Request_MSHR);
	}
	if (unblock) {
	    unblock = !mq.isFull();
	    cause = Blocked_NoMSHRs;
	}
    }
    if (unblock) {
	cache->clearBlocked(cause);
    }
}

void
MissQueue::handleResponse(MemReqPtr &req, Tick time)
{
    MSHR* mshr = req->mshr;
#ifndef NDEBUG
    int num_targets = mshr->getNumTargets();
#endif
    
    bool unblock = false;
    bool unblock_target = false;
    BlockedCause cause = NUM_BLOCKED_CAUSES;

    if (req->isCacheFill()) {
	mshr_miss_latency[mshr->originalCmd][req->thread_num] +=
	    curTick - req->time;
	// targets were handled in the cache tags
	if (mshr == noTargetMSHR) {
	    // we always clear at least one target
	    unblock_target = true;
	    cause = Blocked_NoTargets;
	    noTargetMSHR = NULL;
	}

	if (mshr->hasTargets()) {
	    // Didn't satisfy all the targets, need to resend
	    MemCmd cmd = mshr->getTarget()->cmd;
	    mq.markPending(mshr, cmd);
	    mshr->order = order++;
	    cache->setMasterRequest(Request_MSHR, time);
	    return;
	}
	unblock = mq.isFull();
	mq.deallocate(mshr);
	if (unblock) {
	    unblock = !mq.isFull();
	    cause = Blocked_NoMSHRs;
	}
    } else {
	if (mshr->hasTargets()) {
	    // Should only have 1 target if we had any
	    assert(num_targets == 1);
	    MemReqPtr target = mshr->getTarget();
	    mshr->popTarget();
	    if (cache->doData() && req->cmd.isRead()) {
		memcpy(target->data, req->data, target->size);
	    }
	    cache->respond(target, time);
	    assert(!mshr->hasTargets());
	}

	if (req->cmd.isWrite()) {
	    // If the wrtie buffer is full, we might unblock now
	    unblock = wb.isFull(); 
	    wb.deallocate(mshr);
	    if (unblock) {
		// Did we really unblock?
		unblock = !wb.isFull();
		cause = Blocked_NoWBBuffers;
	    }
	} else {
	    unblock = mq.isFull();
	    mq.deallocate(mshr);
	    if (unblock) {
		unblock = !mq.isFull();
		cause = Blocked_NoMSHRs;
	    }
	}
    }
    if (unblock || unblock_target) {
	cache->clearBlocked(cause);
    }
}

void
MissQueue::squash(int thread_number)
{
    bool unblock = false;
    BlockedCause cause = NUM_BLOCKED_CAUSES;

    if (noTargetMSHR && noTargetMSHR->threadNum == thread_number) {
	noTargetMSHR = NULL;
	unblock = true;
	cause = Blocked_NoTargets;
    }
    if (mq.isFull()) {
	unblock = true;
	cause = Blocked_NoMSHRs;
    }
    mq.squash(thread_number);
    if (!mq.havePending()) {
	cache->clearMasterRequest(Request_MSHR);
    }
    if (unblock && !mq.isFull()) {
	cache->clearBlocked(cause);
    }

}

MSHR*
MissQueue::findMSHR(Addr addr, int asid) const
{
    return mq.findMatch(addr,asid);
}

bool
MissQueue::findWrites(Addr addr, int asid, vector<MSHR*> &writes) const
{
    return wb.findMatches(addr,asid,writes);
}

void
MissQueue::doWriteback(Addr addr, int asid, ExecContext *xc,
		       int size, uint8_t *data, bool compressed)
{
    // Generate request
    MemReqPtr req = buildWritebackReq(addr, asid, xc, size, data,
				      compressed);

    writebacks[req->thread_num]++;

    allocateWrite(req, 0, curTick);
}


void
MissQueue::doWriteback(MemReqPtr &req)
{
    writebacks[req->thread_num]++;
    assert(req->xc || !cache->doData());
    allocateWrite(req, 0, curTick);
}


MSHR* 
MissQueue::allocateTargetList(Addr addr, int asid)
{
   MSHR* mshr = mq.allocateTargetList(addr, asid, blkSize);
   mshr->req->flags |= CACHE_LINE_FILL;
   if (mq.isFull()) {
       cache->setBlocked(Blocked_NoMSHRs);
   }
   return mshr;
}
