/* $Id: s.functional_memory.cc 1.25 04/04/02 22:57:07-08:00 stever@isabel.reinhardt.house $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

#include <string>

#include "sim/host.hh"
#include "sim/param.hh"
#include "base/cprintf.hh"
#include "sim/debug.hh"
#include "mem/functional_mem/functional_memory.hh"
#include "mem/functional_mem/memory_control.hh"
#include "mem/functional_mem/main_memory.hh"
#include "base/misc.hh"
#include "cpu/full_cpu/smt.hh"
#include "sim/universe.hh"

using namespace std;

Addr break_addr = 0;
bool break_physical = false;
int break_thread = 0;
FunctionalMemory *debug_mem = NULL;
Addr debug_addr = 0;
bool debug_physical = false;

void
print_data(const uint8_t *data)
{
    cprintf("%#016x\n"
	    "%#08x\n"
	    "%#04x\n"
	    "%#02x\n",
	    *(uint64_t *)data,
	    *(uint32_t *)data,
	    *(uint16_t *)data,
	    *(uint8_t *)data);
}

void
debug_mem_read(Addr addr, bool phy)
{
    uint8_t buf[16];
    uint8_t *bufp = buf;
    if (debug_mem) {
#ifdef FULL_SYSTEM
	MemReqPtr req = new MemReq(addr & ~7, NULL, 8, (phy ? PHYSICAL : 0));
	if (phy) req->paddr = req->vaddr & ULL(0xFFFFFFFFFF);
#else
	MemReqPtr req = new MemReq(addr & ~7, NULL, 8);
#endif
	debug_mem->read(req, bufp);
	print_data(bufp);
    }
}

void
debug_mem_read()
{ debug_mem_read(debug_addr, debug_physical); }

void
mem_break(const uint8_t *data)
{
    if (data)
	print_data(data);

    debug_break();
}

FunctionalMemory::FunctionalMemory(const string &name)
    : SimObject(name)
{ }

FunctionalMemory::~FunctionalMemory()
{ }


DEFINE_SIM_OBJECT_CLASS_NAME("FunctionalMemory", FunctionalMemory)

void
FunctionalMemory::prot_read(Addr addr, uint8_t *p, int size)
{ fatal("unimplemented"); }

void
FunctionalMemory::prot_write(Addr addr, const uint8_t *p, int size)
{ fatal("unimplemented"); }

void
FunctionalMemory::prot_memset(Addr addr, uint8_t val, int size)
{ fatal("unimplemented"); }


void
FunctionalMemory::access(MemCmd cmd, Addr addr, void *p, int nbytes)
{
    switch(cmd.toIndex()) {
      case Read:
	prot_read(addr, (uint8_t *)p, nbytes);
	return;

      case Write:
	prot_write(addr, (uint8_t *)p, nbytes);
	return;

      default:
	panic("unimplemented");
	return;
    }
}

Fault
FunctionalMemory::writeString(Addr addr, const char *str)
{
    uint8_t c;
    Fault fault;
    MemReqPtr req = new MemReq(addr, 0, 0);

    // EGH This is a hack to "translate" the address correctly
    // Only works in non FULL_SYSTEM
#ifdef FULL_SYSTEM
    fatal("Should not call this in full system mode.");
#endif
    req->paddr = req->vaddr;

    // copy until string terminator ('\0') is encountered
    do {
	c = *str++;
	fault = write(req, c);
	req->paddr++;
	req->vaddr++;
	if (fault != No_Fault)
	    return fault;
    } while (c);

    return No_Fault;
}


Fault
FunctionalMemory::readString(string &str, Addr addr)
{
    uint8_t c;
    Fault fault;
    MemReqPtr req = new MemReq(addr, 0, 0);

    // EGH This is a hack to "translate" the address correctly
    // Only works in non FULL_SYSTEM
#ifdef FULL_SYSTEM
    fatal("Should not call this in full system mode.");
#endif

   req->paddr = req->vaddr;

    // copy until string terminator ('\0') is encountered or n chars
    // have been copied
    do {
	fault = read(req, c);
	//EGH again, because of faxe translation, need to advance both of them.
	req->vaddr++;
	req->paddr++;
	if (fault != No_Fault)
	    return fault;
	str += c;
    } while (c);

    return No_Fault;
}


#if 0
void
FunctionalMemory::memory_breakpoint(Addr addr, int nbytes, bool read,
				    const uint8_t *p) const
{
    int i, j, n;

    if (break_address && (break_address == addr) &&
	(break_size == nbytes)) {
	ccprintf(cerr, "reached memory break point @ cycle = %d : ", curTick);

	if (read)
	    cerr << "(RD) : ";
	else
	    cerr << "(WR) : ";

	for (j = 0; j < (nbytes % 4) + 1; j++) {
	    n = (nbytes > 3) ? 4 : nbytes;
	    for (i = n - 1; i >= 0; i--)
		ccprintf(cerr, "%02X ", *(((unsigned char *) p) + i));
	    cerr << "; ";
	}
	cerr << "\n";
    }
}
#endif

#if 0
// dump a block of memory, returns any faults encountered
Fault
FunctionalMemory::dump(Addr addr,  //target address to dump
		       int len,      //number bytes to dump
		       ostream &stream) // output stream
{
    uint32_t data;
    Fault fault;

    if (!stream)
	stream = stderr;

    addr &= ~sizeof(uint32_t);
    len = (len + (sizeof(uint32_t) - 1)) & ~sizeof(uint32_t);

    MemReqPtr req = new MemReq(addr, 0, sizeof(uint32_t));
    while (len-- > 0) {
	fault = read4(req, data);
	if (fault != No_Fault)
	    return fault;

	ccprintf(stream, "%#08x: %08x\n", addr, data);
	req->vaddr += sizeof(uint32_t);
    }

    // no faults
    return No_Fault;
}
#endif
