/* $Id: s.main_memory.hh 1.45 04/05/21 13:03:14-04:00 binkertn@ziff.eecs.umich.edu $ */

/* @file
 */

#ifndef __MAIN_MEMORY_HH__
#define __MAIN_MEMORY_HH__

#include "mem/functional_mem/functional_memory.hh"

#include "base/statistics.hh"
#include "sim/stats.hh"

// number of entries in page translation hash table (must be power-of-two)
#define MEM_PTAB_SIZE		(32*1024)
#define MEM_LOG_PTAB_SIZE	15

/*
 * Model of infinite virtual memory for a standalone application
 */
class MainMemory : public FunctionalMemory
{
  public:
    friend class simple_disk;

  private:
    // prevent copying of a MainMemory object
    MainMemory(const MainMemory &specmem);
    const MainMemory &operator=(const MainMemory &specmem);

    /*
     *
     */
    class LockedAddr {
	// on alpha, minimum LL/SC granularity is 16 bytes, so lower
	// bits need to masked off.
	static const Addr Addr_Mask = 0xf;

	Addr addr;			// locked address
	ExecContext *xc;	// locking context

      public:
	// mask off unneeded address bits
	static Addr maskAddr(Addr _addr) { return _addr & ~Addr_Mask; }

	// change locked address
	void setAddr(Addr _addr) { addr = maskAddr(_addr); }

	// check for matching reference address
	bool matchesAddr(Addr addr2) { return (addr == maskAddr(addr2)); }

	// check for matching execution context
	bool matchesContext(ExecContext *xc2)
	{ return (xc == xc2); }

	// return pointer to execution context
	ExecContext *getContext() { return xc; }

	LockedAddr(Addr _addr, ExecContext *_xc)
	    : addr(maskAddr(_addr)), xc(_xc)
	{
	}
    };

    std::list<LockedAddr> lockedAddrList;

    // helper function for checkLockedAddrs(): we really want to
    // inline a quick check for an empty locked addr list (hopefully
    // the common case), and do the full list search (if necessary) in
    // this out-of-line function
    bool checkLockedAddrList(MemReqPtr &req);

  protected:
    // page table entry
    struct entry
    {
	entry *next;		// next translation in this bucket
	Addr tag;			// virtual page number tag
	uint8_t *page;		// page pointer
    };

    entry *ptab[MEM_PTAB_SIZE];	// inverted page table

    // memory object stats
    Stats::Scalar<> page_count;		// total number of pages allocated
    Stats::Scalar<> ptab_misses;	// total first level page tbl misses
    Stats::Scalar<> ptab_accesses;	// total page table accesses

    Stats::Formula page_mem;
    Stats::Formula ptab_miss_rate;

    static Addr offset(Addr addr);
    static Addr ptab_set(Addr addr);
    static Addr ptab_tag(Addr addr);

    uint8_t *page(Addr addr);
    uint8_t *translate(Addr addr);
    uint8_t *newpage(Addr addr);

    // Check for access faults to page data
    Fault page_check(Addr addr, int size) const;

    // Read/Write arbitrary amounts of data within a page
    Fault page_read(Addr addr, uint8_t *p, int size);
    Fault page_write(Addr addr, const uint8_t *p, int size);
    Fault page_set(Addr addr, uint8_t val, int size);

    // Speed up access for certain data access sizes
    template <class T> Fault page_read(Addr addr, T &data);
    template <class T> Fault page_write(Addr addr, T data);

    // Record the address of a load-locked operation so that we can
    // clear the execution context's lock flag if a matching store is
    // performed
    void trackLoadLocked(MemReqPtr &req);

    // Compare a store address with any locked addresses so we can
    // clear the lock flag appropriately.  Return value set to 'false'
    // if store operation should be suppressed (because it was a
    // conditional store and the address was no longer locked by the
    // requesting execution context), 'true' otherwise.  Note that
    // this method must be called on *all* stores since even
    // non-conditional stores must clear any matching lock addresses.
    bool checkLockedAddrs(MemReqPtr &req) {
	if (lockedAddrList.empty()) {
	    // no locked addrs: nothing to check, store_conditional fails
	    return !(req->flags & LOCKED);
	} else {
	    // iterate over list...
	    return checkLockedAddrList(req);
	}
    }

  public:
    MainMemory(const std::string &n);
    virtual ~MainMemory();

    // Read/Write arbitrary amounts of data to simulated memory space
    virtual void prot_read(Addr addr, uint8_t *p, int size);
    virtual void prot_write(Addr addr, const uint8_t *p, int size);
    virtual void prot_memset(Addr addr, uint8_t val, int size);

    virtual Fault read(MemReqPtr &req, uint8_t *data);
    virtual Fault write(MemReqPtr &req, const uint8_t *data);

    virtual Fault read(MemReqPtr &req, uint8_t &data);
    virtual Fault read(MemReqPtr &req, uint16_t &data);
    virtual Fault read(MemReqPtr &req, uint32_t &data);
    virtual Fault read(MemReqPtr &req, uint64_t &data);

    virtual Fault write(MemReqPtr &req, uint8_t data);
    virtual Fault write(MemReqPtr &req, uint16_t data);
    virtual Fault write(MemReqPtr &req, uint32_t data);
    virtual Fault write(MemReqPtr &req, uint64_t data);

  public:
    virtual void regStats();
    virtual void regFormulas();
};

// compute address of access within a host page
inline Addr
MainMemory::offset(Addr addr)
{ return addr & (VMPageSize - 1); }

// compute page table set
inline Addr
MainMemory::ptab_set(Addr addr)
{ return (addr >> LogVMPageSize) & (MEM_PTAB_SIZE - 1); }

// compute page table tag
inline Addr
MainMemory::ptab_tag(Addr addr)
{ return addr >> (LogVMPageSize + MEM_LOG_PTAB_SIZE); }

inline Fault
MainMemory::page_read(Addr addr, uint8_t *data, int size)
{
    uint8_t *p = page(addr);

    if (p)
	::memcpy(data, p + offset(addr), size);
    else
	::memset(data, 0, size);

    MEM_ADDR_TEST(addr);
    return No_Fault;
}

inline Fault
MainMemory::page_write(Addr addr, const uint8_t *data, int size)
{
    uint8_t *p = page(addr);

    // allocate page for address if we don't have one.
    if (!p)
	p = newpage(addr);

    ::memcpy(p + offset(addr), data, size);

    MEM_ADDR_TEST(addr);
    return No_Fault;
}

inline Fault
MainMemory::page_set(Addr addr, uint8_t val, int size)
{
    uint8_t *p = page(addr);

    // allocate page for address if we don't have one.
    if (!p)
	p = newpage(addr);

    ::memset(p + offset(addr), val, size);

    MEM_ADDR_TEST(addr);
    return No_Fault;
}

template <class T>
inline Fault
MainMemory::page_read(Addr addr, T &data)
{
    uint8_t *p = page(addr);

    if (p)
	data = *((T *)(p + offset(addr)));
    else
	// page not yet allocated, return zero value
	data = T();

    MEM_ADDR_TEST(addr);
    return No_Fault;
}

template <class T>
inline Fault
MainMemory::page_write(Addr addr, T data)
{
    uint8_t *p = page(addr);

    // allocate page for address if we don't have one.
    if (!p)
	p = newpage(addr);

    *((T *)(page(addr) + offset(addr))) = data;

    MEM_ADDR_TEST(addr);
    return No_Fault;
}

inline Fault
MainMemory::read(MemReqPtr &req, uint8_t &data)
{
    MEM_BLOCK_TEST(req->paddr, uint8_t);
    if (req->flags & LOCKED)
	trackLoadLocked(req);
    page_read(req->paddr, data);
    return No_Fault;
}

inline Fault
MainMemory::read(MemReqPtr &req, uint16_t &data)
{
    MEM_BLOCK_TEST(req->paddr, uint16_t);
    if (req->paddr & (sizeof(uint16_t) - 1)) return Alignment_Fault;
    if (req->flags & LOCKED)
	trackLoadLocked(req);
    page_read(req->paddr, data);
    return No_Fault;
}

inline Fault
MainMemory::read(MemReqPtr &req, uint32_t &data)
{
    MEM_BLOCK_TEST(req->paddr, uint32_t);
    if (req->paddr & (sizeof(uint32_t) - 1)) return Alignment_Fault;
    if (req->flags & LOCKED)
	trackLoadLocked(req);
    page_read(req->paddr, data);
    return No_Fault;
}

inline Fault
MainMemory::read(MemReqPtr &req, uint64_t &data)
{
    MEM_BLOCK_TEST(req->paddr, uint64_t);
    if (req->paddr & (sizeof(uint64_t) - 1)) return Alignment_Fault;
    if (req->flags & LOCKED)
	trackLoadLocked(req);
    page_read(req->paddr, data);
    return No_Fault;
}

inline Fault
MainMemory::write(MemReqPtr &req, uint8_t data)
{
    MEM_BLOCK_TEST(req->paddr, uint8_t);
    if (checkLockedAddrs(req))
	page_write(req->paddr, data);
    return No_Fault;
}

inline Fault
MainMemory::write(MemReqPtr &req, uint16_t data)
{
    MEM_BLOCK_TEST(req->paddr, uint16_t);
    if (req->paddr & (sizeof(uint16_t) - 1)) return Alignment_Fault;
    if (checkLockedAddrs(req))
	page_write(req->paddr, data);
    return No_Fault;
}

inline Fault
MainMemory::write(MemReqPtr &req, uint32_t data)
{
    MEM_BLOCK_TEST(req->paddr, uint32_t);
    if (req->paddr & (sizeof(uint32_t) - 1)) return Alignment_Fault;
    if (checkLockedAddrs(req))
	page_write(req->paddr, data);
    return No_Fault;
}

inline Fault
MainMemory::write(MemReqPtr &req, uint64_t data)
{
    MEM_BLOCK_TEST(req->paddr, uint64_t);
    if (req->paddr & (sizeof(uint64_t) - 1)) return Alignment_Fault;
    if (checkLockedAddrs(req))
	page_write(req->paddr, data);
    return No_Fault;
}

#endif // __MAIN_MEMORY_HH__
