/* $Id: s.memory_control.cc 1.30 04/04/02 22:57:07-08:00 stever@isabel.reinhardt.house $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

#include <string>
#include <vector>

#include "base/misc.hh"
#include "mem/functional_mem/memory_control.hh"
#include "sim/builder.hh"
#include "sim/sim_object.hh"

using namespace std;

MemoryController::MemoryController(const string &name, int cap)
    : FunctionalMemory(name), mapsize(0), mapcap(cap)
{
    map = new devmap[cap];
}

MemoryController::~MemoryController()
{
    if (map)
	delete [] map;
}

void
MemoryController::add_child(FunctionalMemory *obj, const Range<Addr> &range)
{
    if (mapsize == mapcap)
	panic("map is not large enough to add another child\n");

    // check for overlapping ranges!
    for (int i = 0; i < mapsize; ++i) {
	devmap *dm = &map[i];
	if (range <= dm->range && range >= dm->range)
	    panic("range of %s(%#x) overlaps with range of %s(%#x)\n",
		  obj->name(), range, dm->obj->name(), dm->range);
    }

    devmap *mp = &map[mapsize++];

    mp->obj = obj;
    mp->range = range;
    mp->range.start &= PA_IMPL_MASK;
    mp->range.end &= PA_IMPL_MASK;

    DPRINTF(MMU, "add %s: start=%#x end=%#x\n",
	    obj->name(), mp->range.start, mp->range.end);
}

bool
MemoryController::badaddr(Addr paddr) const
{
    for (int i = 0; i < mapsize; i++)
	if (paddr == map[i].range)
	    return map[i].obj->badaddr(paddr);

    return true;
}

MemoryController::devmap *
MemoryController::find_child(MemReqPtr &req) const
{
    for (int i = 0; i < mapsize; i++) {
	if (req->paddr == map[i].range) {
	    int index = i;

	    // Move up one.
	    if (i > 0) {
		devmap temp = map[i - 1];
		map[i - 1] = map[i];
		map[i] = temp;

		index = i - 1;
	    }

	    return &map[index];
	}
    }

    DPRINTF(MMU, "child not found. vaddr=%#x paddr=%#x\n",
	    req->vaddr, req->paddr);

    return NULL;
}

void
MemoryController::dump() const
{
    for (int i = 0; i < mapsize; i++)
	cprintf("%d: %s: range=%#x\n", i, map[i].obj->name(), map[i].range);
}

BEGIN_DECLARE_SIM_OBJECT_PARAMS(MemoryController)

    Param<int> capacity;

END_DECLARE_SIM_OBJECT_PARAMS(MemoryController)

BEGIN_INIT_SIM_OBJECT_PARAMS(MemoryController)

    INIT_PARAM_DFLT(capacity, "Maximum Number of children", 32)

END_INIT_SIM_OBJECT_PARAMS(MemoryController)

CREATE_SIM_OBJECT(MemoryController)
{
    return new MemoryController(getInstanceName(), capacity);
}

REGISTER_SIM_OBJECT("MemoryController", MemoryController)

