/* $Id: s.physical_memory.cc 1.46 04/04/02 22:57:07-08:00 stever@isabel.reinhardt.house $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include <cstdio>
#include <iostream>
#include <string>


#include "base/misc.hh"
#ifdef FULL_SYSTEM
#include "mem/functional_mem/memory_control.hh"
#endif
#include "mem/functional_mem/physical_memory.hh"
#include "sim/host.hh"
#include "sim/builder.hh"

using namespace std;

PhysicalMemory::PhysicalMemory(const string &n, Addr addr, Addr size,
			       MemoryController *mmu, const std::string &fname)
    : FunctionalMemory(n), base_addr(addr), pmem_size(size), pmem_addr(NULL)
{
    mmu->add_child(this, Range<Addr>(base_addr, base_addr + pmem_size));

    int fd = -1;

    if (!fname.empty()) {
	fd = open(fname.c_str(), O_RDWR | O_CREAT, 0644);
	if (fd == -1) {
	    perror("open");
	    fatal("Could not open physical memory file: %s\n", fname);
	}
	ftruncate(fd, pmem_size);
    }

    int map_flags = (fd == -1) ? (MAP_ANON | MAP_PRIVATE) : MAP_SHARED;
    pmem_addr = (uint8_t *)mmap(NULL, pmem_size, PROT_READ | PROT_WRITE,
				map_flags, fd, 0);

    if (fd != -1)
	close(fd);

    if (pmem_addr == (void *)MAP_FAILED) {
	perror("mmap");
	fatal("Could not mmap!\n");
    }
}

PhysicalMemory::~PhysicalMemory()
{
    if (pmem_addr)
	munmap(pmem_addr, pmem_size);
}

//
// little helper for better prot_* error messages
//
void
PhysicalMemory::prot_access_error(Addr addr, int size, const string &func)
{
    panic("invalid physical memory access!\n"
          "%s: %s(addr=%#x, size=%d) out of range (max=%#x)\n",
	  name(), func, addr, size, pmem_size - 1);
}

void
PhysicalMemory::prot_read(Addr addr, uint8_t *p, int size)
{
    MEM_ADDR_TEST(addr);
    if (addr + size >= pmem_size)
	prot_access_error(addr, size, "prot_read");

    memcpy(p, pmem_addr + addr - base_addr, size);
}

void
PhysicalMemory::prot_write(Addr addr, const uint8_t *p, int size)
{
    MEM_ADDR_TEST(addr, p);
    if (addr + size >= pmem_size)
	prot_access_error(addr, size, "prot_write");

    memcpy(pmem_addr + addr - base_addr, p, size);
}

void
PhysicalMemory::prot_memset(Addr addr, uint8_t val, int size)
{
    MEM_ADDR_TEST(addr);
    if (addr + size >= pmem_size)
	prot_access_error(addr, size, "prot_memset");

    memset(pmem_addr + addr - base_addr, val, size);
}

Fault
PhysicalMemory::read(MemReqPtr &req, uint8_t *data)
{
    MEM_ADDR_TEST(req->paddr);
    off_t offset = req->paddr - base_addr;
    if (offset + req->size >= pmem_size)
	return Machine_Check_Fault;

    memcpy(data, pmem_addr + offset, req->size);
    return No_Fault;
}

Fault
PhysicalMemory::write(MemReqPtr &req, const uint8_t *data)
{
    MEM_ADDR_TEST(req->paddr, data);
    off_t offset = req->paddr - base_addr;
    if (offset + req->size >= pmem_size)
	return Machine_Check_Fault;

    memcpy(pmem_addr + offset, data, req->size);
    return No_Fault;
}

Fault
PhysicalMemory::read(MemReqPtr &req, uint8_t &data)
{
    MEM_BLOCK_TEST(req->paddr, uint64_t);
    off_t offset = req->paddr - base_addr;
    if (offset + sizeof(uint8_t) > pmem_size)
	return Machine_Check_Fault;

    data = *(uint8_t *)(pmem_addr + offset);
    return No_Fault;
}

Fault
PhysicalMemory::read(MemReqPtr &req, uint16_t &data)
{
    MEM_BLOCK_TEST(req->paddr, uint64_t);
    off_t offset = req->paddr - base_addr;
    if (offset + sizeof(uint16_t) > pmem_size)
	return Machine_Check_Fault;

    data = *(uint16_t *)(pmem_addr + offset);
    return No_Fault;
}

Fault
PhysicalMemory::read(MemReqPtr &req, uint32_t &data)
{
    MEM_BLOCK_TEST(req->paddr, uint64_t);
    off_t offset = req->paddr - base_addr;
    if (offset + sizeof(uint32_t) > pmem_size)
	return Machine_Check_Fault;

    data = *(uint32_t *)(pmem_addr + offset);
    return No_Fault;
}

Fault
PhysicalMemory::read(MemReqPtr &req, uint64_t &data)
{
    MEM_BLOCK_TEST(req->paddr, uint64_t);
    off_t offset = req->paddr - base_addr;
    if (offset + sizeof(uint64_t) > pmem_size)
	return Machine_Check_Fault;

    data = *(uint64_t *)(pmem_addr + offset);
    return No_Fault;
}

Fault
PhysicalMemory::write(MemReqPtr &req, uint8_t data)
{
    MEM_BLOCK_TEST(req->paddr, uint64_t, (uint8_t *)&data);
    off_t offset = req->paddr - base_addr;
    if (offset + sizeof(uint8_t) > pmem_size)
	return Machine_Check_Fault;

    *(uint8_t *)(pmem_addr + offset) = data;
    return No_Fault;
}

Fault
PhysicalMemory::write(MemReqPtr &req, uint16_t data)
{
    MEM_BLOCK_TEST(req->paddr, uint64_t, (uint8_t *)&data);
    off_t offset = req->paddr - base_addr;
    if (offset + sizeof(uint16_t) > pmem_size)
	return Machine_Check_Fault;

    *(uint16_t *)(pmem_addr + offset) = data;
    return No_Fault;
}

Fault
PhysicalMemory::write(MemReqPtr &req, uint32_t data)
{
    MEM_BLOCK_TEST(req->paddr, uint64_t, (uint8_t *)&data);
    off_t offset = req->paddr - base_addr;
    if (offset + sizeof(uint32_t) > pmem_size)
	return Machine_Check_Fault;

    *(uint32_t *)(pmem_addr + offset) = data;
    return No_Fault;
}

Fault
PhysicalMemory::write(MemReqPtr &req, uint64_t data)
{
    MEM_BLOCK_TEST(req->paddr, uint64_t, (uint8_t *)&data);
    off_t offset = req->paddr - base_addr;
    if (offset + sizeof(uint64_t) > pmem_size)
	return Machine_Check_Fault;

    *(uint64_t *)(pmem_addr + offset) = data;
    return No_Fault;
}

void
PhysicalMemory::serialize(ostream &os)
{
    string filename = Checkpoint::dir() + name() + ".physmem";

    SERIALIZE_SCALAR(pmem_size);
    SERIALIZE_SCALAR(filename);

    // write memory file
    int fd = creat(filename.c_str(), 0664);
    if (fd < 0) {
	perror("creat");
	fatal("Can't open physical memory checkpoint file '%s'\n", filename);
    }

    if (::write(fd, pmem_addr, pmem_size) != pmem_size) {
	fatal("Write failed on physical memory checkpoint file '%s'\n",
	      filename);
    }

    close(fd);
}

void
PhysicalMemory::unserialize(Checkpoint *cp, const string &section)
{
    // unmap file that was mmaped in the constructor
    munmap(pmem_addr, pmem_size);

    string filename;

    UNSERIALIZE_SCALAR(pmem_size);
    UNSERIALIZE_SCALAR(filename);

    // mmap memoryfile
    int fd = open(filename.c_str(), O_RDONLY);
    if (fd < 0) {
	perror("open");
	fatal("Can't open physical memory checkpoint file '%s'", filename);
    }

    pmem_addr = (uint8_t *)mmap(NULL, pmem_size,
				PROT_READ | PROT_WRITE, MAP_PRIVATE, fd, 0);

    close(fd);

    if (pmem_addr == (void *)MAP_FAILED) {
	perror("mmap");
	fatal("Could not mmap physical memory!\n");
    }
}

#ifdef FULL_SYSTEM
BEGIN_DECLARE_SIM_OBJECT_PARAMS(PhysicalMemory)

    Param<string> file;
    SimObjectParam<MemoryController *> mmu;
    Param<Addr> addr;
    Param<unsigned> size;

END_DECLARE_SIM_OBJECT_PARAMS(PhysicalMemory)

BEGIN_INIT_SIM_OBJECT_PARAMS(PhysicalMemory)

    INIT_PARAM_DFLT(file, "memory mapped file", ""),
    INIT_PARAM(mmu, "Memory Controller"),
    INIT_PARAM(addr, "Device Address"),
    INIT_PARAM(size, "memory size")

END_INIT_SIM_OBJECT_PARAMS(PhysicalMemory)

CREATE_SIM_OBJECT(PhysicalMemory)
{
    return new PhysicalMemory(getInstanceName(), addr, size, mmu, file);
}

REGISTER_SIM_OBJECT("PhysicalMemory", PhysicalMemory)

#endif // FULL_SYSTEM
