/* $Id$ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

#include <assert.h>

#include "base/inifile.hh"
#include "base/misc.hh"
#include "sim/builder.hh"
#include "sim/configfile.hh"
#include "sim/config_node.hh"
#include "sim/host.hh"
#include "sim/sim_object.hh"
#include "sim/universe.hh"

using namespace std;


ostream &
builderStream()
{
    static ofstream file;
    static ostream *stream = NULL;

    if (!stream) {
	if (!outputDirectory.empty()) {
	    string filename = outputDirectory + "builder.txt";
	    file.open(filename.c_str());
	    stream = &file;
	} else {
	    stream = outputStream;
	}
    }

    return *stream;
}

SimObjectBuilder::SimObjectBuilder(const string &_configClass,
				   const string &_instanceName,
				   ConfigNode *_configNode,
				   const string &_simObjClassName)
    : ParamContext(_configClass, true),
      instanceName(_instanceName),
      configNode(_configNode),
      simObjClassName(_simObjClassName)
{
}

SimObjectBuilder::~SimObjectBuilder()
{
}

///////////////////////////////////////////
//
// SimObjectBuilder member definitions
//
///////////////////////////////////////////

// override ParamContext::parseParams() to check params based on
// instance name first.  If not found, then check based on iniSection
// (as in default ParamContext implementation).
void
SimObjectBuilder::parseParams(IniFile &iniFile)
{
    iniFilePtr = &iniFile;	// set object member

    ParamList::iterator i;

    for (i = paramList->begin(); i != paramList->end(); ++i) {
	string string_value;

	if (iniFile.findDefault(instanceName, (*i)->name, string_value)) {
	    (*i)->parse(string_value);
	}
	else if (iniFile.findDefault(iniSection, (*i)->name, string_value)) {
	    (*i)->parse(string_value);
	}
    }
}


void
SimObjectBuilder::printErrorProlog(ostream &os)
{
    os << "Error creating object '" << getInstanceName()
       << "' of type '" << simObjClassName
       << "', section '" << iniSection << "':" << endl;
}


////////////////////////////////////////////////////////////////////////
//
// SimObjectClass member definitions
//
////////////////////////////////////////////////////////////////////////

// Map of class names to SimObjectBuilder creation functions.  Need to
// make this a pointer so we can force initialization on the first
// reference; otherwise, some SimObjectClass constructors may be invoked
// before the classMap constructor.
map<string,SimObjectClass::CreateFunc> *SimObjectClass::classMap = NULL;

// SimObjectClass constructor: add mapping to classMap
SimObjectClass::SimObjectClass(const string &className, CreateFunc createFunc)
{
    if (classMap == NULL)
	classMap = new map<string,SimObjectClass::CreateFunc>();

    if ((*classMap)[className])
    {
	cerr << "Error: simulation object class " << className << " redefined"
	     << endl;
	fatal("");
    }

    // add className --> createFunc to class map
    (*classMap)[className] = createFunc;
}


//
//
SimObject *
SimObjectClass::createObject(IniFile &configDB,
			     const string &configClassName,
			     const string &objName,
			     ConfigNode *configNode)
{
    // find simulation object class name from configuration class
    // (specified by 'type=' parameter)
    string simObjClassName;

    if (!configNode->find("type", simObjClassName)) {
	cerr << "Configuration class '" << configClassName << "' not found."
	     << endl;
	abort();
    }

    // look up className to get appropriate createFunc
    if (classMap->find(simObjClassName) == classMap->end()) {
	cerr << "Simulator object class '" << simObjClassName << "' not found."
	     << endl;
	abort();
    }

    CreateFunc createFunc = (*classMap)[simObjClassName];

    // call createFunc with config hierarchy node to get object
    // builder instance (context with parameters for object creation)
    SimObjectBuilder *objectBuilder = (*createFunc)(configClassName,
						    objName, configNode,
						    simObjClassName);

    assert(objectBuilder != NULL);

    // parse all parameters in context to generate parameter values
    objectBuilder->parseParams(configDB);

    // now create the actual simulation object
    SimObject *object = objectBuilder->create();

    assert(object != NULL);

    // echo object parameters to stats file (for documenting the
    // config used to generate the associated stats)
    builderStream() << "[" << object->name() << "]" << endl;
    builderStream() << "type=" << simObjClassName << endl;
    objectBuilder->showParams(builderStream());
    builderStream() << endl;

    // done with the SimObjectBuilder now
    delete objectBuilder;

    return object;
}


//
// static method:
//
void
SimObjectClass::describeAllClasses(ostream &os)
{
    map<string,CreateFunc>::iterator iter;

    for (iter = classMap->begin(); iter != classMap->end(); ++iter) {
	const string &className = iter->first;
	CreateFunc createFunc = iter->second;

	os << "[" << className << "]\n";

	// create dummy object builder just to instantiate parameters
	SimObjectBuilder *objectBuilder = (*createFunc)("", "", NULL, "");

	// now get the object builder to describe ite params
	objectBuilder->describeParams(os);

	os << endl;

	// done with the object builder now
	delete objectBuilder;
    }
}
