/* $Id: s.configfile.cc 1.37 04/04/05 11:00:47-07:00 stever@isabel.reinhardt.house $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <string>
#include <vector>

#include "base/cprintf.hh"
#include "base/inifile.hh"
#include "base/misc.hh"
#include "base/str.hh"
#include "base/trace.hh"
#include "sim/builder.hh"
#include "sim/config_node.hh"
#include "sim/configfile.hh"
#include "sim/sim_object.hh"

using namespace std;

ConfigHierarchy::ConfigHierarchy(IniFile &f)
    : configDB(f), root(NULL)
{
}

void
ConfigHierarchy::build()
{
    root = new Node("universe", "Universe", "", NULL, this);
    root->buildSubtree();
}

//
// build the subtree under current node (recursively)
//
void
ConfigHierarchy::Node::buildSubtree()
{
    DPRINTFR(Config, "buildSubtree: %s\n", name);

    // get 'children=' option
    string children_string;
    if (!find("children", children_string))
	return;

    // parse & create child ConfigHierarchy::Node objects
    if (!parseChildren(children_string))
	fatal("Could not parse string");

    // recursively build subtree under each child
    for (int i = 0; i < children.size(); i++)
	children[i]->buildSubtree();
}


//
// parse 'children=' option string and create corresponding Node
// objects as children of current node
//
// if a 'checkpoint=' is present, add a checkpoint node to the
// child list, and pass down the root name
bool
ConfigHierarchy::Node::parseChildren(const string &str)
{
    string s = str;
    eat_white(s);
    vector<string> v;

    tokenize(v, s, ' ');

#if 0
    string cp;
    if (find(configClass, "checkpoint", cp)) {
	// if a checkpoint is specified, add a node
	string childPath = hierarchyPath;
	if (childPath != "")
	    childPath += ".";
	childPath += "checkpoint";

	children.push_back(new Node("checkpoint", cp, childPath,"",
				    this, myHierarchy));
	// set checkpoint hierarchy to root
	find(cp, "root_category", checkpointPath);
	if (checkpointPath == "")
	    checkpointPath = hierarchyPath;
    }
#endif

    for (int i = 0; i < v.size(); i++) {
	// specification is of the form:
	//    <name>:<configClass>
	//      or
	//    <name>:<configClass>{<count>}

	// used to delimit current substring while parsing
	string::size_type startpos, endpos;

	// find name
	startpos = 0;
	endpos = v[i].find(':', startpos);
	string name = (endpos == string::npos) ?
	    v[i].substr(startpos) : v[i].substr(startpos, endpos - startpos);
	if (endpos == string::npos) {
	    addChild(name, "__unknown");
	    continue;
	}

	// find the configClass
	startpos = endpos + 1;
	endpos = v[i].find('{', startpos);
	string childConfigClass = (endpos == string::npos) ?
	    v[i].substr(startpos) : v[i].substr(startpos, endpos - startpos);

	// if there's a count field, parse it
	if (endpos != string::npos) {
	    startpos = endpos + 1;
	    endpos = v[i].find('}', startpos);
	    if (endpos == string::npos) {
		cerr << "Missing '}' in object specification \""
		     << v[i] << "\"" << endl;
		return false;
	    }

	    // check for junk after '}'
	    if (endpos != (v[i].length() - 1)) {
		cerr << "Junk after '}' in object specification \""
		     << v[i] << "\"" << endl;
		return false;
	    }

	    string count_string = v[i].substr(startpos, endpos - startpos);
	    int count_number;
	    if (!to_number(count_string, count_number)) {
		cerr << "Can't parse count in object specification \""
		     << v[i] << "\"" << endl;
		return false;
	    }

	    for (int j = 0; j < count_number; j++) {
		addChild(csprintf("%s%d", name, j), childConfigClass);
	    }
	} else {
	    addChild(name, childConfigClass);
	}
    }

    return true;
}


const ConfigHierarchy::Node*
ConfigHierarchy::Node::addChild(const string &childName,
				const string &childConfigClass)
{
    // check for duplicate names (not allowed as it makes future
    // references to the name ambiguous)
    if (findChild(childName)) {
	fatal("Duplicate name '%s' in config node '%s'\n",
	      childName, hierarchyPath);
    }

    // extend the hierarchy path
    string childPath = hierarchyPath;
    if (childPath != "")
	childPath += ".";
    childPath += childName;

    DPRINTFR(Config, "addChild: adding child %s of class %s\n",
	     childPath, childConfigClass);

    Node *child = new Node(childName, childConfigClass, childPath,
			   this, myHierarchy);

    children.push_back(child);

    return child;
}


const ConfigHierarchy::Node *
ConfigHierarchy::Node::findChild(const string &name) const
{
    for (int i = 0; i < children.size(); i++) {
	const Node *n = children[i];
	if (n->name == name)
	    return n;
    }

    return NULL;
}

const ConfigHierarchy::Node *
ConfigHierarchy::Node::resolveNode(const string &name, int level) const
{
    const ConfigNode *n;

    vector<string> v;

    tokenize(v, name, '.');

    if ((n = findChild(v[level])) != NULL) {

	// have a match at current level of path

	if (level == v.size() - 1){
	    // this is the last level... we're done
	    return n;
	} else {
	    // look for match at next level
	    return n->resolveNode(name, level + 1);
	}
    }

    if (level == 0 && parent != NULL) {
	// still looking for initial match: recurse up the tree if we can
	return parent->resolveNode(name, 0);
    }

    // nowhere else to look... give up
    return NULL;
}


SimObject *
ConfigHierarchy::Node::resolveSimObject(const string &name) const
{
    const ConfigNode *node = resolveNode(name);

    // check fro config node not found
    if (node == NULL)
	return NULL;

    // if config node exists, but corresponding SimObject has not yet
    // been created, force creation
    if (node->simObject == NULL) {
	ConfigNode *mnode = const_cast<ConfigNode *>(node);
	mnode->createSimObject();
    }

    return node->simObject;
}


void
ConfigHierarchy::createSimObjects()
{
    root->createSimObjects();
}

void
ConfigHierarchy::Node::createSimObject()
{
    // make this safe to call more than once by skipping if object
    // already created
    if (simObject != NULL)
	return;

    // This hierarchy node corresponds to a simulation object iff
    // there is a 'type=' parameter for it
    string sim_obj_type;

    if (find("type", sim_obj_type)) {
	simObject = SimObjectClass::createObject(getConfigDB(), configClass,
						 getFullName(), this);

	if (simObject == NULL) {
	    cerr << "Error creating object." << endl;
	    abort();
	}
    }
}


void
ConfigHierarchy::Node::createSimObjects()
{
    // do a depth-first post-order traversal, creating each hierarchy
    // node's corresponding simulation object as we go
    for (int i = 0; i < children.size(); i++)
	children[i]->createSimObjects();

    // now create this node's object
    createSimObject();
}


void
ConfigHierarchy::dump(ostream &stream) const
{
    dump(stream, root, 0);
}


void
ConfigHierarchy::dump(ostream &stream, const Node *n, int l) const
{
    if (!n) return;
    for (int i = 0; i < l; i++)
	stream << "  ";
    stream << n->name << ":" << n->configClass << "\n";

    for (int i = 0; i < n->children.size(); i++)
	dump(stream, n->children[i], l + 1);
}

bool
ConfigHierarchy::find(const string &configClass, const string &name,
		      string &value) const
{
    return configDB.findDefault(configClass, name, value);
}

void
ConfigHierarchy::unserializeSimObjects()
{
    root->unserialize(NULL, "");
}

// Walks tree in a pre-order traversal and unserializes any
// SimObjects with a checkpoint provided
void
ConfigHierarchy::Node::unserialize(Checkpoint *parentCkpt,
				   const std::string &section)
{
    Checkpoint *cp = parentCkpt; // use parent's checkpoint by default
    string cpSection = section;  // and parent-provided section name

    string checkpointName;

    if (find("checkpoint", checkpointName)) {
	DPRINTFR(Config, "Loading checkpoint dir '%s'\n",
		 checkpointName);
	cp = new Checkpoint(checkpointName, section, this);
	Serializable::unserializeGlobals(cp);
	cpSection = "";
    }

    if (simObject && cp) {
	if (cpSection == "")
	    fatal("Can't have SimObject at root of checkpoint\n");
	DPRINTFR(Config, "Unserializing '%s' from section '%s'\n",
		 simObject->name(), cpSection);
	if(cp->sectionExists(cpSection))
	    simObject->unserialize(cp, cpSection);
	else
	    warn("Not unserializing '%s': no section found in checkpoint.\n",
		 cpSection);
    }

    for (int i = 0; i < children.size(); i++) {
	string childSection = cpSection;
	if (childSection != "")
	    childSection += ".";
	childSection += children[i]->name;
	children[i]->unserialize(cp, childSection);
    }
}
