/* $Id: s.int_stats.cc 1.34 04/01/29 00:33:44-05:00 binkertn@ziff.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

//
//  Interval statistics
//
//
//

#include <string>
#include <sstream>
#include <vector>
#include <iomanip>

#include "base/callback.hh"
#include "sim/int_stats.hh"
#include "sim/sim_events.hh"
#include "sim/sim_exit.hh"
#include "sim/param.hh"

using namespace std;

vector<string> istats_names;

void
dump_istats()
{
    if (istats_names.empty())
	return;

    ccprintf(cerr, "%s", istats_names[0]);
    for (int i = 1; i < istats_names.size(); i++)
	ccprintf(cerr, ", %s", istats_names[i]);

    ccprintf(cerr, "\n");
}

StatsVector::StatsVector()
{}

StatsVector::StatsVector(const StatsVector &sv)
    : stats(sv.stats)
{}

StatsVector::~StatsVector()
{}

void
StatsVector::dump()
{
    ccprintf(cerr, "%d", (uint64_t)stats[0]);
    for (int i = 1; i < stats.size(); i++)
	ccprintf(cerr, ", %d", (uint64_t)stats[i]);

    ccprintf(cerr, "\n");
}

void
StatsVector::load()
{
    int size = istats_names.size();
    stats.resize(size);
    for (int i = 0; i < size; ++i) {
	stat_stat_t *stat = stat_find_stat(sim_sdb, istats_names[i].c_str());
	if (!stat) {
	    warn("can't find stat: \"%s\"", istats_names[i]);
	    continue;
	}

	double val = stat_value_as_double(sim_sdb, stat);
	stats[i] = val;
    }
}

const StatsVector &
StatsVector::operator=(const StatsVector &sv)
{
    if (&sv != this)
	stats = sv.stats;
    return *this;
}

const StatsVector &
StatsVector::operator+=(const StatsVector &sv)
{
    assert(stats.size() == 0 || sv.stats.size() == 0 ||
	   stats.size() == sv.stats.size());

    if (stats.size() > 0 && sv.stats.size() > 0) {
	int num = stats.size();
	for (int i = 0; i < num; i++)
	    stats[i] += sv.stats[i];
    } else if (stats.size() == 0 && sv.stats.size() > 0) {
	stats = sv.stats;
    }

    return *this;
}

const StatsVector &
StatsVector::operator-=(const StatsVector &sv)
{
    assert(sv.stats.size() == 0 || stats.size() == sv.stats.size());

    if (stats.size() > 0 && sv.stats.size() > 0) {
	int num = stats.size();
	for (int i = 0; i < num; i++)
	    stats[i] -= sv.stats[i];
    }

    return *this;
}

StatsVector
operator+(const StatsVector &l, const StatsVector &r)
{
    StatsVector sv = l;
    sv += r;
    return sv;
}

StatsVector
operator-(const StatsVector &l, const StatsVector &r)
{
    StatsVector sv = l;
    sv -= r;
    return sv;
}

//===========================================================================
//
//  Initialization:
//

class StatEvent : public Event
{
  private:
    IntervalStats *istats;

  public:
    StatEvent(IntervalStats *i)
	: Event(&mainEventQueue), istats(i) {
	schedule(0);
    }
    void process() { istats->init(); delete this; }
    virtual const char *description() { return "istat configure"; }
};


IntervalStats::IntervalStats(stat_sdb_t *sdb, unsigned thread,
    const string &stat_file, bool exit_when_done, bool enable_stat_triggers,
    Counter stat_cycle_int, const string &stat_range,
    const vector<string> &statistics)
{
    //  assume we're not valid to start
    valid = false;

    numThreads = thread;
    statDB = sdb;

    if (stat_file.empty())
	return;


    statsFilename = stat_file;
    exitAfterLastInterval = exit_when_done;

    enableTriggers = enable_stat_triggers;
    if (exitAfterLastInterval && enableTriggers) {
	exitAfterLastInterval = false;
	warn("istats: disabling exit_when_done due to enable_stat_triggers");
    }

    if (statistics.empty())
	return;

    statNames = statistics;

    if (stat_cycle_int == 0)
	enableCycleIntervals = false;
    else
	enableCycleIntervals = true;

    cycleRange = stat_range;

    //  If the user specified cycle-interval statistics...
    interval = stat_cycle_int;
    if (interval == 0)
	enableCycleIntervals = false;

    new StatEvent(this);
}

void
IntervalStats::init()
{
    event = new CycleIntStatEvent(this);

    numStats = buildStatList();
    if (numStats == 0)
	return;

    //==================================================================
    //
    //  At this point we have verified the parameters we need for stats...
    //
    valid = true;
    currentOutputLabel = "";

    //
    //  Open the output file
    //
    outfile = new ofstream(statsFilename.c_str());
    if (!outfile) {
	fatal("unable to open output file \"%s\"", statsFilename);
    }

    //
    //  Make room for storing the previous stat values
    //
    lastValue.resize(numStats);

    //
    //  Print the output file header
    //
    *outfile << "# curTick, ";

    for (int i = 0; i < statLabels.size(); i++) {
	if (dumpValue[i])
	    *outfile << statLabels[i] << ", ";
	else
	    *outfile << statLabels[i] << "(delta), ";

	// set up the "last values"
	lastValue[i] = 0;
    }

    *outfile << endl;

    //
    //  Schedule the first invocation of the event
    //
    if (enableCycleIntervals)
	event->schedule(cycleRange.start);
}


void
IntervalStats::dumpStats(Tick now)
{
    if (!valid)
	return;

    *outfile << currentOutputLabel << " : " << now;

    outfile->setf(std::ios::fixed);

    for (int i = 0; i < statList.size(); i++) {
        double new_val = stat_value_as_double(sim_sdb, statList[i]);

        if (dumpValue[i]) {
            *outfile << ", " << std::setw(12)
                     << std::setprecision(3) << new_val;
        } else {
            double delta = new_val - lastValue[i];

            *outfile << ", " << std::setw(12)
                     << std::setprecision(3) << delta;

            lastValue[i] = new_val;
        }
    }

    *outfile << endl;

    outfile->flush();
}

//
//  Parse the vector of strings naming the stats we want, and place pointers
//  to the stat objects into stat_list. Elements of just_dump_value[]
//  are initialized here also.
//
unsigned
IntervalStats::buildStatList()
{
    unsigned num_stats = 0;

    //
    //  Build the vector of stat pointers
    //
    for (int i = 0; i < statNames.size(); ++i) {
	string stat = statNames[i];

	if (stat.size() >= 2) {
	    bool dump_vals = false;

	    //
	    //  Exclamation marks stat where we don't want to print the
	    //  delta... just the value
	    //
	    if (stat[0] == '!') {
		dump_vals = true;
		// remove exclamation
		stat.erase(stat.begin());
	    }

	    //
	    //  Astericks marks stat where we want this stat for all
	    //  threads
	    //
	    if (stat[stat.size() - 1] == '*') {
		// remove astericks
		stat.erase(stat.end() - 1);

		for (int t = 0; t < numThreads; ++t) {
		    ostringstream n;
		    stat_stat_t *p;

		    n << stat << "_" << t;
		    p = stat_find_stat(sim_sdb, n.str().c_str());

		    if (p) {
			statList.push_back(p);
			dumpValue.push_back(dump_vals);
			statLabels.push_back(n.str().c_str());
			++num_stats;
		    } else {
			warn("can't find stat: \"%s\"", n.str());
		    }
		}
	    } else {
		stat_stat_t *p = stat_find_stat(sim_sdb, stat.c_str());

		if (p) {
		    statList.push_back(p);
		    dumpValue.push_back(dump_vals);
		    statLabels.push_back(stat.c_str());
		    ++num_stats;
		} else {
		    warn("can't find stat: \"%s\"", stat);
		}
	    }
	} else {
	    fatal("\"%s\" is an illegal stat name", stat);
	}
    }

    return num_stats;
}


//==========================================================================
//
//  These functions are called with a label to dump stats at the current cycle
//
void
IntervalStats::NewInterval(string label)
{
    if (!enableTriggers)
	return;

    event->schedule(curTick + 1);

    currentOutputLabel = label;
}


void
IntervalStats::StartInterval(string label)
{
    if (!enableTriggers)
	return;

    event->schedule(curTick + 1);

    if (currentOutputLabel != "")
	outputLabelStack.push_back(currentOutputLabel);

    currentOutputLabel = "START:" + label;
}


void
IntervalStats::EndInterval(string label)
{
    if (!enableTriggers)
	return;

    event->schedule(curTick + 1);

    if (outputLabelStack.size() != 0) {
	string l = outputLabelStack.back();
	outputLabelStack.pop_back();

	if (l != label)
	    cerr << "warning: istat popped label " << l <<
		", expecting label " << label << endl;

	currentOutputLabel = "END" + label;
    } else {
	currentOutputLabel = "";
    }
}


//==========================================================================
//
//  Event Constructor
//
IntervalStats::CycleIntStatEvent::CycleIntStatEvent(IntervalStats *sp) :
    Event(&mainEventQueue)
{
    stats = sp;
}


void
IntervalStats::CycleIntStatEvent::process()
{
    if (!stats->isValid()) {
	delete this;
    } else {

	//
	//  Print the stats
	//
	stats->dumpStats(curTick);

	if (stats->getInterval() > 0) {
	    //
	    //  Optionally, Reschedule the event
	    //
	    Counter next = curTick + stats->getInterval();
	    if (next > stats->getRange()) {
		//
		//  Done with this event...
		//
		stats->closeOutfile();

		if (stats->exitWhenDone())
		    new SimExitEvent("done with interval stats");

		delete this;
	    } else {
		//  re-schedule
		schedule(next);
	    }
	}
    }
}

const char *
IntervalStats::CycleIntStatEvent::description()
{
    return "interval stat event";
}


//
//  Parameters
//

class IntervalStatParamContext : public ParamContext
{
  public:
    IntervalStatParamContext(const string &_iniSection)
	: ParamContext(_iniSection)
    {
    }
    void checkParams();
};


IntervalStatParamContext intervalStatParams("istat");

Param<string> stat_file(&intervalStatParams,
			"file",
			"output file for interval statistics");

Param<bool> exit_when_done(&intervalStatParams,
			   "exit_when_done",
			   "exit after last cycle interval");

Param<bool> enable_stat_triggers(&intervalStatParams,
			   "enable_stat_triggers",
			   "enable statistic dump instruction triggers");

Param<Counter> stat_cycle_int(&intervalStatParams,
				"cycle",
				"dump statistics every 'n' cycles");

Param<string> stat_range(&intervalStatParams, "range",
			 "cycle range to dump", "0:0");

VectorParam<string> statistics(&intervalStatParams,
			       "stats",
			       "statistics to dump");

#if 0
Param<Counter> stat_inst_int(&intervalStatParams,
			       "inst",
			       "dump statistics every 'n' instructions");
#endif

IntervalStats *istats = NULL;

void
IntervalStatParamContext::checkParams()
{
    if (statistics.isValid()) {
	string filename = stat_file;
	if (!filename.empty() && !outputDirectory.empty())
		filename = outputDirectory + filename;

	istats_names = statistics;
	istats = new IntervalStats(sim_sdb, 1, stat_file, exit_when_done,
				   enable_stat_triggers, stat_cycle_int,
				   stat_range, statistics);
    }
}

