/* $Id: s.sim_events.cc 1.38 04/05/04 17:00:59-04:00 binkertn@ziff.eecs.umich.edu $ */

/*
 * Copyright (c) 2003
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, and Lisa Hsu.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

#include <string>

#include "base/callback.hh"
#include "base/hostinfo.hh"
#include "sim/eventq.hh"
#include "sim/param.hh"
#include "sim/sim_events.hh"
#include "sim/sim_exit.hh"
#include "sim/sim_init.hh"
#include "sim/stats.hh"

using namespace std;

//
// handle termination event
//
void
SimExitEvent::process()
{
    // This event does not autodelete because exitNow may be called,
    // and the function will never be allowed to finish.
    if (theQueue() == &mainEventQueue) {
	string _cause = cause;
	int _code = code;
	delete this;
        exitNow(_cause, _code);
    } else {
	new SimExitEvent(cause, code);
	delete this;
    }
}


const char *
SimExitEvent::description()
{
    return "simulation termination";
}

//
// constructor: automatically schedules at specified time
//
CountedExitEvent::CountedExitEvent(EventQueue *q, const std::string &_cause,
				   Tick _when, int &_downCounter)
    : Event(q, Sim_Exit_Pri),
      cause(_cause),
      downCounter(_downCounter)
{
    // catch stupid mistakes
    assert(downCounter > 0);

    schedule(_when);
}


//
// handle termination event
//
void
CountedExitEvent::process()
{
    if (--downCounter == 0) {
	new SimExitEvent(cause, 0);
    }
}


const char *
CountedExitEvent::description()
{
    return "counted exit";
}

#ifdef CHECK_SWAP_CYCLES
new CheckSwapEvent(&mainEventQueue, CHECK_SWAP_CYCLES);
#endif

void
CheckSwapEvent::process()
{
    /*  Check the amount of free swap space  */
    long swap;

    /*  returns free swap in KBytes  */
    swap = procInfo("/proc/meminfo", "SwapFree:");

    if (swap < 1000)
	ccprintf(cerr, "\a\a\aWarning! Swap space is low (%d)\n", swap);

    if (swap < 100) {
	cerr << "\a\aAborting Simulation! Inadequate swap space!\n\n";
	new SimExitEvent("Lack of swap space");
    }

    schedule(curTick + interval);
}

const char *
CheckSwapEvent::description()
{
    return "check swap";
}


///////////////////////////////////////////////////
//
// Simulation termination parameters
//
///////////////////////////////////////////////////

class TermParamContext : public ParamContext
{
  public:
    TermParamContext(const string &_iniSection)
	: ParamContext(_iniSection) {}
    void checkParams();
};

TermParamContext simTerminationParams("max");

Param<Tick> max_cycle(&simTerminationParams, "cycle",
			"maximum number of cycles to execute");

void
TermParamContext::checkParams()
{
    // if a max cycle count was specified, put a termination event on
    // the event queue at that point
    if (max_cycle.isValid())
	new SimExitEvent(max_cycle, "reached maximum cycle count");
}

//
// Progress event: print out cycle every so often so we know we're
// making forward progress.
//
class ProgressEvent : public Event
{
  protected:
    Tick interval;

  public:
    ProgressEvent(EventQueue *q, Tick interval);

    void process();	// process event
    virtual const char *description();
};

//
// constructor: schedule at specified time
//
ProgressEvent::ProgressEvent(EventQueue *q, Tick _interval)
    : Event(q), interval(_interval)
{
    schedule(curTick + interval);
}

//
// handle progress event: print message and reschedule
//
void
ProgressEvent::process()
{
    DPRINTFN("ProgressEvent\n");
    // reschedule for next interval
    schedule(curTick + interval);
}


const char *
ProgressEvent::description()
{
    return "progress message";
}

/////////
//
// Periodic progress message support: print out a message every n
// cycles so we know we're making forward progress.
//
/////////

// Parameter space for execution address tracing options.  Derive
// from ParamContext so we can override checkParams() function.
class ProgressParamContext : public ParamContext
{
  public:
    ProgressParamContext(const string &_iniSection)
	: ParamContext(_iniSection) {}
    void checkParams();
};

ProgressParamContext progessMessageParams("progress");

Param<Tick> progress_interval(&progessMessageParams, "cycle",
				"cycle interval for progress messages");

namespace {
    struct SetupProgress : public Callback
    {
	Tick interval;
	SetupProgress(Tick tick) : interval(tick) {}

	virtual void process()
	{
	    new ProgressEvent(&mainEventQueue, interval);
	    delete this;
	}
    };
}

/* check execute options */
void
ProgressParamContext::checkParams()
{
    if (progress_interval.isValid())
	registerInitCallback(new SetupProgress(progress_interval));
}
