#! /usr/bin/python

# This script automates about 90% of generating .odesc files from the
# existing .cc files.  There's still some manual tweaking required
# (particularly for enum types), so we should just keep the .cc and
# .odesc files in sync and not run this again.  I used this initially
# to generate the .odesc files and am committing it mostly for
# posterity.

import re
import sys
import os

def extract_bases(dummy, dirpath, namelist):
    i = 0
    while i < len(namelist):
        if namelist[i] == "SCCS":
            del namelist[i]
        else:
            i = i + 1
    for name in namelist:
        if name.endswith('.hh'):
            extract_base(name, dirpath)

base_re_string = r'''
(?:class|struct) \s* (\w+) \s* : \s* public \s* (\w+)
'''

base_re = re.compile(base_re_string, re.VERBOSE)

bases = {}

def extract_base(filename, path):
    filename = os.path.join(path, filename)
    source = open(filename, "r").read()

    for base_match in base_re.finditer(source):
        bases[base_match.group(1)] = base_match.group(2)

def extract_odescs(dummy, dirpath, namelist):
    i = 0
    while i < len(namelist):
        if namelist[i] == "SCCS":
            del namelist[i]
        else:
            i = i + 1
    for name in namelist:
        if name.endswith('.cc'):
            extract_odesc(name, dirpath)

def_re_string = r'''
DEFINE_SIM_OBJECT_CLASS_NAME\( \s* \"([^"]*)\" \s* , \s* (\w+) \s* \)
'''

reg_re_string = r'''
REGISTER_SIM_OBJECT\( \s* \"([^"]*)\" \s* , \s* (\w+) \s* \)
'''

sec_re_template = r'''
BEGIN_(%s)_SIM_OBJECT_PARAMS\( \s* (\w+) \s* \)
(.*)
END_\1_SIM_OBJECT_PARAMS\( \s* \2 \s* \)
'''

decl_re_string = sec_re_template % 'DECLARE'
init_re_string = sec_re_template % 'INIT'

decl_re = re.compile(decl_re_string, re.DOTALL | re.VERBOSE)
init_re = re.compile(init_re_string, re.DOTALL | re.VERBOSE)
def_re  = re.compile(def_re_string,  re.DOTALL | re.VERBOSE)
reg_re  = re.compile(reg_re_string,  re.DOTALL | re.VERBOSE)

re_items = {
'id': r'([\w_:]+)',
'str': r'(\"[^"]*\"(?:\s*\"[^"]*\")*)',
'dflt': r'([\w_&:()<>,]+|\"[^"]*\")'
}

for k, v in re_items.items():
    re_items[k] = r'\s*' + v + r'\s*'

init_item_re = \
    re.compile(r'INIT_PARAM\(%(id)s,%(str)s\)' % re_items)
init_dflt_item_re = \
    re.compile(r'INIT_PARAM_DFLT\(%(id)s,%(str)s,%(dflt)s\)' % re_items)
init_enum_item_re = \
    re.compile(r'INIT_ENUM_PARAM\(%(id)s,%(str)s,%(id)s\)' % re_items)
init_enum_dflt_item_re = \
    re.compile(r'INIT_ENUM_PARAM_DFLT\(%(id)s,%(str)s,%(id)s,%(dflt)s\)' % re_items)

def gen_odesc(path, name, param_decls, param_inits):
    odescname = os.path.join(path, name + ".odesc1")

    odesc = open(odescname, "w")

    base = bases.get(name)
    if not base:
        print "No base found for", name
        base = "SimObject"

    print >> odesc, "OBJECT:"
    print >> odesc, "    %s(%s)" % (name, base)

    param_list = re.findall(r'(?s)\s*(\w*Param)<([\w: *]+)>\s*([\w_]+);', param_decls)

    params = {}

    class Param:
        def __init__(self, pclass, ptype):
            self.pclass = pclass
            self.ptype = ptype

    for pclass, ptype, name in param_list:
        params[name] = Param(pclass, ptype)

    init_list = init_item_re.findall(param_inits)

    init_dflt_list = init_dflt_item_re.findall(param_inits)

    init_enum_list = init_enum_item_re.findall(param_inits)

    init_enum_dflt_list = init_enum_dflt_item_re.findall(param_inits)

    for name, desc in init_list:
        params[name].desc = desc

    for name, desc, dflt in init_dflt_list:
        params[name].desc = desc
        params[name].dflt = dflt

    for name, desc, map in init_enum_list:
        params[name].desc = desc
        params[name].map = map

    for name, desc, map, dflt in init_enum_dflt_list:
        params[name].desc = desc
        params[name].map = map
        params[name].dflt = dflt

    sparams = params.keys()
    sparams.sort()

    for name in sparams:
        p = params[name]
        if not hasattr(p, 'pclass'):
            print "No declaration found for param '%s'" % name
            sys.exit(1)
        if not hasattr(p, 'desc'):
            print "No initialization found for param '%s'" % name
            sys.exit(1)

    type_map = {
        'bool' : 'Bool',
        'int' : 'Int', 'unsigned' : 'Int', 'unsigned int' : 'Int',
        'uint32_t' : 'Int', 'time_t' : 'Int',
        'string' : 'String', 'std::string' : 'String'
        }

    print >> odesc, "PARAMS:"

    for name in sparams:
        p = params[name]
        if p.pclass.startswith('SimObject'):
            p.pclass = p.pclass[len('SimObject'):]
            # strip trailing '*' (C++ pointer)
            p.ptype = re.sub(r'\s*\*\s*$', '', p.ptype)
        if p.pclass.startswith('SimpleEnum'):
            p.pclass = p.pclass[len('Simple'):]
        if type_map.has_key(p.ptype):
            p.ptype = type_map[p.ptype]
        if hasattr(p, 'dflt'):
            if p.dflt == '&defaultHierParams':
            # these are going to require some manual effort...
                print "%s: removing default '%s' for %s" \
                            % (odescname, p.dflt, name)
                del p.dflt
        if p.pclass.find('Enum') != -1:
            if not hasattr(p, 'map'):
                print "Map field in non-enum class '%s'" % name
                sys.exit(1)
            # these are going to require some manual effort...
            print "%s: enum %s" % (odescname, name)
        else:
            if hasattr(p, 'map'):
                print "Map field in non-enum class '%s'" % name
                sys.exit(1)

        s = "    %s = %s(%s, %s" % (name, p.pclass, p.ptype, p.desc)
        if hasattr(p, 'map'): s += ', ' + p.map
        if hasattr(p, 'dflt'): s += ', ' + p.dflt
        s += ')'
        print >> odesc, s

    odesc.close()

def extract_odesc(filename, path = "."):
    filename = os.path.join(path, filename)
    source = open(filename, "r").read()

    for reg_match, decl_match, init_match in zip(reg_re.finditer(source),
                                                 decl_re.finditer(source),
                                                 init_re.finditer(source)):
        name = reg_match.group(1)
        cpp_name = reg_match.group(2)
        if name != cpp_name:
            print "REG/C++ name mismatch:", name, "!=", cpp_name
            exit

        decl_name = decl_match.group(2)
        init_name = init_match.group(2)
        if name != decl_name:
            print "REG/DECL name mismatch:", name, "!=", decl_name
            exit
        if name != init_name:
            print "REG/INIT name mismatch:", name, "!=", init_name
            exit


        gen_odesc(path, name, decl_match.group(3), init_match.group(3))

    for def_match in def_re.finditer(source):
        name = def_match.group(1)
        name = def_match.group(2)
        if name != name:
            print "Warning:", name, "!=", name
        gen_odesc(path, name, "", "")

filename = sys.argv[1]

if (os.path.isdir(filename)):
    os.path.walk(filename, extract_bases, 1)
    os.path.walk(filename, extract_odescs, 1)
else:
    extract_odesc(filename)

