import os
import re

#####################################################################
#
# Object description loading.
#
# The final step is to define the classes corresponding to M5 objects
# and their parameters.  These classes are described in .od ("object
# description") files in the source tree.  This code walks the tree to
# find those files and loads up the descriptions (by evaluating them
# in pieces as Python code).
#
#
# Because SimObject classes inherit from other SimObject classes, and
# can use arbitrary other SimObject classes as parameter types, we
# have to do this in three steps:
#
# 1. Walk the tree to find all the .od files.  Note that the base
# of the filename *must* match the class name.  This step builds a
# mapping from class names to file paths.
#
# 2. Start generating empty class definitions (via def_class()) using
# the OBJECT field of the .od files to determine inheritance.
# def_class() recurses on demand to define needed base classes before
# derived classes.
#
# 3. Now that all of the classes are defined, go through the .od
# files one more time loading the parameter descriptions.
#
#####################################################################

# dictionary: maps object names to file paths
odesc_file = {}

# dictionary: maps object names to boolean flag indicating whether
# class definition was loaded yet.  Since SimObject is defined in
# m5.config.py, count it as loaded.
odesc_loaded = { 'SimObject': True }

# Find odesc files in namelist and initialize odesc_file and
# odesc_loaded dictionaries.  Called via os.path.walk() (see below).
def find_odescs(process, dirpath, namelist):
    # Prune out SCCS directories so we don't process s.*.od files.
    i = 0
    while i < len(namelist):
        if namelist[i] == "SCCS":
            del namelist[i]
        else:
            i = i + 1
    # Find .od files and record them.
    for name in namelist:
        if name.endswith('.od'):
            objname = name[:name.rindex('.od')]
            path = os.path.join(dirpath, name)
            if odesc_file.has_key(objname):
                print "Warning: duplicate object names:", \
                      odesc_file[objname], path
            odesc_file[objname] = path
            odesc_loaded[objname] = False


# Regular expression string for parsing .od files.
file_re_string = r'''
^OBJECT: \s* (\w+) \s* \( \s* (\w+) \s* \)
\s*
^PARAMS: \s*\n ( (\s+.*\n)* )
'''

# Compiled regular expression object.
file_re = re.compile(file_re_string, re.MULTILINE | re.VERBOSE)

# .od file parsing function.  Takes a filename and returns tuple of
# object name, object base, and parameter description section.
def parse_file(path):
    f = open(path, 'r').read()
    m = file_re.search(f)
    if not m:
        print "Can't parse", path
        sys.exit(1)
    return (m.group(1), m.group(2), m.group(3))

# Define SimObject class based on description in specified filename.
# Class itself is empty except for _name attribute; parameter
# descriptions will be loaded later.  Will recurse to define base
# classes as needed before defining specified class.
def def_class(path):
    # load & parse file
    (obj, parent, params) = parse_file(path)
    # check to see if base class is defined yet; define it if not
    if not odesc_loaded.has_key(parent):
        print "No .od file found for", parent
        sys.exit(1)
    if not odesc_loaded[parent]:
        def_class(odesc_file[parent])
    # define the class.  The _name attribute of the class lets us
    # track the actual SimObject class name even when we derive new
    # subclasses in scripts (to provide new parameter value settings).
    print "class %s(%s): _name = '%s'" % (obj, parent, obj)
    # mark this file as loaded
    odesc_loaded[obj] = True

# Munge an arbitrary Python code string to get it to execute (mostly
# dealing with indentation).  Stolen from isa_parser.py... see
# comments there for a more detailed description.
def fixPythonIndentation(s):
    # get rid of blank lines first
    s = re.sub(r'(?m)^\s*\n', '', s);
    if (s != '' and re.match(r'[ \t]', s[0])):
        s = 'if 1:\n' + s
    return s

# Load parameter descriptions from .od file.  Object class must
# already be defined.
def def_params(path):
    # load & parse file
    (obj_name, parent_name, param_code) = parse_file(path)
    # initialize param dict
    param_dict = {}
    # execute parameter descriptions.
    print "%s.init_params('''%s''', globals())" % \
          (obj_name, fixPythonIndentation(param_code))
#    try:
#        # "in globals(), param_dict" makes exec use the current
#        # globals as the global namespace (so all of the Param
#        # etc. objects are visible) and param_dict as the local
#        # namespace (so the newly defined parameter variables will be
#        # entered into param_dict).
#        exec fixPythonIndentation(param_code) in globals(), param_dict
#    except Exception, exc:
#        print "Param error in %s:" % path, exc
#        return
#    # Convert object name string to Python class object
#    obj = eval(obj_name)
#    # Set the object's parameter description dictionary (see MetaConfigNode).
#    obj.set_param_dict(param_dict)


# Walk directory tree to find .od files.
# Someday we'll have to make the root path an argument instead of
# hard-coding it.  For now the assumption is you're running this in
# util/config.
root = '/z/stever/bk/m5'
os.path.walk(root, find_odescs, None)

print '''
# DO NOT EDIT.
# This file is automatically generated from load_odescs.py.
# To update, run 'python load_odescs.py > m5odescs.py'.
'''
print 'from m5config import *'

# Iterate through file dictionary and define classes.
for objname, path in odesc_file.iteritems():
    if not odesc_loaded[objname]:
        def_class(path)

sim_object_list = odesc_loaded.keys()
sim_object_list.sort()

# Iterate through files again and load parameters.
for path in odesc_file.itervalues():
    def_params(path)

