#!/usr/bin/env python

from __future__ import division
import copy, math, operator, re, string, sys, types

collections = []

flags_init      = 0x00000001
flags_printable = 0x00000002
flags_total     = 0x00000010
flags_pdf       = 0x00000020
flags_cdf       = 0x00000040
flags_dist      = 0x00000080
flags_nozero    = 0x00000100
flags_nonan     = 0x00000200

descriptions = False
display_all = False

def issequence(t):
    return isinstance(t, types.TupleType) or isinstance(t, types.ListType)


def total(f):
    if isinstance(f, FormulaStat):
        v = f.value
    else:
        v = f

    f = FormulaStat()
    if issequence(v):
        f.value = reduce(operator.add, v)
    else:
        f.value = v

    return f

def unaryop(op, f):
    if isinstance(f, FormulaStat):
        v = f.value
    else:
        v = f

    if issequence(v):
        return map(op, v)
    else:
        return op(v)

def zerodiv(lv, rv):
    if rv == 0.0:
        return 0.0
    else:
        return operator.truediv(lv, rv)

def wrapop(op, lv, rv):
    if isinstance(lv, str):
        return lv

    if isinstance(rv, str):
        return rv

    return op(lv, rv)

def binaryop(op, lf, rf):
    if isinstance(lf, FormulaStat):
        lv = lf.value
    else:
        lv = lf

    if isinstance(rf, FormulaStat):
        rv = rf.value
    else:
        rv = rf

    if issequence(lv):
        if issequence(rv):
            return map(lambda x, y: binaryop(op, x, y), lv, rv)
        else:
            return map(lambda x, y: binaryop(op, x, y), lv, [rv]*len(lv))
    else:
        if issequence(rv):
            return map(lambda x, y: binaryop(op, x, y), [lv]*len(rv), rv)
        else:
            return wrapop(op, lv, rv)

def sums(x, y):
    if issequence(x):
        return map(lambda x, y: x + y, x, y)
    else:
        return x + y

def alltrue(list):
    return reduce(lambda x, y: x and y, list)

def allfalse(list):
    return not reduce(lambda x, y: x or y, list)

def enumerate(list):
    return map(None, range(len(list)), list)

def cmp(a, b):
    if a < b:
        return -1
    elif a == b:
        return 0
    else:
        return 1

def statcmp(a, b):
    v1 = a.split('.')
    v2 = b.split('.')

    last = min(len(v1), len(v2)) - 1
    for i,j in zip(v1[0:last], v2[0:last]):
        if i != j:
            return cmp(i, j)

    # Special compare for last element.
    if len(v1) == len(v2):
        return cmp(v1[last], v2[last])
    else:
        return cmp(len(v1), len(v2))

class Value:
    def __init__(self, value, precision, percent = False):
        self.value = value
        self.precision = precision
        self.percent = percent
    def __str__(self):
        if isinstance(self.value, str):
            if self.value.lower() == 'nan':
                value = 'NaN'
            if self.value.lower() == 'inf':
                value = 'Inf'
        else:
            if self.precision >= 0:
                format = "%%.%df" % self.precision
            elif self.value == 0.0:
                format = "%.0f"
            elif self.value % 1.0 == 0.0:
                format = "%.0f"
            else:
                format = "%f"
            value = self.value
            if self.percent:
                value = value * 100.0
            value = format % value

        if self.percent:
            value = value + "%"

        return value

class Print:
    def __init__(self, **vals):
        self.__dict__.update(vals)

    def __str__(self):
        value = Value(self.value, self.precision)
        pdf = ''
        cdf = ''
        if self.__dict__.has_key('pdf'):
            pdf = Value(self.pdf, 2, True)
        if self.__dict__.has_key('cdf'):
            cdf = Value(self.cdf, 2, True)

        output = "%-40s %12s %8s %8s" % (self.name, value, pdf, cdf)

        if descriptions and self.__dict__.has_key('desc') and self.desc:
            output = "%s # %s" % (output, self.desc)

        return output

    def doprint(self):
        if display_all:
            return True
        if self.value == 0.0 and (self.flags & flags_nozero):
            return False
        if isinstance(self.value, str):
            if self.value == 'NaN' and (self.flags & flags_nonan):
                return False
        return True

    def display(self):
        if self.doprint():
            print self

class Statistic:
    def __init__(self, name, desc, **vals):
        self.name = name
        self.desc = desc
        self.__dict__.update(vals)
        if self.__dict__.has_key('subnames'):
            self.subnames = list(self.subnames)
        if self.__dict__.has_key('subdescs'):
            self.subdescs = list(self.subdescs)
        if self.__dict__.has_key('ysubnames'):
            self.ysubnames = list(self.ysubnames)

    def zero(self):
        return False
    def __ne__(self, other):
        return not (self == other)

class FormulaStat:
    def __add__(self, other):
        f = FormulaStat()
        f.value = binaryop(operator.add, self, other)
        return f
    def __sub__(self, other):
        f = FormulaStat()
        f.value = binaryop(operator.sub, self, other)
        return f
    def __mul__(self, other):
        f = FormulaStat()
        f.value = binaryop(operator.mul, self, other)
        return f
    def __truediv__(self, other):
        f = FormulaStat()
        f.value = binaryop(zerodiv, self, other)
        return f
    def __mod__(self, other):
        f = FormulaStat()
        f.value = binaryop(operator.mod, self, other)
        return f
    def __radd__(self, other):
        f = FormulaStat()
        f.value = binaryop(operator.add, other, self)
        return f
    def __rsub__(self, other):
        f = FormulaStat()
        f.value = binaryop(operator.sub, other, self)
        return f
    def __rmul__(self, other):
        f = FormulaStat()
        f.value = binaryop(operator.mul, other, self)
        return f
    def __rtruediv__(self, other):
        f = FormulaStat()
        f.value = binaryop(zerodiv, other, self)
        return f
    def __rmod__(self, other):
        f = FormulaStat()
        f.value = binaryop(operator.mod, other, self)
        return f
    def __neg__(self):
        f = FormulaStat()
        f.value = unaryop(operator.neg, self)
        return f
    def __getitem__(self, idx):
        f = FormulaStat()
        try:
            f.value = self.value[idx]
        except TypeError:
            if idx != 0:
                raise
            f.value = self.value
        return f

class Scalar(Statistic,FormulaStat):
    def display(self):
        p = Print()
        p.name = self.name
        p.desc = self.desc
        p.value = self.value
        p.flags = self.flags
        p.precision = self.precision
        if display_all or (self.flags & flags_printable):
            p.display()

    def comparable(self, other):
        return self.name == other.name

    def __eq__(self, other):
        return self.value == other.value

    def __isub__(self, other):
        self.value -= other.value
        return self

    def __iadd__(self, other):
        self.value += other.value
        return self

    def __itruediv__(self, other):
        if not other:
            return self
        self.value /= other
        return self
    
class VectorDisplay:
    def display(self):
        p = Print()
        p.flags = self.flags
        p.precision = self.precision

        if issequence(self.value):
            if not len(self.value):
                return
            
            mytotal = reduce(lambda x,y: float(x) + float(y), self.value)
            mycdf = 0.0

            value = self.value

            if display_all:
                subnames = [ '[%d]' % i for i in range(len(value)) ]
            else:
                subnames = [''] * len(value)

            if self.__dict__.has_key('subnames'):
                for i,each in enumerate(self.subnames):
                    if len(each) > 0:
                        subnames[i] = '.%s' % each

            subdescs = [self.desc]*len(value)
            if self.__dict__.has_key('subdescs'):
                for i in xrange(min(len(value), len(self.subdescs))):
                    subdescs[i] = self.subdescs[i]

            for val,sname,sdesc in map(None, value, subnames, subdescs):
                if mytotal > 0.0:
                    mypdf = float(val) / float(mytotal)
                    mycdf += mypdf
                    if (self.flags & flags_pdf):
                        p.pdf = mypdf
                        p.cdf = mycdf

                if len(sname) == 0:
                    continue

                p.name = self.name + sname
                p.desc = sdesc
                p.value = val
                p.display()

            if (self.flags & flags_total):
                if (p.__dict__.has_key('pdf')): del p.__dict__['pdf']
                if (p.__dict__.has_key('cdf')): del p.__dict__['cdf']
                p.name = self.name + '.total'
                p.desc = self.desc
                p.value = mytotal
                p.display()

        else:
            p.name = self.name
            p.desc = self.desc
            p.value = self.value
            p.display()

class Vector(Statistic,FormulaStat):
    def display(self):
        if not display_all and not (self.flags & flags_printable):
            return

        d = VectorDisplay()
        d.__dict__.update(self.__dict__)
        d.display()

    def comparable(self, other):
        return self.name == other.name and \
               len(self.value) == len(other.value)

    def __eq__(self, other):
        if issequence(self.value) != issequence(other.value):
            return false

        if issequence(self.value):
            if len(self.value) != len(other.value):
                return False
            else:
                for v1,v2 in zip(self.value, other.value):
                    if v1 != v2:
                        return False
                return True
        else:
            return self.value == other.value

    def __isub__(self, other):
        self.value = binaryop(operator.sub, self.value, other.value)
        return self

    def __iadd__(self, other):
        self.value = binaryop(operator.add, self.value, other.value)
        return self

    def __itruediv__(self, other):
        if not other:
            return self
        if issequence(self.value):
            for i in xrange(len(self.value)):
                self.value[i] /= other
        else:
            self.value /= other
        return self
    
class Formula(Vector):
    def comparable(self, other):
        return self.name == other.name and \
               compare(self.dist, other.dist)

    def __eq__(self, other):
        return self.value == other.value

    def __isub__(self, other):
        return self

    def __iadd__(self, other):
        return self

    def __itruediv__(self, other):
        if not other:
            return self
        return self
    
class SimpleDist:
    def __init__(self, sums, squares, samples):
        self.sums = sums
        self.squares = squares
        self.samples = samples

    def display(self, name, desc, flags, precision):
        p = Print()
        p.flags = flags
        p.precision = precision

        if self.samples > 0:
            p.name = name + ".mean"
            p.value = self.sums / self.samples
            p.display()

            p.name = name + ".stdev"
            if self.samples > 1:
                var = (self.samples * self.squares - self.sums ** 2) \
                      / (self.samples * (self.samples - 1))
                if var >= 0:
                    p.value = math.sqrt(var)
                else:
                    p.value = 'NaN'
            else:
                p.value = 0.0
            p.display()

        p.name = name + ".samples"
        p.value = self.samples
        p.display()

    def comparable(self, other):
        return True

    def __eq__(self, other):
        return self.sums == other.sums and self.squares == other.squares and \
               self.samples == other.samples

    def __isub__(self, other):
        self.sums -= other.sums
        self.squares -= other.squares
        self.samples -= other.samples
        return self

    def __iadd__(self, other):
        self.sums += other.sums
        self.squares += other.squares
        self.samples += other.samples
        return self

    def __itruediv__(self, other):
        if not other:
            return self
        self.sums /= other
        self.squares /= other
        self.samples /= other
        return self
    
class FullDist(SimpleDist):
    def __init__(self, sums, squares, samples, minval, maxval,
                 under, vec, over, min, max, bsize, size):
        self.sums = sums
        self.squares = squares
        self.samples = samples
        self.minval = minval
        self.maxval = maxval
        self.under = under
        self.vec = vec
        self.over = over
        self.min = min
        self.max = max
        self.bsize = bsize
        self.size = size

    def display(self, name, desc, flags, precision):
        p = Print()
        p.flags = flags
        p.precision = precision

        p.name = name + '.min_val'
        p.value = self.minval
        p.display()

        p.name = name + '.max_val'
        p.value = self.maxval
        p.display()

        p.name = name + '.underflow'
        p.value = self.under
        p.display()

        i = self.min
        for val in self.vec[:-1]:
            p.name = name + '[%d:%d]' % (i, i + self.bsize - 1)
            p.value = val
            p.display()
            i += self.bsize

        p.name = name + '[%d:%d]' % (i, self.max)
        p.value = self.vec[-1]
        p.display()


        p.name = name + '.overflow'
        p.value = self.over
        p.display()

        SimpleDist.display(self, name, desc, flags, precision)

    def comparable(self, other):
        return self.min == other.min and self.max == other.max and \
               self.bsize == other.bsize and self.size == other.size

    def __eq__(self, other):
        return self.sums == other.sums and self.squares == other.squares and \
               self.samples == other.samples

    def __isub__(self, other):
        self.sums -= other.sums
        self.squares -= other.squares
        self.samples -= other.samples

        if other.samples:
            self.minval = min(self.minval, other.minval)
            self.maxval = max(self.maxval, other.maxval)
            self.under -= under
            self.vec = map(lambda x,y: x - y, self.vec, other.vec)
            self.over -= over
        return self

    def __iadd__(self, other):
        if not self.samples and other.samples:
            self = other
            return self

        self.sums += other.sums
        self.squares += other.squares
        self.samples += other.samples

        if other.samples:
            self.minval = min(self.minval, other.minval)
            self.maxval = max(self.maxval, other.maxval)
            self.under += other.under
            self.vec = map(lambda x,y: x + y, self.vec, other.vec)
            self.over += other.over
        return self

    def __itruediv__(self, other):
        if not other:
            return self
        self.sums /= other
        self.squares /= other
        self.samples /= other

        if self.samples:
            self.under /= other
            for i in xrange(len(self.vec)):
                self.vec[i] /= other
            self.over /= other
        return self
    
class Dist(Statistic):
    def display(self):
        if not display_all and not (self.flags & flags_printable):
            return

        self.dist.display(self.name, self.desc, self.flags, self.precision)

    def comparable(self, other):
        return self.name == other.name and \
               self.dist.compareable(other.dist)

    def __eq__(self, other):
        return self.dist == other.dist

    def __isub__(self, other):
        self.dist -= other.dist
        return self

    def __iadd__(self, other):
        self.dist += other.dist
        return self

    def __itruediv__(self, other):
        if not other:
            return self
        self.dist /= other
        return self
    
class VectorDist(Statistic):
    def display(self):
        if not display_all and not (self.flags & flags_printable):
            return

        if isinstance(self.dist, SimpleDist):
            return

        for dist,sn,sd,i in map(None, self.dist, self.subnames, self.subdescs,
                                range(len(self.dist))):
            if len(sn) > 0:
                name = '%s.%s' % (self.name, sn)
            else:
                name = '%s[%d]' % (self.name, i)

            if len(sd) > 0:
                desc = sd
            else:
                desc = self.desc

            dist.display(name, desc, self.flags, self.precision)

        if (self.flags & flags_total) or 1:
            if isinstance(self.dist[0], SimpleDist):
                disttotal = SimpleDist( \
                    reduce(sums, [d.sums for d in self.dist]),
                    reduce(sums, [d.squares for d in self.dist]),
                    reduce(sums, [d.samples for d in self.dist]))
            else:
                disttotal = FullDist( \
                    reduce(sums, [d.sums for d in self.dist]),
                    reduce(sums, [d.squares for d in self.dist]),
                    reduce(sums, [d.samples for d in self.dist]),
                    min([d.minval for d in self.dist]),
                    max([d.maxval for d in self.dist]),
                    reduce(sums, [d.under for d in self.dist]),
                    reduce(sums, [d.vec for d in self.dist]),
                    reduce(sums, [d.over for d in self.dist]),
                    dist[0].min,
                    dist[0].max,
                    dist[0].bsize,
                    dist[0].size)

            name = '%s.total' % (self.name)
            desc = self.desc
            disttotal.display(name, desc, self.flags, self.precision)

    def comparable(self, other):
        return self.name == other.name and \
               alltrue(map(lambda x, y : x.comparable(y),
                           self.dist,
                           other.dist))

    def __eq__(self, other):
        return alltrue(map(lambda x, y : x == y, self.dist, other.dist))

    def __isub__(self, other):
        if issequence(self.dist) and issequence(other.dist):
            for sd,od in zip(self.dist, other.dist):
                sd -= od
        else:
            self.dist -= other.dist
        return self

    def __iadd__(self, other):
        if issequence(self.dist) and issequence(other.dist):
            for sd,od in zip(self.dist, other.dist):
                sd += od
        else:
            self.dist += other.dist
        return self

    def __itruediv__(self, other):
        if not other:
            return self
        if issequence(self.dist):
            for dist in self.dist:
                dist /= other
        else:
            self.dist /= other
        return self
    
class Vector2d(Statistic):
    def display(self):
        if not display_all and not (self.flags & flags_printable):
            return
        d = VectorDisplay()
        d.__dict__.update(self.__dict__)

        if self.__dict__.has_key('ysubnames'):
            ysubnames = list(self.ysubnames)
            slack = self.x - len(ysubnames)
            if slack > 0:
                ysubnames.extend(['']*slack)
        else:
            ysubnames = range(self.x)

        for x,sname in enumerate(ysubnames):
            o = x * self.y
            d.value = self.value[o:o+self.y]
            d.name = '%s[%s]' % (self.name, sname)
            d.display()

        if self.flags & flags_total:
            d.value = []
            for y in range(self.y):
                xtot = 0.0
                for x in range(self.x):
                    xtot += self.value[y + x * self.x]
                d.value.append(xtot)

            d.name = self.name + '.total'
            d.display()

    def comparable(self, other):
        return self.name == other.name and self.x == other.x and \
               self.y == other.y

    def __eq__(self, other):
        return True

    def __isub__(self, other):
        return self

    def __iadd__(self, other):
        return self

    def __itruediv__(self, other):
        if not other:
            return self
        return self
    
class Collection:
    class Node: pass
    def __init__(self, name, subname, bin, host, date, *args):
        self.name = name
        self.subname = subname
        self.bin = bin
        self.host = host
        self.date = date
        self.dict = {}
        self.list = []
        if len(args) > 1:
            sys.exit('too many arguments')

        if len(args):
            for stat in args[0]:
                self.append(stat)

    def append(self, stat):
        statname = re.sub(':', '__', stat.name)
        path = string.split(statname, '.')

        x = self
        while len(path) > 1:
            name = path.pop(0)
            if not x.__dict__.has_key(name):
                x.__dict__[name] = Collection.Node()
            x = x.__dict__[name]

        name = path.pop(0)
        x.__dict__[name] = stat

        self.dict[statname] = stat
        self.list.append(statname)

    def display_one(self, stat):
        if isinstance(stat, Formula):
            if not len(stat.formula):
                return
            formula = re.sub(':', '__', stat.formula)
            fstat = eval(formula, globals(), self.__dict__)
            if isinstance(fstat, FormulaStat):
                stat.value = fstat.value
            else:
                stat.value = fstat
            stat.display()
        else:
            if display_all or not stat.__dict__.has_key('prereq') or \
                   not self.dict[stat.prereq].zero():
                stat.display()

    def display(self, stats, regexps):
        list = []
        if len(stats) == 0 and len(regexps) == 0:
            list.extend(self.list)
            list.sort(statcmp)
        else:
            dict = {}
            for stat in stats:
                dict[stat] = 1
            list.extend(stats)
            for regexp in regexps:
                for stat in self.list:
                    if not dict.has_key(stat) and regexp.match(stat):
                        dict[stat] = stat
                        list.append(stat)
            list.sort(statcmp)

        bin = ''
        if len(self.bin):
            bin = '(%s)' % self.bin
        print 'statistics dump %s%s %s\ngenerated using %s on %s' % \
              (self.name, bin, self.subname, self.host, self.date)
        for name in list:
            if self.dict.has_key(name):
                stat = self.dict[name]
                if isinstance(stat, Statistic):
                    self.display_one(stat)

    def same_run(self, other):
        return self.name == other.name and self.host == other.host \
               and self.date == other.date

    def same_bin(self, other):
        return self.bin == other.bin

def collection_total(collections):
    if len(collections) == 1:
        return collections[0]
    
    if not alltrue(map(lambda x: isinstance(x, Collection), collections)):
        sys.exit("can only add a Collection to another Collection")

    name = 'total'
    subname = 'total'
    bin = 'total'
    nocompare = True

    if alltrue(map(lambda x: x.name == collections[0].name, collections[1:])):
        name = collections[0].name
        if alltrue(map(lambda x: x.subname == collections[0].subname,
                       collections[1:])):
            subname = collections[0].subname
            nocompare = False

    if alltrue(map(lambda x: x.bin == collections[0].bin, collections[1:])):
        bin = collections[0].bin

    import socket, time
    host = socket.gethostname()
    date = time.ctime(time.time())

    dict = {}

    for collection in collections:
        for statname in collection.list:
            stat = collection.dict[statname]
            if isinstance(stat, Formula):
                dict[statname] = stat
            else:
                if nocompare or stat.binned:
                    if dict.has_key(statname):
                        dict[statname] += stat
                    else:
                        dict[statname] = copy.deepcopy(stat)
                else:
                    if not dict.has_key(statname):
                        dict[statname] = stat
                    elif dict[statname] != stat and not nocompare:
                        sys.exit('error: non binned stats should be the same!')

    result = Collection(name, subname, bin, host, date)
    keys = dict.keys()
    keys.sort(statcmp)
    for k in keys:
        result.append(dict[k])

    return result

def collection_average(collections):
    if len(collections) == 1:
        return collections[0]
    
    if not alltrue(map(lambda x: isinstance(x, Collection), collections)):
        sys.exit("can only add a Collection to another Collection")

    name = 'average'
    subname = 'average'
    bin = 'average'
    nocompare = True

    if alltrue(map(lambda x: x.name == collections[0].name, collections[1:])):
        name = collections[0].name
        if alltrue(map(lambda x: x.subname == collections[0].subname,
                       collections[1:])):
            subname = collections[0].subname
            nocompare = False

    if alltrue(map(lambda x: x.bin == collections[0].bin, collections[1:])):
        bin = collections[0].bin

    import socket, time
    host = socket.gethostname()
    date = time.ctime(time.time())

    dict = {}
    count = {}

    import copy
    for collection in collections:
        for statname in collection.list:
            stat = collection.dict[statname]
            if isinstance(stat, Formula):
                dict[statname] = copy.deepcopy(stat)
                count[statname] = 1
            else:
                if nocompare or stat.binned:
                    if dict.has_key(statname):
                        dict[statname] += copy.deepcopy(stat)
                        count[statname] += 1
                    else:
                        dict[statname] = copy.deepcopy(stat)
                        count[statname] = 1
                else:
                    if not dict.has_key(statname):
                        dict[statname] = copy.deepcopy(stat)
                        count[statname] = 1
                    elif dict[statname] != stat and not nocompare:
                        sys.exit('error: non binned stats should be the same!')

    result = Collection(name, subname, bin, host, date)
    keys = dict.keys()
    keys.sort(statcmp)
    for k in keys:
        dict[k] /= count[k]
        result.append(dict[k])

    return result

def usage():
    if usage.__dict__.has_key('extra'):
        extra = usage.extra
    else:
        extra = ''

    print '''
Usage: %(progname)s [-A] [-D] [-B <regexp>] [-b <bins>] [-N <regexp>]
       [-n <subnames>] [-S <regexp>] [-s <stats>] [-T{a|b|n} | -t{a|b|n}]
       %(extra)s
       -A             Display All
       -D             Print Descriptions
       -B <regexp>    Print only the bins matching the regular expression
       -b <bins>      Print only the specified bins (comma separated list)
       -N <regexp>    Print only the subnames matching the regular expression
       -n <subnames>  Print only the specified subnames (comma separated list)
       -S <regexp>    Print only the stats matching the regular expression
       -s <stats>     Print only the specified stats (comma seprated list)
       -T[a|b|n]      total all/bins/subnames
       -t[a|b|n]      Print only total of all/bins/subnames
''' % { 'progname' : sys.argv[0], 'extra' : extra }
    sys.exit(1)


def parse_options(argv):
    try:
        import getopt
        opts, args = getopt.getopt(argv, '-ADB:b:N:n:S:s:T:t:')
    except getopt.GetoptError:
        usage()

    class Options : pass
    options = Options()
    options.args = args
    options.bin_regexps = []
    options.bins = {}
    options.subnames_regexps = []
    options.subnames = {}
    options.stat_regexps = []
    options.stats = []
    options.total_all = False
    options.total_bins = False
    options.total_subnames = False
    options.only_total = False

    for o,a in opts:
        if o == '-A':
            global display_all
            display_all = True
        if o == '-D':
            global descriptions
            descriptions = True
        if o == '-B':
            options.bin_regexps.append(re.compile(a))
        if o == '-b':
            for bin in a.split(','):
                options.bins[bin] = 1
        if o == '-N':
            options.subnames_regexps.append(re.compile(a))
        if o == '-n':
            for subname in a.split(','):
                options.subnames[subname] = 1
        if o == '-S':
            options.stat_regexps.append(re.compile(a))
        if o == '-s':
            options.stats.extend(map(lambda x: x.strip(), a.split(',')))
        if o == '-T' or o == '-t':
            if o == '-t':
                options.only_total = True

            if a == 'a':
                options.total_all = True
            elif a == 'b':
                options.total_bins = True
            elif a == 'n':
                options.total_subnames = True
            else:
                usage()
                
    return options

def display_stats(options):
    list = []
    list.extend(collections)

    if len(options.subnames) > 0:
        templist = list
        list = []
        for item in templist:
            if options.subnames.has_key(item.subname):
                list.append(item)

    if len(options.bins) > 0:
        templist = list
        list = []
        for item in templist:
            if options.bins.has_key(item.bin):
                list.append(item)

    if len(options.subnames_regexps) > 0:
        templist = list
        list = []
        for item in templist:
            for exp in options.subnames_regexps:
                if exp.match(item):
                    list.append(item)
                    break

    if len(options.bin_regexps) > 0:
        templist = list
        list = []
        for item in templist:
            for exp in options.bin_regexps:
                if exp.match(item):
                    list.append(item)
                    break

    stats = options.stats
    regexps = options.stat_regexps
    if not options.only_total:
        for item in list:
            item.display(stats, regexps)

    if options.total_bins:
        bins = {}
        for item in list:
            if not bins.has_key(item.bin):
                bins[item.bin] = []
            bins[item.bin].append(item)

        values = bins.values()
        values.sort()
        
        for bin in values:
            total = collection_average(bin)
            total.bin = 'total'
            total.display(stats, regexps)

    if options.total_subnames:
        subnames = {}
        for item in list:
            if not subnames.has_key(item.subname):
                subnames[item.subname] = []
            subnames[item.subname].append(item)

        
        values = subnames.values()
        values.sort()
        for subname in values:
            total = collection_average(subname)
            total.bin = 'total'
            total.display(stats, regexps)
        
    if options.total_all:
        total = collection_average(list)
        total.subname = 'total'
        total.bin = 'total'
        total.display(stats, regexps)

def program_display():
    options = parse_options(sys.argv[1:])
    if len(options.args) > 0:
        usage()
    display_stats(options)

if __name__ == '__main__':
    usage.extra = ' <stats files>'
    options = parse_options(sys.argv[1:])

    numargs = len(options.args)
    if numargs == 0:
        usage()

    __name__ = "stats"
    for file in options.args:
        execfile(file)

    display_stats(options)
